#!/bin/bash
set -e #-x

# DO NOT ADD NEW EXPORTED VARIABLES HERE

# Add new operands in this list.
available_operands="custom rootfromvcs nosync regendefconfig noconf nocleanstaging nobuild notrackws noforcesync noccache fast enable_tracing component_size createrootpath continuous_integration opensource_only noautoconf strip_kernel"
available_commands="fullclean buildimage buildclean synconly doc tests createpackage buildpackage modular_packages"

usage() {
  echo "Usage:" >&2
  echo "  $0 <project directory> [<command>|<modifier>]..." >&2
  echo "  $0 help [<operand>]" >&2
  echo "where command is one of: $available_commands" >&2
  echo "where modifier is one or more of: $available_operands" >&2
  echo "-- commands --" >&2
  echo "  nosync : do not perform a sync with the version control system. Can be used offline" >&2
  echo "  nocleanstaging : do not clean the staging directory" >&2
  echo "  nobuild : do not perform an actual build. Only sync the sources from version control and prepare the build" >&2
  echo "  fast : perform a parallel build" >&2
  echo "-- operands --" >&2
  echo "  createpackage : creates an offline buildable package with both binaries and sources as defined in the project.list" >&2
  echo "  buildpackage : builds a package created with createpackage. Can be used offline" >&2
  exit 1
}

listcontains() {
    local word
    for word in $1; do 
       [[ "$word" == "$2" ]] && return 0
  done
    return 1
}

if [ -z "$1" ] || [ ! -d "$1" ]; then
  if [ ! "help" = "$1" ]; then
  usage
fi
fi

commands=""
endmessage=""
opt_createpackage=""
opt_buildpackage=""

parse_arguments() {
    prjdef=$1
    # Set variables to "1" for the optional parameters. Name those variables to the parameters
    shift 1
    while [ -n "$1" ]; do
      if listcontains "$available_commands" $1; then
         if [ -n "$commands" ]; then
             echo "Not possible to specify more than 1 command. Command $commands allready active"
	     usage
    	 else
             commands=$1
	 fi
      elif listcontains "$available_operands" $1; then
         echo $1 is set
         export $1=1
      else
         echo "$1: invalid operand."
         exit 1
      fi
      shift 1
    done
}

initializeEnvironment() {
curdir=`pwd`
prjdir=`cd $prjdef;pwd`
starttime=`date +%Y-%m-%d_%H:%M:%S`

echo $prjdir > prjdir.txt
wsdir=`echo "${prjdir%config*}"`

echo "-- Build started at: ${starttime}"
echo "-- Project directory: ${prjdir}"
echo "-- Workspace directory = $wsdir"

prjdep="$prjdir/project.list"
defconfig="$prjdir/defconfig"
hardcofile="$prjdir/hardco.cfg"
osrcfile="$prjdir/opensource_include.lst"
compexclfile="$prjdir/component_exclude.lst"

for f in "$prjdep"; do
  if ! [ -r "$f" ]; then
    n=$(basename "$f")
    echo "unable to read config file $n"
    echo "check that it exists in $prjdef"
    exit 1
  fi
done

if [ -f server.cfg ]; then
  servercfgfile="server.cfg"
else
  if [ -f "$HOME/.server.cfg" ]; then
   servercfgfile="$HOME/.server.cfg"
  else
    if [ -f "$HOME/server.cfg" ]; then
      servercfgfile="$HOME/server.cfg"
    else
      if [ -f "$wsdir/.server.cfg" ]; then
        servercfgfile="$wsdir/.server.cfg"  
      else
        echo -e "-- \033[31mERROR: server.cfg not found.\033[0m This file should be in this directory or"
        echo "-- create a file .server.cfg or server.cfg in your home directory."
        echo "-- It should contain a list of perforce servers and your workspace configuration."
        echo "-- Check the sample directory for sample files."
        exit 1
      fi
    fi
  fi
fi
echo "-- Server configuration file: $servercfgfile"

if [ -n "${OUTPUTDIR}" ]; then
  if [ "`echo ${OUTPUTDIR} | cut -c1`" = "/" ]; then
    outputdir="${OUTPUTDIR}"
  else
    outputdir="`(cd "${OUTPUTDIR}"; pwd)`"
  fi
else
  outputdir="$curdir/output"
fi
echo "-- Output directory: $outputdir"

STAGINGDIR=$outputdir/staging
echo "-- Staging directory: $STAGINGDIR"

# Only auto-clean staging dir when noconf isn't set
if [ "${nocleanstaging}" != "1" ] && [ "${noconf}" != "1" ] && [ "${tests}" != "1" ] && [ "${opt_buildpackage}" != "-w" ]; then
  echo "-- Cleaning $STAGINGDIR"
  rm -rf $STAGINGDIR
fi

mkdir -p $STAGINGDIR

RELEASEDIR=$outputdir/release
DELIVERYDIR=$RELEASEDIR/delivery
BUILD_CUSTOMER=""
BUILD_PROJECT=""
BUILD_HARDWARE=""
BUILD_CONFIGTYPE=""
k=0
for i in `echo $prjdef |sed 's/\// /g'`; do
    arr[$k]=$i
    let k=$k+1
done
# n is total number of directories found
n=${#arr[*]}
BUILD_CUSTOMER=${arr[$n-4]}
BUILD_PROJECT=${arr[$n-3]}
BUILD_HARDWARE=${arr[$n-2]}
BUILD_CONFIGTYPE=${arr[$n-1]}

if [ "${nocleanstaging}" != "1" ]; then
  # remove the root filesystem directory
  echo "-- Cleaning $RELEASEDIR/root_fs"
  rm -rf $RELEASEDIR/root_fs
fi

echo "-- Release directory: $RELEASEDIR"
mkdir -p $RELEASEDIR

spinnumber=`grep CONFIG_BUILD_SPIN_NUMBER= $defconfig | sed -e 's/CONFIG_BUILD_SPIN_NUMBER=//' | sed -e 's/\"//g' `
if [ ${#spinnumber} -eq 0 ]; then
    spinnumber="00"
fi

echo "-- Spin number: $spinnumber"

bldcfg="buildconfig.default"
versioninfo="versioninfo.txt"
licenceinfo="licenceinfo.txt"
changelistinfo="changelistinfo.txt"
buildinfo="$STAGINGDIR/buildconfig.info"

# Store some information for use in the build information of the image.
rm -f $changelistinfo
rm -f "${buildinfo}"
echo "BUILD_VER=`echo $curdir | sed -e 's/.*\/build\///'`" >> "${buildinfo}"
echo "BUILD_DEF=${prjdef}" >> "${buildinfo}"
echo "BUILD_START=${starttime}" >> "${buildinfo}"
echo "BUILD_HOST=`uname -n`" >> "${buildinfo}"
echo "BUILD_USER=$USER" >> "${buildinfo}"
echo "BUILD_CUSTOMER=$BUILD_CUSTOMER" >> "${buildinfo}"
echo "BUILD_PROJECT=$BUILD_PROJECT" >> "${buildinfo}"
echo "BUILD_HARDWARE=$BUILD_HARDWARE" >> "${buildinfo}"
echo "BUILD_CONFIGTYPE=$BUILD_CONFIGTYPE" >> "${buildinfo}"
echo "BUILD_SPIN_NUMBER=$spinnumber" >> "${buildinfo}"

if [ "${custom}" = "1" ]; then
  echo "-- Ready for customisation, press enter to start build."
  read
fi

if [ "${strip_kernel}" == "1" ]; then
    if [ "${opt_createpackage}" == "-z" ]; then
        echo "-- Stripping kernel at the end of the build process"
    else
        echo "-- Strip Kernel only possible when creating a package. Disabling option."
        strip_kernel=""
    fi
fi

opt_notrackws=""
if [ "${notrackws}" == "1" ]; then
  opt_notrackws="-n"
fi

opt_tracing=""
if [ "${enable_tracing}" == 1 ] ; then
  opt_tracing="-t"
fi

opt_component_size=""
if [ "${component_size}" == 1 ] ; then
  opt_component_size="-x"
fi

opt_noforcesync=""
if [ "${noforcesync}" == "1" ]; then
  opt_noforcesync="-nf"
fi

opt_fast=""
if [ "${fast}" == "1" ]; then
  if [ "${opt_createpackage}" == "-z" ]; then
    echo "-- parallel build not supported when creating package - deactivating parallel build"
    opt_fast=""
    fast=""
  else
    echo "-- parallel build enabled" 
    opt_fast="-f"
  fi
fi

opt_ci=""
if [ "${continuous_integration}" == "1" ]; then
  opt_ci="-ci"
fi

opt_opensource_only=""
if [ "${opensource_only}" == "1" ]; then
  opt_opensource_only="-opensource_only"
fi

opt_nosync=""
if [ "${nosync}" == "1" ]; then
  opt_nosync="-nosync"
fi

if [ "${createrootpath}" == "1" ]; then
    ./build_rootpathcfg -s $servercfgfile
fi

if [ "${opt_buildpackage}" == "-w" ]; then
    ./build_rootpathcfg_package $wsdir
fi
}

help_modular_packages() {
    cat >&2 << EOF
    Synopsis: $0 modular_packages

    Dries will provide some explanations about what this command does
EOF
}

command_modular_packages() {
  initializeEnvironment
  echo "-- Get source code (original config)"
  # if hardcofile is present in package config, use that one,
  # otherwise take the original file
  if [ -e "$hardcofile" ]; then
    HARDCOFILE=$hardcofile
  else
    HARDCOFILE=$prjdir/../../../hardco.cfg
  fi
  ./modular_packages.tcl $opt_notrackws -defconfig $defconfig -s $servercfgfile -p $prjdep -h $HARDCOFILE -o $outputdir -b $bldcfg $opt_rootfromvcs $ccache_enable $legacy_releasedir $opt_tracing $opt_createpackage $opt_component_size $opt_ci $opt_fast $opt_opensource_only  
  cat ${prjdir}/project.list | grep -v sah_packages_ | grep -v "#" | sed 's/\>.*//g' | sed 's/=.*//g' > ./output/staging/modpkg_dependencies
  install -m 644 ./scripts/modpkg.mk ${STAGINGDIR}
  echo 'include $(STAGINGDIR)/modpkg.mk' >> ${STAGINGDIR}/sahcommon.mk

  install -d ${STAGINGDIR}/config
  chmod -R +w ${STAGINGDIR}/config
  cp -aR $prjdir/../../../* ${STAGINGDIR}/config
  chmod -R +w ${STAGINGDIR}/config
  cp -aR $prjdir/* ${STAGINGDIR}/config
  chmod -R +w ${STAGINGDIR}/config
  install -d ${STAGINGDIR}/utils
  cp -aR $curdir/utils/* ${STAGINGDIR}/utils
  chmod -R +wx ${STAGINGDIR}/utils
  originaldefconfig=$(dirname $defconfig|awk '{print  $1"/../../../defconfig"}')
  install -m 644 $originaldefconfig .config
  cat $defconfig >> .config
  install -m 664 -p .config ${STAGINGDIR}/components.config
  make install-component-config
  ./scripts/install_scripts/install_to_staging.sh ${STAGINGDIR}

  # start making the package(s)
  make -f Makefile.components modular_packages

  rootdir=`head -n1 rootpath.cfg | cut -d= -f2`
  ln -s $outputdir $rootdir/output

  exit 0
}

syncsource() {
    initializeEnvironment
if [ "${regendefconfig}" == "1" ]; then
  echo "-- Get source code"
    ./getsource.tcl $opt_notrackws $opt_noforcesync $opt_ci -s $servercfgfile -p $prjdep -r -defconfig $defconfig
  # Create fixed link to the output dir
  rootdir=`head -n1 rootpath.cfg | cut -d= -f2`
  rm -f $rootdir/output
  ln -s $outputdir $rootdir/output
elif [ "${nosync}" != "1" ]; then
  echo "-- Get source code"
  getsource_cmd="./getsource.tcl $opt_notrackws $opt_noforcesync $opt_ci $opt_tracing $opt_buildpackage $opt_createpackage $opt_component_size -s $servercfgfile -p $prjdep -defconfig $defconfig -osrc $osrcfile -compexcl $compexclfile $co$opt_opensource_only"
  echo ${getsource_cmd}
  ${getsource_cmd}
  # Create fixed link to the output dir
  rootdir=`head -n1 rootpath.cfg | cut -d= -f2`
  rm -f $rootdir/output
  ln -s $outputdir $rootdir/output
fi
}

help_synconly() {
    cat >&2 << EOF
    Synopsis: $0 synconly

    Fetch components from version control system after having initialized the staging dir.
    
    Warning: by default, this command will overwrite files !
EOF
}

command_synconly() {
    initializeEnvironment
    syncsource
    endmessage="Synchronisation"
  rm -rf "$outputdir"
}

configureBuildTree() {
if [ -f $prjdir/test.config ]
then
  echo "test.config exists (copy $prjdir/test.config to $outputdir/test)"
  install -d $outputdir/test
  install -m 664 $prjdir/test.config $outputdir/test/
fi

if [ "${noconf}" != "1" ]; then
  echo "-- Configure build tree"

  # check for ccache presence
  ccache_where=`whereis ccache | wc -w`
  ccache_block=`grep "CONFIG_BUILD_NO_CCACHE=y" $defconfig | wc -w`
  if [ "$ccache_where" -gt "1" ] && [ "${noccache}" != "1" ] && [ "$ccache_block" != "1" ]; then
    echo "-- Ccache is present: enabling"
    ccache_enable="-c"
  else
    echo "-- Ccache not present"
    ccache_enable=""
  fi

  opt_rootfromvcs=""
  if [ "${rootfromvcs}" == "1" ]; then
    opt_rootfromvcs="-r"
  fi

  if [ "${opensource_only}" == "1" ]; then
    prjdep=cur_prj.list
    echo "Resetting project.list to $prjdep"
  fi

  config_source_cmd="./configsource.tcl -defconfig $defconfig -s $servercfgfile -p $prjdep -h $hardcofile -o $outputdir -b $bldcfg -osrc $osrcfile -compexcl $compexclfile $opt_rootfromvcs $ccache_enable $opt_tracing $opt_nosync $opt_buildpackage $opt_createpackage $opt_component_size $opt_ci $opt_fast $opt_opensource_only"
  echo ${config_source_cmd}
  ${config_source_cmd}
  cat $versioninfo | sort > $STAGINGDIR/buildconfig.dep
  cat $licenceinfo | sort > $STAGINGDIR/licence.txt
  if [ "${opt_buildpackage}" == "" ]; then
  ./gentestconfig.tcl -s $servercfgfile -p $prjdep -h $hardcofile -o $outputdir -b $bldcfg $opt_nosync $opt_rootfromvcs $ccache_enable 
  fi
fi
}

generatePackage() {
  generate_package_cmd="./genPkg.tcl -defconfig $defconfig -s $servercfgfile -p cur_prj.list -h $hardcofile -o $outputdir -b $bldcfg -osrc $osrcfile -compexcl $compexclfile $opt_rootfromvcs $ccache_enable $opt_tracing $opt_nosync $opt_buildpackage $opt_createpackage $opt_component_size $opt_ci $opt_fast $opt_opensource_only"
  echo ${generate_package_cmd}
  ${generate_package_cmd}
}

populateStagingDir() {
### Copy the project dir to the staging dir
install -d ${STAGINGDIR}/config
cp -aR $prjdir/* ${STAGINGDIR}/config
chmod -R +w ${STAGINGDIR}/config

### Copy the utils dir to the staging dir
install -d ${STAGINGDIR}/utils
cp -aR $curdir/utils/* ${STAGINGDIR}/utils
chmod -R +wx ${STAGINGDIR}/utils
}

installBuildConfiguration() {
### Install build configuration
echo "-- Install build configuration"
if [ "${opensource_only}" == "1" ]; then
    install -m 664 -p rebuild_defconfig .config
else
    install -m 664 -p $defconfig .config
fi
# components.config is needed to be able to run Makefile (since its is included from Makefile.global)
install -m 664 -p .config ${STAGINGDIR}/components.config
if grep -q "^# Automatically generated make config: don't edit" .config; then
    make install-component-config
else
    # this is a minimal config with only basic project details, automatically fill it with defaults
    make oldconfig </dev/null
fi

if [ "${noautoconf}" != "1" ]; then
    make install-autoconf-config
fi
    ./scripts/install_scripts/install_to_staging.sh ${STAGINGDIR}

}

preparebuild() {
    initializeEnvironment
    syncsource

    configureBuildTree
    populateStagingDir
    installBuildConfiguration
    if [ "${opt_createpackage}" == "-z" ]; then
      generatePackage
    fi
}

preparepackagebuild() {
    initializeEnvironment
    syncsource
    configureBuildTree
    populateStagingDir
    installBuildConfiguration
    if [ "${opt_createpackage}" == "-z" ]; then
      generatePackage
    fi
}

help_fullclean() {
cat >&2 << EOF
Synopsis: $0 <project directory> fullclean

Delete all build artefacts and other untracked files. This allows to make a build equivalent to a build from scratch.
EOF
}

command_fullclean() {
    initializeEnvironment
    echo "-- Get source code"
    getsource_cmd="./getsource.tcl -fullclean $opt_notrackws $opt_ci $opt_tracing -s $servercfgfile -p $prjdep -defconfig $defconfig"
    echo ${getsource_cmd}
    ${getsource_cmd}
    # Create fixed link to the output dir
    rootdir=`head -n1 rootpath.cfg | cut -d= -f2`
    rm -f $rootdir/output
    rm -rf $outputdir
}

help_doc() {
cat >&2 << EOF
Synopsis: $0 <project directory> doc

  Perform all steps necessary to build the documentation:
  1. Compute dependencies
  2. Fetch components from version control system (overwrites files !)
  3. Perform a build to populate the staging directory
  4. Build the documentation
EOF
}

command_doc() {
  command_buildimage

  echo "-- Build documentation"
  make doc
  endmessage="Building documentation"
}

help_tests() {
cat >&2 << EOF
Synopsis: $0 <project directory> tests

  Perform all steps necessary to build the tests:
  1. Compute dependencies
  2. Fetch components from version control system (overwrites files !)
  3. Build the tests
EOF
}

command_tests() {
    preparebuild
  echo "-- Build tests"
  make -i tests
    endmessage="Building tests"
}


help_buildimage() {
cat >&2 << EOF
Synopsis: $0 <project directory> buildimage

  Perform all steps necessary to build an image:
  1. Compute dependencies
  2. Fetch components from version control system (overwrites files !)
  3. Build each component
  4. Create root file system
  5. Generate an image from the root file system
EOF
}

command_buildimage() {
    preparebuild

### Activate parallel build
JOBS=""
if [ "${fast}" = "1" ]; then
    if [ "$PARALLEL" != "" ]; then
        CPUS=$PARALLEL
    else
        CPUS=$(grep processor /proc/cpuinfo|wc -l)
    fi
    JOBS="-j$CPUS"
    echo "-- Fast mode activated (${CPUS} cpu(s))"
else
    echo "-- Fast mode not activated or deactivated by createpackage"
fi

if [ "${nobuild}" = "1" ]; then
  endtime=`date +%Y-%m-%d_%H:%M:%S`
  echo "-- Synchronisation and configuration completed at ${endtime}"
  echo "BUILD_END=${endtime}" >> "${buildinfo}"
  exit 0
fi


### Build all components
echo "-- Build all components"
make ${JOBS}

### Create image
if [ "${A}" != "1" ]; then  
  echo "-- Create image"
  make image  
else
  echo "-- CLANG check is enabled, do not create an image"
fi

    ### Build ready
    endmessage="Build"
}

help_buildclean() {
cat >&2 << EOF
Synopsis: $0 <project directory> buildimage

  Perform all steps necessary to clean all components:
  1. Compute dependencies
  2. Fetch components from version control system (overwrites files !)
  3. Clean each component
EOF
}

command_buildclean() {
    preparebuild

### Activate parallel build
JOBS=""
if [ "${fast}" = "1" ]; then
    if [ "$PARALLEL" != "" ]; then
        CPUS=$PARALLEL
    else
        CPUS=$(grep processor /proc/cpuinfo|wc -l)
    fi
    JOBS="-j$CPUS"
    echo "-- Fast mode activated (${CPUS} cpu(s))"
fi

if [ "${nobuild}" = "1" ]; then
  endtime=`date +%Y-%m-%d_%H:%M:%S`
  echo "-- Synchronisation and configuration completed at ${endtime}"
  echo "BUILD_END=${endtime}" >> "${buildinfo}"
  exit 0
fi


### Build all components
echo "-- Build all components"
make ${JOBS} clean

    ### Build ready
    endmessage="Build"
}

help_createpackage() {
cat >&2 << EOF
Synopsis: $0 <project directory> createpackage

  Perform all steps necessary to create a package:
  1. Compute dependencies
  2. Fetch components from version control system (overwrites files !)
  3. Build each component
  4. Create root file system
  5. Generate an image from the root file system
  6. Create a package from which an image can be built
EOF
}

command_createpackage() {
  opt_createpackage="-z"
  command_buildimage
  echo "-- Copy the staging directory to the package directory"
  ./createPackageCopyStagingDir.sh
  echo "-- Copy the config directory to the package directory"
  install -d output/package/config/$BUILD_CUSTOMER/$BUILD_PROJECT/$BUILD_HARDWARE/$BUILD_CONFIGTYPE
  cp -aR $prjdir/* output/package/config/$BUILD_CUSTOMER/$BUILD_PROJECT/$BUILD_HARDWARE/$BUILD_CONFIGTYPE
  if [ "${opensource_only}" == "1" ]; then
    install cur_prj.list output/package/config/$BUILD_CUSTOMER/$BUILD_PROJECT/$BUILD_HARDWARE/$BUILD_CONFIGTYPE/project.list
    install .config output/package/config/$BUILD_CUSTOMER/$BUILD_PROJECT/$BUILD_HARDWARE/$BUILD_CONFIGTYPE/defconfig
  fi
  chmod -R +w output/package/config
  if [ "${strip_kernel}" == 1 ] ; then
    echo "-- Strip the kernel"
    # look for broadcom kernel
    line=`cat Makefile.components | grep "sah_kernel_broadcom -f Component.mk clean"`
    if [ "${line}" != "" ] ; then
        echo "## Broadcom kernel found!"
        arr=($line)
        kernel_location=${arr[2]}
        cp tools/do_consumer_release_sah $kernel_location/release
        cd $kernel_location/release
        echo "## Starting kernel strip"
        ./do_consumer_release_sah -y
        echo "## copying stripped kernel to package"
        rm -rf $curdir/output/package/sah_kernel_broadcom
        cd $kernel_location/..
        cp -prf sah_kernel_broadcom output/package/
        cd $curdir
    else
        # look for ikanos kernel
        line1=`cat Makefile.components | grep "sah/kernel/IKANOS" | grep install`
        line2=`cat Makefile.components | grep "sah_kernel_ikanos -f Component.mk clean"`
        if [ "${line1}" != "" -o "${line2}" != "" ] ; then
            echo "## Ikanos kernel found! Stripping not (yet) supported"
        else
            echo "## No supported kernel found. Unable to strip"
        fi
    fi
  fi
  echo "-- Tar the package directory"
  rm -rf output/release/delivery/Package.tar.gz
  mkdir -p output/release/delivery
  tar -C output/package -zcf output/release/delivery/Package.tar.gz .
  echo "-- Remove temporary location for package build"
  tmp_build=`grep "# tmp path"  Makefile.components | cut -c 14-`
  rm -rf $tmp_build
  endmessage="Build"
}

help_buildpackage() {
cat >&2 << EOF
Synopsis: $0 <project directory> buildpackage

  Perform all steps necessary to build a package:
  1. Compute dependencies
  2. Build each component
  3. Create root file system
  4. Generate an image from the root file system
EOF
}

command_buildpackage() {
  opt_buildpackage="-w"
  preparepackagebuild
### Activate parallel build
  JOBS=""
  if [ "${fast}" = "1" ]; then
	if [ "$PARALLEL" != "" ]; then
		CPUS=$PARALLEL
	else
		CPUS=$(grep processor /proc/cpuinfo|wc -l)
	fi
	JOBS="-j$CPUS"
	echo "-- Fast mode activated (${CPUS} cpu(s))"
  fi

  if [ "${nobuild}" = "1" ]; then
	endtime=`date +%Y-%m-%d_%H:%M:%S`
	echo "-- Synchronisation and configuration completed at ${endtime}"
	echo "BUILD_END=${endtime}" >> "${buildinfo}"
	exit 0
  fi


### Build all components
  echo "-- Build all components"
  make ${JOBS}

### Create image
  if [ "${A}" != "1" ]; then  
  echo "-- Create image"
  make image  
  else
  echo "-- CLANG check is enabled, do not create an image"
  fi

### Build ready
  endmessage="Build"
}

if [ "help" = "$1" ]; then
    if [[ ! -z "$2" ]] || listcontains "$available_commands""$2"; then
        help_$2
    else
        usage
    fi
    exit 1
else
    parse_arguments $@
    if [ -z $commands ]; then
       command_buildimage
    else
       for i in $commands; do
          command_$i
       done
fi
    if [ ! -z "$endmessage" ]; then
endtime=`date +%Y-%m-%d_%H:%M:%S`
        echo "-- $endmessage completed at ${endtime}"
        if [ -d ${DELIVERYDIR} ]; then
            echo "-- creating checksum on all files in $DELIVERYDIR"
            cd $DELIVERYDIR
            md5sum * > checklist.chk
        fi
        if [ ! -z "${buildinfo}" ] && [ -f "${buildinfo}" ]; then
echo "BUILD_END=${endtime}" >> "${buildinfo}"
        fi
    fi
fi
