#!/bin/bash
set -e #-x

# DO NOT ADD NEW EXPORTED VARIABLES HERE

# Add new operands in this list.
available_operands="opensource_only"
available_commands="clean"

usage() {
  echo "Usage:" >&2
  echo "  $0 <project directory> [<command>]..." >&2
  echo "  $0 help" >&2
  echo "No extra Command:" >&2
  echo "  Perform all steps necessary to build an image:" >&2
  echo "  1. Compute dependencies" >&2
  echo "  2. Build each source-component or untar each binary component" >&2
  echo "  3. Create root file system" >&2
  echo "  4. Generate an image from the root file system" >&2
  echo "opensource_only: build only the open-source part." >&2
  echo "clean: clean up all components" >&2
  exit 1
}

listcontains() {
    local word
    for word in $1; do
       [[ "$word" == "$2" ]] && return 0
  done
    return 1
}

if [ -z "$1" ] || [ ! -d "$1" ]; then
  if [ ! "help" = "$1" ]; then
  usage
fi
fi

commands=""
endmessage=""
opt_buildpackage="-w"

parse_arguments() {
    prjdef=$1
    shift 1
    while [ -n "$1" ]; do
      if listcontains "$available_commands" $1; then
         commands+=" "$1
     elif listcontains "$available_operands" $1; then
        echo $1 is set
        export $1=1
     else
        echo "$1: invalid operand."
        exit 1
      fi
      shift 1
    done
}

initializeEnvironment() {
curdir=`pwd`
prjdir=`cd $prjdef;pwd`
starttime=`date +%Y-%m-%d_%H:%M:%S`

echo $prjdir > prjdir.txt
wsdir=`awk -Fconfig '{ printf "%s", $1 }' prjdir.txt`

echo "-- Build started at: ${starttime}"
echo "-- Project directory: ${prjdir}"
echo "-- Workspace directory = $wsdir"

prjdep="$prjdir/project.list"
defconfig="$prjdir/defconfig"
hardcofile="$prjdir/hardco.cfg"
osrcfile="$prjdir/opensource_include.lst"

for f in "$prjdep"; do
  if ! [ -r "$f" ]; then
    n=$(basename "$f")
    echo "unable to read config file $n"
    echo "check that it exists in $prjdef"
    exit 1
  fi
done

if [ -f server.cfg ]; then
  servercfgfile="server.cfg"
else
  if [ -f "$HOME/.server.cfg" ]; then
   servercfgfile="$HOME/.server.cfg"
  else
    if [ -f "$HOME/server.cfg" ]; then
      servercfgfile="$HOME/server.cfg"
    else
      echo "-- ERROR: server.cfg not found. This file should be in this directory or"
      echo "-- create a file .server.cfg or server.cfg in your home directory."
      echo "-- It should contain a list of perforce servers and your workspace configuration."
      echo "-- Check the sample directory for sample files."
      exit 1
    fi
  fi
fi
echo "-- Server configuration file: $servercfgfile"

if [ -n "${OUTPUTDIR}" ]; then
  if [ "`echo ${OUTPUTDIR} | cut -c1`" = "/" ]; then
    outputdir="${OUTPUTDIR}"
  else
    outputdir="`(cd "${OUTPUTDIR}"; pwd)`"
  fi
else
  outputdir="$curdir/output"
fi
echo "-- Output directory: $outputdir"

STAGINGDIR=$outputdir/staging
echo "-- Staging directory: $STAGINGDIR"

mkdir -p $STAGINGDIR

RELEASEDIR=$outputdir/release
BUILD_CUSTOMER=""
BUILD_PROJECT=""
BUILD_HARDWARE=""
BUILD_CONFIGTYPE=""
k=0
for i in `echo $prjdef |sed 's/\// /g'`; do
    arr[$k]=$i
    let k=$k+1
done
# n is total number of directories found
n=${#arr[*]}
BUILD_CUSTOMER=${arr[$n-4]}
BUILD_PROJECT=${arr[$n-3]}
BUILD_HARDWARE=${arr[$n-2]}
BUILD_CONFIGTYPE=${arr[$n-1]}

echo "-- Cleaning $RELEASEDIR/root_fs"
rm -rf $RELEASEDIR/root_fs

echo "-- Release directory: $RELEASEDIR"
mkdir -p $RELEASEDIR

bldcfg="buildconfig.default"
versioninfo="versioninfo.txt"
buildinfo="$STAGINGDIR/buildconfig.info"

# Store some information for use in the build information of the image.
rm -f "${buildinfo}"
echo "BUILD_VER=`echo $curdir | sed -e 's/.*\/build\///'`" >> "${buildinfo}"
echo "BUILD_DEF=${prjdef}" >> "${buildinfo}"
echo "BUILD_START=${starttime}" >> "${buildinfo}"
echo "BUILD_HOST=`uname -n`" >> "${buildinfo}"
echo "BUILD_USER=$USER" >> "${buildinfo}"
echo "BUILD_CUSTOMER=$BUILD_CUSTOMER" >> "${buildinfo}"
echo "BUILD_PROJECT=$BUILD_PROJECT" >> "${buildinfo}"
echo "BUILD_HARDWARE=$BUILD_HARDWARE" >> "${buildinfo}"
echo "BUILD_CONFIGTYPE=$BUILD_CONFIGTYPE" >> "${buildinfo}"
echo "BUILD_SPIN_NUMBER=$spinnumber" >> "${buildinfo}"

opt_opensource_only=""
if [ "${opensource_only}" == "1" ]; then
  opt_opensource_only="-opensource_only"
fi

./build_rootpathcfg_package $wsdir
}

configureBuildTree() {
if [ -f $prjdir/test.config ]
then
  echo "test.config exists (copy $prjdir/test.config to $outputdir/test)"
  install -d $outputdir/test
  install -m 664 $prjdir/test.config $outputdir/test/
fi
echo "-- Configure build tree"
ccache_enable=""
config_source_cmd="./configsource.tcl -defconfig $defconfig -s $servercfgfile -p $prjdep -h $hardcofile -o $outputdir -b $bldcfg -osrc $osrcfile $ccache_enable $opt_buildpackage $opt_opensource_only"
echo ${config_source_cmd}
${config_source_cmd}
cat $versioninfo | sort > $STAGINGDIR/buildconfig.dep
}

populateStagingDir() {
### Copy the project dir to the staging dir
install -d ${STAGINGDIR}/config
cp -aR $prjdir/* ${STAGINGDIR}/config
chmod -R +w ${STAGINGDIR}/config

### Copy the utils dir to the staging dir
install -d ${STAGINGDIR}/utils
cp -aR $curdir/utils/* ${STAGINGDIR}/utils
chmod -R +wx ${STAGINGDIR}/utils
}

installBuildConfiguration() {
### Install build configuration
echo "-- Install build configuration"
install -m 664 -p $defconfig .config
# components.config is needed to be able to run Makefile (since its is included from Makefile.global)
install -m 664 -p .config ${STAGINGDIR}/components.config
if grep -q "^# Automatically generated make config: don't edit" .config; then
    make install-component-config
else
    # this is a minimal config with only basic project details, automatically fill it with defaults
    make oldconfig </dev/null
fi


    make install-autoconf-config

    ./scripts/install_scripts/install_to_staging.sh ${STAGINGDIR}
}


preparepackagebuild() {
    initializeEnvironment
    configureBuildTree
    populateStagingDir
    installBuildConfiguration
}

help_clean() {
cat >&2 << EOF
Synopsis: $0 <project directory> cleanoffline

Delete all build artefacts and other untracked files when not connected to version control system. This allows to make a build equivalent to a build from scratch.
EOF
}

command_clean() {
    initializeEnvironment
    echo "-- delete all non-version controlled files"
    echo "-- make list of all files in workspace and compare to list of files under version-control"
    # determine builddir
    builddir=`echo ${outputdir:0:${#outputdir}-7}`
    echo "-- builddir: $builddir"
    rootdir=`head -n1 rootpath.cfg | cut -d= -f2`
    cd $rootdir
    find -L . -type f ! -path "./config/*" ! -path "./output/*" ! -path "./build/*" > p4synced.txt
    sort p4synced.txt > p4synced_sorted.txt
    diff p4synced_sorted.txt $builddir/p4files_sorted.txt | grep "<" > diff.txt
    sed -i -e 's/< //g' diff.txt
    sed -i '/diff.txt/d' diff.txt

    echo "Beginning to remove files."

    for file in $(cat diff.txt); do
       echo "Now removing the file $file..."
       rm -f "$file"
    done

    echo "Finished removing files."
}

help_buildpackage() {
cat >&2 << EOF
Synopsis: $0 <project directory> buildpackage

  Perform all steps necessary to build a package:
  1. Compute dependencies
  2. Build each component
  3. Create root file system
  4. Generate an image from the root file system
EOF
}

command_buildpackage() {
  opt_buildpackage="-w"
  preparepackagebuild

### Build all components
  echo "-- Build all components"
  make ${JOBS}

### Create image
  if [ "${A}" != "1" ]; then
  echo "-- Create image"
  make image
  else
  echo "-- CLANG check is enabled, do not create an image"
  fi

### Build ready
  endmessage="Build"
}

if [ "help" = "$1" ]; then
        usage
    exit 1
else
    parse_arguments $@
    if [ -z $commands ]; then
       command_buildpackage
    else
       for i in $commands; do
          command_$i
       done
fi
    if [ ! -z "$endmessage" ]; then
endtime=`date +%Y-%m-%d_%H:%M:%S`
        echo "-- $endmessage completed at ${endtime}"
        if [ ! -z "${buildinfo}" ] && [ -f "${buildinfo}" ]; then
echo "BUILD_END=${endtime}" >> "${buildinfo}"
        fi
    fi
fi

help_buildimage() {
cat >&2 << EOF
Synopsis: $0 <project directory> buildimage

  Perform all steps necessary to build an image:
  1. Compute dependencies
  2. Fetch components from version control system (overwrites files !)
  3. Build each component
  4. Create root file system
  5. Generate an image from the root file system
EOF
}
