package provide buildsystem 1.0

namespace eval buildsystem {
    variable globaloptions
        variable globalbuildcfg
        variable globalcomponentlst
        variable globalrootpathlst

        set packagetargetdir "./output/package"
        set package_buildconfig_default "buildconfig.default.m4"

        set globaloptions("buildconfigfile") "$::env(HOME)/.server.cfg"

        namespace export parse_cs_commandline read_cs_deps generatePackage
}

proc buildsystem::find_alias {name} {
    foreach candidate [buildsystem::getComponents] {
        if { ! [string equal $name $candidate] && [string first $name $candidate] == 0} {
            log::log debug "Translated $name to $candidate"
            return $candidate
        }
    }
    log::log debug "No translation for $name"
    return 0
}

proc buildsystem::translate_aliases_in_deps {} {
    foreach name [buildsystem::getComponents] {
        set version [buildsystem::getComponent_version $name]
        if [buildsystem::is_undefined_version $version] {
            set fullname [buildsystem::find_alias $name]
            if { $fullname != 0 } {
                set translations($name) $fullname
                buildsystem::delComponents $name
            }
        }
    }
    foreach comp [buildsystem::getComponents] {
        set dependencies ""
        foreach {name assg version} [buildsystem::getComponent_raw_dependencies $comp] {
            if [info exists translations($name)] {
                lappend dependencies $translations($name) $assg $version
            } else {
                lappend dependencies $name $assg $version
            }
        }
        log::log debug "Deps for $comp $dependencies"
        buildsystem::setComponent_dependencies $comp $dependencies
    }
}

proc buildsystem::detect_undefined_references {} {
    # undefined references in dependencies are marked with '?' or '*'
    # "*" means a mandatory component
    # "?" means an optional component
    # As "?" is lexicographically greater than '*', it could override a '*'.
    # Thus all dependencies have to be checked again.
    set fail_build 0
    foreach comp [buildsystem::getComponents] {
        foreach {depname assg version} [buildsystem::getComponent_raw_dependencies $comp] {
            if { [string equal $version "*"] && [is_undefined_version [buildsystem::getComponent_version $depname]] } {
                puts [buildsystem::getComponent_version $depname]
                puts "\033\[31mError: unresolved component $depname=$version from $comp\033\[0m"
                set fail_build 1
            }
        }
    }
    if {$fail_build} {
        exit 1
    }
}

proc buildsystem::setOpensourceAsSource {} { 
    foreach name [buildsystem::getComponents] {
        if {[isComponent_completely_opensource $name]} {
            puts "setting release of opensource component $name as source"
            buildsystem::setComponent_rel_as_src $name
        }
        continue
    }
}

proc buildsystem::detect_gplv3_component {} { 
    foreach name [buildsystem::getComponents] {
        if {[isComponent_opensource $name]} {
            set lic [buildsystem::getComponent_license $name]
            if { [string match -nocase *gplv3* $lic] } {
                puts "\033\[31mError: component $name has GPLv3 licence .. Aborting the build"
                exit 1
            }
        }
        continue
    }
}

proc buildsystem::detect_excluded_component {} { 
    foreach name [buildsystem::getComponents] {
        if {[is_component_excluded $name]} {
            puts "\033\[31mError: component $name is on the blacklist .. Aborting the build"
            exit 1
        }
        continue
    }
        
}

proc buildsystem::setchangelistnumber {} { 
    variable globaloptions

    set changelistinfo $globaloptions("changelist_info")

    if { [file exists $changelistinfo] } {
    
        foreach name [buildsystem::getComponents] {

            if { [catch { set cl [ exec cat $changelistinfo | grep $name= | sed s/.*=// ] } ] }  {
                set cl "Unknown"
            }
            buildsystem::setComponent_changelist $name $cl
            continue
        }
    }
        
}

proc buildsystem::check_deps { } {
    variable opensource_types
    variable bs_is_opensource
    variable globaloptions

    if {$bs_is_opensource} {
        puts "Deleting none opensource components"
        delNoneOpensourceComponents
    }
    
    buildsystem::translate_aliases_in_deps
    buildsystem::detect_undefined_references
    #buildsystem::detect_gplv3_component
    buildsystem::detect_excluded_component
    buildsystem::setchangelistnumber
    if { $globaloptions("os_as_source") == 1 } {
        puts "### -> Setting opensource components as Source in package"
        buildsystem::setOpensourceAsSource
    } else {
        puts "### -> Leaving opensource components as they are"
    }

    foreach name [buildsystem::getComponents] {
        set version [buildsystem::getComponent_version $name]
        if { $version  == "?"} {
                puts "\033\[33mWarning: unresolved component $name=$version\033\[0m"
                buildsystem::delComponents $name
                continue
        }
        if {!$bs_is_opensource} {
            foreach dep_name [getComponent_dependencies $name] {
                if {![componentExists $dep_name]} {
                    puts "\033\[31mError: Component $name is depending on a not existing component $dep_name\033\[0m"
                    set fail_build 1
                }
            }
        }
    }
}

# parse configsource commandline options
# options are returned in commandlineoptions as an associative list
proc buildsystem::parseConfigSourceArgs {argv} {
    variable globaloptions
    set options {
        {defconfig.arg  "set to defconfig"}
        {osrc.arg           "set to opensource configuration file"}
        {compexcl.arg   "set the component exclude file"}
        {p.arg          "set to projectdependenciesfile"}
        {h.arg          "set to hardcofile"}
        {s.arg          "set to configurationfile"}
        {o.arg          "set to outputdirectory"}
        {b.arg          "set to buildconfig.default file"}
        {v.arg          "set to version_info file"}
        {lic.arg        "set to licence_info file"}
        {cl.arg         "set to changelist_info file"}
        {l.arg          "use specific release directory "}
        {g.arg          "use specific staging directory "}
        {f              "enable parallel building"}
        {r              "get the rootpath from the version control system"}
        {t              "enable tracing"}
        {ci             "continuous integration"}
        {c              "use ccache"}
        {nosync              "nosync"}
        {w		        "build package"}
        {x              "differentComponentReleaseDir"}
        {y              "use legacy releasedir "}
        {z              "create package"}
        {opensource_only     "Build only the opensource packages"}
    }

    set globaloptions("gendefaultbuildconfig") 1
    set globaloptions("ccache") 0
    set globaloptions("differentComponentReleaseDir") 0
    set globaloptions("createPackage") 0
    set globaloptions("buildPackage") 0
    set globaloptions("nosync") 0
    set globaloptions("version_info") "versioninfo.txt"
    set globaloptions("licence_info") "licenceinfo.txt"
    set globaloptions("changelist_info") "changelistinfo.txt"
    set globaloptions("parallel_make") 0
    set globaloptions("continuous_integration") 0
    set globaloptions("opensource_include") 0
    set globaloptions("os_as_source") 0
    set globaloptions("component_exclude") 0
    set globaloptions("opensource_only") 0

    if {[llength $argv]} {
        while {[llength $argv]} {

            set err [::cmdline::getopt argv {defconfig.arg osrc.arg compexcl.arg s.arg p.arg h.arg o.arg b.arg v.arg lic.arg cl.arg l.arg g.arg f r t ci c d nosync w x y z opensource_only} optVar optValue]
                if {$err == 1} {
                    switch $optVar {
                        "compexcl" {
                            set globaloptions("component_exclude") [file join $optValue]
                        }
                        "defconfig" {
                            set globaloptions("defconfig") [file join $optValue]
                        }
                        "osrc" {
                            set globaloptions("opensource_include") [file join $optValue]
                        }

                        "p" {
                            set globaloptions("projectdependenciesfile") [file join $optValue]
                        }
                        "h" {
                            set globaloptions("hardcofile") $optValue
                        }
                        "s" {
                            set globaloptions("buildconfigfile") $optValue
                        }
                        "o" {
                            set globaloptions("outputdirectory") $optValue
                        }
                        "b" {
                            set globaloptions("buildconfigdefault") $optValue
                        }
                        "v" {
                            set globaloptions("version_info") $optValue
                        }
                        "lic" {
                            set globaloptions("licence_info") $optValue
                        }
                        "cl" {
                            set globaloptions("changelist_info") $optValue
                        }
                        "r" {
                            set globaloptions("rootpathfromvcs") "1"
                        }
                        "t" {
                            log::lvSuppress debug 0
                        }
                        "ci" {
                            set globaloptions("continuous_integration") 1
                        }
                        "f" {
                            set globaloptions("parallel_make") 1
                        }
                        "l" {
                            set globaloptions("releasedir") $optValue
                        }
                        "g" {
                            set globaloptions("stagingdir") $optValue
                        }
                        "c" {
                            set globaloptions("ccache") 1
                        }
                        "nosync" {
                            set globaloptions("nosync") 1
                        }
                        "w" {
                            set globaloptions("buildPackage") 1
                        }
                        "x" {
                            set globaloptions("differentComponentReleaseDir") 1
                        }
                        "y" {
                            set globaloptions("legacyreleasedir") 1
                        }
                        "z" {
                            set globaloptions("createPackage") 1
                        }
                        "opensource_only" {
                            set globaloptions("opensource_only") 1
                            setOpenSourceBuild 1
                        }
                    }
                } elseif {$err < 0 } {
                    set usage [::cmdline::usage $options]
                        puts $usage
                        exit 1
                } else {
                    exit 1
                }

        }
    } else {
        set usage [::cmdline::usage $options]
            puts $usage
            exit 1
    }
}

#read all dependencies recursively
proc buildsystem::read_cs_deps {componentnames} {
    return [buildsystem::read_deps $componentnames 0]
}

proc buildsystem::getGlobalOptions {} {
    variable globaloptions
    return [array get globaloptions]
}

# Create buildconfig.default and Component.all
proc buildsystem::buildConfigAndComponentAll {scriptdir} {
    variable globaloptions
    variable globalrootpathlst
    variable globalcomponentlst
    variable package_buildconfig_default

    set fail_on_exit 0
    if { [ info exists globaloptions("stagingdir") ] } {
        set stagingdir [ string trim "$globaloptions("stagingdir")" ]
    } else {
        set stagingdir [string trim "$globaloptions("outputdirectory")/staging"]
    }

    ########################################
    # generate buildconfig.default, buildconfig.default.m4 and Component.all
    set buildconfig_package [open "$package_buildconfig_default" "w"]
    set buildconfig [open $globaloptions("buildconfigdefault") "w"]
    set version_info [open $globaloptions("version_info") "w"]
    set licence_info [open $globaloptions("licence_info") "w"]
    set componentall [open "$scriptdir/Component.all" "w"]
    puts $componentall "source \"$stagingdir\/buildconfig.in\""
    puts $componentall "source \"Global.in\""

    set lastpath []

    foreach name [lsort [buildsystem::getComponents]] {
        set deppath [buildsystem::getComponent_deppath $name]
        puts $buildconfig "[string toupper $name]=$deppath"
        puts $version_info "$name=[buildsystem::getComponent_version $name] Changelist=[buildsystem::getComponent_changelist $name]"
        puts $licence_info "$name=[buildsystem::getComponent_license $name]"
        puts $buildconfig_package "[string toupper $name]=__CWD__$deppath"

        set err_file_list "Component.mk Component.in Component.lic"
        foreach chk_file $err_file_list {
            if {![defconfig_name_is_set $name]} {
                continue
            }

            if {![file exists $deppath/$chk_file]} {
                puts "\033\[31mError: $name: no $chk_file at $deppath!\033\[0m"
                #For now exit for everything except Component.dep
                set fail_on_exit 1
            }
        }

        set wrn_file_list "Component.dep"
        foreach chk_file $wrn_file_list {
            if {![file exists $deppath/$chk_file]} {
                puts "\033\[33mWarning: $name: no $chk_file at $deppath!\033\[0m"
            }
        }

        array set servercomponent $globalcomponentlst($name)
        set paths [split $name "_"]

        set commonCount 0
        for {set i 0} {$i<[llength $paths] && $i<[llength $lastpath]} {incr i} {
            if {[lindex $paths $i] == [lindex $lastpath $i]} {
                incr commonCount
            } else {
                break
            }
        }

# Close old menus
        for {set i $commonCount} {$i<[llength $lastpath]} {incr i} {
            puts $componentall "endmenu"
        }

# Open new menus
        for {set i $commonCount} {$i<[llength $paths]} {incr i} {
            puts $componentall "menu \"[lindex $paths $i]\""
        }

# write entry in Component.all
        puts $componentall "source \"$deppath/Component.in\""
        set uname [string toupper $name]
        foreach option "CFLAGS LDFLAGS" {
            puts $componentall "config ${uname}_${option}"
            puts $componentall "string ${name}_${option}"
            puts $componentall "default \"\""
        }
        set lastpath $paths
    }

# Close menus for the last entry
    for {set i 0} {$i<[llength $lastpath]} {incr i} {
        puts $componentall "endmenu"
    }

    close $componentall
    close $buildconfig
    close $version_info
    close $licence_info
    close $buildconfig_package

    if {$fail_on_exit} {
        exit 1
    }
}

# determine the relative position from the build dir compared to the other directories
proc buildsystem::getRelativeBuildLocation {} {
    set buildsystem_version [buildsystem::getComponent_version "build"]
    set buildsystem_srvpath [buildsystem::get_server_path "build"]
    set rel_package_loc     [file join "$buildsystem_srvpath/$buildsystem_version"]

    set relativePosCount [expr [llength [split $rel_package_loc /]] - 1]
    set relativeBuildLocation ""
    for { set i 0 } { $i < $relativePosCount} {incr i} {
        if [string length $relativeBuildLocation ] {
            set relativeBuildLocation "$relativeBuildLocation/"
        }
        set relativeBuildLocation "$relativeBuildLocation.."
    }
    return $relativeBuildLocation
}

# generate component Makefile
proc buildsystem::generateCompMakeFile {generateForPackageCreation} {
    variable globaloptions
    variable package_default
    variable packagetargetdir
    set tmp_path "/tmp/build_"
    append tmp_path [expr int(rand() * 10000 + 1)] 

    if {$generateForPackageCreation == 1} {
        puts "calling generateCompMakeFile with generateForPackageCreation set to 1"
        set relPath [buildsystem::getComponent_relpath "build"]
        set component_makefile [open "$packagetargetdir/$relPath/Makefile.components" "w"]
        set relativeBuildLocation [buildsystem::getRelativeBuildLocation]
    } else {
        puts "calling generateCompMakeFile with generateForPackageCreation set to $generateForPackageCreation"
        set component_makefile [open "Makefile.components" "w"]
    }

    if { [ info exists globaloptions("stagingdir") ] } {
            set stagingdir [ string trim "$globaloptions("stagingdir")" ]
        } else {
            set stagingdir [string trim "$globaloptions("outputdirectory")/staging"]
        }
    if {[info exists globaloptions("releasedir")]} {
                set releasedir [string trim "$globaloptions("releasedir")"]
            } else {
                set releasedir [string trim "$globaloptions("outputdirectory")/release"]
            }
    set component_makedep [open "Makefile.deps" "w"]

    puts $component_makefile "# Makefile.components"
    puts $component_makefile "# tmp path = $tmp_path"
    puts $component_makefile ""
    puts $component_makefile "include Makefile.global"
    puts $component_makefile "include Makefile.recipes"
    puts $component_makefile ""
    puts $component_makefile ""
    puts $component_makefile "all: components"
    puts $component_makefile ""

    set has_stb_hardco_toolchain [buildsystem::componentExists "stb_hardco_toolchain"]
    if { $has_stb_hardco_toolchain } {
        # Everything must depend on stb_hardco_toolchain, except itself and its dependencies
        set must_not_depend_on_toolchain(stb_hardco_toolchain) 1
        foreach component [buildsystem::getComponent_dependencies "stb_hardco_toolchain"] {
            set must_not_depend_on_toolchain($component) 1
        }
    }

    # ################################################
    foreach name [buildsystem::getComponents] {
        set relative_package_location   [buildsystem::getComponent_relpath $name]
        set comp_location               [buildsystem::getComponent_deppath $name]
        set release_as_binary           [buildsystem::getComponent_rel_as_bin $name]
        set line_sah_debug              [buildsystem::getComponent_debug_options $name]
        set enable_debug                [buildsystem::getComponent_enable_debug $name]
        set temp_location               "$tmp_path/$name"
        set temp_package_location       "$tmp_path/package/$name"
        set builddir                     [pwd]

        if {$generateForPackageCreation == 1} {
            set component_location "$relativeBuildLocation$relative_package_location"
        } else {
            set component_location [buildsystem::getComponent_deppath $name]
        }

        # binary_package_dir is the private root_fs directory for a component
        set binary_package_dir {COMP_PACKAGE_DIR=$(STAGINGDIR)}
        set binary_package_dir "$binary_package_dir/binary_packages/$name"
        set binary_package_dir_location {$(STAGINGDIR)}
        set binary_package_dir_location "$binary_package_dir_location/binary_packages/$name"

        set source_package_dir {COMP_SOURCE_DIR=$(STAGINGDIR)}
        set source_package_dir "$source_package_dir/../package$relative_package_location"
        set package_component_dir "$stagingdir/../package$relative_package_location"

        set builddeps [buildsystem::getComponent_dependencies $name]
        # Add stb_hardco_toolchain dependency (only for build rule, not needed for _release and others)
        if { $has_stb_hardco_toolchain && ![info exists must_not_depend_on_toolchain($name)] } {
            append builddeps " stb_hardco_toolchain"
        }

        ### Generate build commands
        if {$globaloptions("parallel_make") == 1} {
            puts $component_makefile "$name: $builddeps"
            puts $component_makefile  "\t\$(MAKE) ${name}_component -f Makefile.components"
        } else {
            puts $component_makefile "$name: $builddeps ${name}_component"
        }
        puts $component_makefile  ""

        puts $component_makefile "${name}_deps: [buildsystem::getComponent_dependencies_with_suffix $name "_deps"]"

        set uname [string toupper $name]
        puts $component_makefile "ifneq (\$(CONFIG_${uname}_CFLAGS),\"\")"
        puts $component_makefile "${name}_component: CFLAGS=\$(subst \$(quote),\$(empty),\$(CONFIG_${uname}_CFLAGS))"
        puts $component_makefile "endif"

        puts $component_makefile "ifneq (\$(CONFIG_${uname}_LDFLAGS),\"\")"
        puts $component_makefile "${name}_component: LDFLAGS=\$(subst \$(quote),\$(empty),\$(CONFIG_${uname}_LDFLAGS))"
        puts $component_makefile "endif"

        puts $component_makefile "${name}_component:"
        set line {ifeq ($(CONFIG_}
        set line "$line[string toupper $name]),y)"
        puts $component_makefile $line
        if { (($generateForPackageCreation == 1) || ($generateForPackageCreation == 2)) && $release_as_binary != 0 } {
            ##-# untar staging.tar
            set line "\tmkdir -p $temp_package_location/staging"
            puts $component_makefile $line
            set line "\ttar -xvf $comp_location/staging.tar -C $stagingdir"
            puts $component_makefile $line
        } elseif { [file exists $comp_location/staging.tar] } {
            # binary component in workspace -> release as binary again
            # untar to the staging dir for dependency purposes
            # leave the tar file in it's place
            set line "\tmkdir -p $temp_package_location/staging"
            puts $component_makefile $line
            set line "\ttar -xvf $comp_location/staging.tar -C $stagingdir"
            puts $component_makefile $line
            set line "\techo \"$name was binary -> releasing as binary again\""
            puts $component_makefile $line 
        } else {
            if { ($globaloptions("createPackage") == 1) && ($generateForPackageCreation == 0) && ($release_as_binary != 0) } {
                ##-# do a pre-install find of the staging dir (with file sizes)
                set line "\tmkdir -p $temp_location"
                puts $component_makefile $line
                set line "\tcd $stagingdir ; find -printf \"%p %s\\n\" > $temp_location/staging_pre.txt"
                puts $component_makefile $line
            }
            puts $component_makefile "ifeq \(\$\(A\),1\)"
            set line "\tinstall -d \"\$\(OUTPUTDIR\)/analysis/$name\""
            puts $component_makefile $line
            set CC_PREFIX "CCC_ANALYZER_CPLUSPLUS=1 scan-build -analyze-headers --use-cc \"\$\(CC_NOCCACHE\)\" --use-c++ \"\$\(CXX_NOCCACHE\)\" -o \"\$\(OUTPUTDIR\)/analysis/$name\" --keep-going"
            set line "\t$CC_PREFIX \$(MAKE) -C $component_location -f Component.mk"
            puts $component_makefile $line
            set line "\t-rmdir \"\$\(OUTPUTDIR\)/analysis/$name\" && echo \"CLANG discovered no errors for component $name, removing the result directory\""
            puts $component_makefile $line
            puts $component_makefile "else"
            puts $component_makefile "\t\$(MAKE) -C $component_location -f Component.mk $line_sah_debug"
            puts $component_makefile "endif"

            set line "\t\$(MAKE) -C $component_location -f Component.mk install $line_sah_debug"
            puts $component_makefile $line
            if { ($globaloptions("createPackage") == 1) && ($generateForPackageCreation == 0) && ($release_as_binary == 1) } {
                ##-# do a post-install find of the staging dir (with file sizes) and tar the differences
                set line "\ttouch $stagingdir/dummy"
                puts $component_makefile $line
                set line "\tcd $stagingdir ; find -printf \"%p %s\\n\" > $temp_location/staging_post.txt"
                puts $component_makefile $line
                set line "\tcd $stagingdir ; diff $temp_location/staging_pre.txt $temp_location/staging_post.txt | grep \"^>\" | sed 's/^..//' | awk '{\$\$(NF--);print}' | tar -cvf $package_component_dir/staging.tar --no-recursion -T -"
                puts $component_makefile $line
                set line "\trm -f $stagingdir/dummy"
                puts $component_makefile $line
            } elseif { ($globaloptions("createPackage") == 1) && ($generateForPackageCreation == 0) && ($release_as_binary == 2) } {
                ##-# do a post-install find of the staging dir (with file sizes) and tar the differences except the header files
                set line "\ttouch $stagingdir/dummy"
                puts $component_makefile $line
                set line "\tcd $stagingdir ; find -printf \"%p %s\\n\" > $temp_location/staging_post.txt"
                puts $component_makefile $line
                set line "\tcd $stagingdir ; diff $temp_location/staging_pre.txt $temp_location/staging_post.txt | grep \"^>\" | sed 's/^..//' | awk '{\$\$(NF--);print}' | tar -cvf $package_component_dir/staging.tar --no-recursion --exclude='*.h' -T -"
                puts $component_makefile $line
                set line "\trm -f $stagingdir/dummy"
                puts $component_makefile $line
            } 
        }
        puts $component_makefile "endif"
        puts $component_makefile ""

### generate release commands
        set strip_cmd {STRIP=$(STRIP)}
        puts $component_makefile "${name}_release: [buildsystem::getComponent_dependencies_with_suffix $name "_release"] ${name}_component_release"
        puts $component_makefile  ""

        puts $component_makefile "${name}_component_release:"
        set line {ifeq ($(CONFIG_}
        set line "$line[string toupper $name]),y)"
        puts $component_makefile $line

        set releasedirlocation ""
        if ($globaloptions("differentComponentReleaseDir")) {
            set releasedirlocation "RELEASEDIR=\$(RELEASEDIR)/$name ROOT_FS_DIR=\$(ROOT_FS_DIR)/$name/root_fs"
        }
        if { (($generateForPackageCreation == 1) || ($generateForPackageCreation == 2)) && $release_as_binary == 1 } {
            set line "\t\$(MAKE) -C $component_location -f Component.mk release $strip_cmd $releasedirlocation $line_sah_debug"
            puts $component_makefile $line
        } else {
            if { ($globaloptions("createPackage") == 1) && ($generateForPackageCreation == 0) && ($release_as_binary != 0) } {
                ##-# do a pre-install find of the release dir (with file sizes)
                set line "\tmkdir -p $temp_location"
                puts $component_makefile $line
                set line "\tcd $releasedir ; find -printf \"%p %s\\n\" > $temp_location/release_pre.txt"
                puts $component_makefile $line
            }

            set line "\t\$(MAKE) -C $component_location -f Component.mk release $strip_cmd $releasedirlocation $line_sah_debug"
            puts $component_makefile $line

            if { ($globaloptions("createPackage") == 1) && ($generateForPackageCreation == 0) && ($release_as_binary != 0) } {
                ##-# do a post-install find of the release dir (with file sizes) and tar the differences
                set line "\ttouch $releasedir/dummy"
                puts $component_makefile $line
                set line "\tcd $releasedir ; find -printf \"%p %s\\n\" > $temp_location/release_post.txt"
                puts $component_makefile $line
                #set line "\tcd $releasedir ; diff $temp_location/release_pre.txt $temp_location/release_post.txt | grep \">\" | sed 's/^..//' | awk '{\$\$(NF--);print}' | awk '{print\(\"\\\"\" \$\$0 \"\\\"\"\)}' | xargs tar -cvf $package_component_dir/release.tar"
                #puts $component_makefile $line
                set line "\trm -f $releasedir/dummy"
                puts $component_makefile $line
            }
        }
        set line "\trm -rf "

        puts $component_makefile "endif"
        puts $component_makefile ""

#generate make clean commands
        puts $component_makefile "${name}_clean: [buildsystem::getComponent_dependencies_with_suffix $name "_clean"] ${name}_component_clean"
        puts $component_makefile  ""

        puts $component_makefile "${name}_component_clean:"
        if { (($generateForPackageCreation == 1) || ($generateForPackageCreation == 2)) && $release_as_binary == 1 } {
            puts $component_makefile "\t@echo Binary component version, nothing to be done for $name\n"
        } else {
            set line {ifeq ($(CONFIG_}
            set line "$line[string toupper $name]),y)"
            puts $component_makefile $line
            set line "\t\$(MAKE) -C $component_location -f Component.mk clean $line_sah_debug"
            puts $component_makefile $line
            puts $component_makefile "endif"
            puts $component_makefile ""
        }
        if { [string match *kernel_broadcom* $name] } {
#generate make prepare_oss commands
            puts $component_makefile "${name}_prepare_oss: [buildsystem::getComponent_dependencies_with_suffix $name "_clean"] ${name}_component_prepare_oss"
            puts $component_makefile  ""

            puts $component_makefile "${name}_component_prepare_oss:"
            if { (($generateForPackageCreation == 1) || ($generateForPackageCreation == 2)) && $release_as_binary == 1 } {
                puts $component_makefile "\t@echo Binary component version, nothing to be done for $name\n"
            } else {
                set line {ifeq ($(CONFIG_}
                set line "$line[string toupper $name]),y)"
                puts $component_makefile $line
                set line "\t\$(MAKE) -C $component_location -f Component.mk prepare_oss $line_sah_debug"
                puts $component_makefile $line
                puts $component_makefile "endif"
                puts $component_makefile ""
            }
        }

#entry for counting lines of code per component
        puts $component_makefile "${name}_countlines: [buildsystem::getComponent_dependencies_with_suffix $name "_countlines"] ${name}_component_countlines"
        puts $component_makefile  ""

        puts $component_makefile "${name}_component_countlines:"
        puts $component_makefile "\tperl ./scripts/cloc/cloc-1.08.pl --quiet --progress-rate=0 $component_location"
        puts $component_makefile ""

#entry for gathering license files
        puts $component_makefile "${name}_getlicenses: [buildsystem::getComponent_dependencies_with_suffix $name "_getlicenses"] ${name}_component_getlicenses"
        puts $component_makefile  ""

        puts $component_makefile "${name}_component_getlicenses:"
        puts $component_makefile "\tperl ./scripts/getlicenses.pl ${name} $component_location"
        puts $component_makefile ""

        puts $component_makefile "${name}_getsourcecode:"
        puts $component_makefile  "\tperl ./scripts/getsourcecode.pl TARGET=COLLECT COMPONENT=$name PATH=$component_location CODE_SOURCE=\$\{CODE_SOURCE\} LICENSE=\$\{LICENSE\} STAGINGDIR=\$\(STAGINGDIR\) ROOT_LIST=[buildsystem::getRootPaths_cvs]"
        puts $component_makefile ""

#generate make doc commands
# do not add dependencies... otherwise components that have no doc target will cause documentation not to be build
        puts $component_makefile "${name}_doc: ${name}_component_doc"
        puts $component_makefile  ""

        puts $component_makefile "${name}_component_doc:"
        set line {ifeq ($(CONFIG_}
        set line "$line[string toupper $name]),y)"
        puts $component_makefile $line

        set line "\t\$(call doc_recipe,$component_location,$name)"
        puts $component_makefile $line
        puts $component_makefile "endif"
        puts $component_makefile ""


#generate make test commands
# do not add dependencies... otherwise components that have no tests target will cause tests not to be build
        puts $component_makefile "${name}_test: ${name}_component_test"
        puts $component_makefile  ""


        puts $component_makefile "${name}_component_test:"
        set line {ifeq ($(CONFIG_}
        set line "$line[string toupper $name]),y)"
        puts $component_makefile $line

        set localtestdir "\$(OUTPUTDIR)/test/$name"
        set testtype "remote"
        if {[file exists "output/test/test.config"] == 1} {
#  Slurp up the data file
            set fp [open "output/test/test.config" r]
            set file_data [read $fp]
            close $fp
#  Process data file
            set data [split $file_data "\n"]
            foreach line $data {
                if {[string length $line] > 1 } {
                    set x [checkconfig "TESTTYPE" $line]
                    if {$x == 1} { set testtype [setconfig $line] }
                }
            }
        }


#gather configuration information needed for xml
        set spin [exec cat output/staging/buildconfig.info | grep BUILD_SPIN_NUMBER= | sed -e s/BUILD_SPIN_NUMBER=//]
        set configlocation [exec cat output/staging/buildconfig.info |  grep BUILD_DEF | sed -e s/BUILD_DEF=//]
        set configindex [expr [ string first "/config/" $configlocation ] + 8 ]
        set strippedlocation [ string range $configlocation $configindex [ string length $configlocation ] ]
        set dirparts [split $strippedlocation "/"]
        set version ""
        foreach dirpart $dirparts {
            if [string length $dirpart] {
                if ![string compare $dirpart [string toupper $dirpart] ] {
                    set version "$version/$dirpart"
                }
            }
        }
        set testconfig [string range $strippedlocation [string length $version] [string length $strippedlocation]]
        set testconfigversion [string range $version 1 [string length $version]]

        set line "\t\$(call test_recipe,$component_location,$name,[buildsystem::getComponent_version $name],$localtestdir,$testconfig,$testconfigversion,$spin,$testtype)"
        puts $component_makefile $line

        puts $component_makefile "endif"
        puts $component_makefile ""
#############################
    }

######################################################
    puts $component_makefile "components: [buildsystem::makefile_getProjComponents ""]"
    puts $component_makefile ""

    puts $component_makefile "releasecomp: [buildsystem::makefile_getProjComponents "_release"]"
    puts $component_makefile ""

    puts $component_makefile "release: releasecomp"
    puts $component_makefile ""
    puts $component_makefile ""

    puts $component_makefile "cleancomp: [buildsystem::makefile_getProjComponents "_clean"]"
    puts $component_makefile ""
    puts $component_makefile "clean: cleancomp"
    puts $component_makefile ""

    puts $component_makefile "doccomp: [buildsystem::makefile_getComponents "_doc"]"
    puts $component_makefile ""
    puts $component_makefile "doc: doccomp"
    puts $component_makefile ""

    puts $component_makefile "modular_packages: [buildsystem::getModularComponents]"
    puts ""

    puts $component_makefile "testcomp: [buildsystem::makefile_getComponents "_test"]"
    puts $component_makefile ""
#gather configuration information needed for xml
    set spin [exec cat output/staging/buildconfig.info | grep BUILD_SPIN_NUMBER= | sed -e s/BUILD_SPIN_NUMBER=//]
    set configlocation [exec cat output/staging/buildconfig.info |  grep BUILD_DEF | sed -e s/BUILD_DEF=//]
    set configindex [expr [ string first "/config/" $configlocation ] + 8 ]
    set strippedlocation [ string range $configlocation $configindex [ string length $configlocation ] ]
    set dirparts [split $strippedlocation "/"]
    set version ""
        foreach dirpart $dirparts {
        if [string length $dirpart] {
            if ![string compare $dirpart [string toupper $dirpart] ] {
                set version "$version/$dirpart"
            }
        }
    }
    set testconfig [string range $strippedlocation [string length $version] [string length $strippedlocation]]
    set testconfigversion [string range $version 1 [string length $version]]

#add the global header
    puts $component_makefile "\techo \"<?xml version=\\\"1.0\\\"?>\" > \$(OUTPUTDIR)/test/testlist.xml"
    puts $component_makefile "\techo \"<testrun config=\\\"$testconfig\\\" configversion=\\\"$testconfigversion\\\" spinnumber=\\\"$spin\\\">\" >> \$(OUTPUTDIR)/test/testlist.xml"

    foreach name [buildsystem::getComponents] {
        puts $component_makefile "\t@if \[ -f \$(OUTPUTDIR)/test/${name}/component.testlist.xml \] && \[ `cat \$(OUTPUTDIR)/test/${name}/component.testlist.xml | wc -l` -gt 5 \]; then cat \$(OUTPUTDIR)/test/${name}/component.testlist.xml | grep -v \"<?xml version=\" | grep -v \"<testrun\" | grep -v \"testrun>\" >> \$(OUTPUTDIR)/test/testlist.xml; fi"
#set line "$line ${name}_test"
    }

#add the global xml footer
    puts $component_makefile "\techo \"</testrun>\" >> \$(OUTPUTDIR)/test/testlist.xml"
    exec install -d output/test
    exec echo $spin > output/test/test.spin
    exec echo $testconfig > output/test/test.configlocation
    exec echo $testconfigversion > output/test/test.configversion

    puts $component_makefile ""
    puts $component_makefile "tests: testcomp"

    puts $component_makefile ""

    puts $component_makefile "countlines: [buildsystem::makefile_getProjComponents "_countlines"]"
    puts $component_makefile ""

    puts $component_makefile "getlicenses: [buildsystem::makefile_getProjComponents "_getlicenses"]"
    puts $component_makefile ""

################ begin get_source_code
    puts $component_makefile "getsourcecode_release: getsourcecode"
    puts $component_makefile  "\tperl ./scripts/getsourcecode.pl TARGET=RELEASE RELEASEDIR=\$\(RELEASEDIR\) STAGINGDIR=\$\(STAGINGDIR\)"
    puts $component_makefile ""

    set souce_code_deps ""
    foreach name [buildsystem::getComponents] {
        set souce_code_deps "${souce_code_deps} ${name}_getsourcecode"
    }

    puts $component_makefile "getsourcecode: getsourcecode_clean $souce_code_deps"
    puts $component_makefile ""

    puts $component_makefile "getsourcecode_clean:"
    puts $component_makefile  "\tperl ./scripts/getsourcecode.pl TARGET=CLEAN STAGINGDIR=\$\(STAGINGDIR\)"
    puts $component_makefile ""
################ end get_source_code
    puts $component_makefile ".PHONY: all build components release"
    close $component_makefile
    close $component_makedep
}

# converts the buildconfig.info to a config.in so it can be used in menuconfigs
proc buildsystem::ConvertBuildconfigInfoToConfigIn {} {
    variable globaloptions

# this list needs to be hardcoded because otherwise we don't know which ones need to be used as a boolean as well!
        set boollist "BUILD_HARDWARE BUILD_CONFIGTYPE BUILD_PROJECT BUILD_CUSTOMER"
        set skiplist "BUILD_VER BUILD_DEF BUILD_START BUILD_HOST BUILD_USER"

        if { [ info exists globaloptions("stagingdir") ] } {
            set stagingdir [ string trim "$globaloptions("stagingdir")" ]
        } else {
            set stagingdir [string trim "$globaloptions("outputdirectory")/staging"]
        }

    set filename "$stagingdir/buildconfig.info"
        set buildconfig [open "$stagingdir/buildconfig.in" "w"]
        if { [ catch { set fp [ open "$filename" "r" ] } ] } {
            puts "\033\[31merror occured while reading $filename\033\[0m"
                exit
        }
#set line "menu \"build settings\""
#puts $buildconfig $line
    set line ""
        fconfigure $fp -buffering line
        gets $fp data
        while {$data != ""} {
            set skip 0
#puts $data
                set record [ split $data "="]
# create string config
                foreach l $skiplist {
#puts "$l = [lindex $record 0 ]"
#after 1000
                    if { [ string equal $l [ lindex $record 0 ] ] } {
                        set skip 1
#puts "break";
                            break;
                    }
                }
            if {$skip} {
                gets $fp data;
                continue
            }
            set line "config [lindex $record 0 ]\n"
                set line "$line \tstring\n"
                set line "$line \tdefault \"[lindex $record 1]\"\n"
                set line "$line \thelp\n\t\t[lindex $record 0] parsed from buildconfig.info"
                puts $buildconfig $line
                foreach index $boollist {
                    if { [string equal $index [lindex $record 0]] } {
# create bool config
                        set line "config [string toupper [lindex $record 1] ]\n"
                            set line "$line \tbool\n\t"
                            set line "$line default y\n"
                            set line "$line \thelp\n\t\t[lindex $record 1] parsed from buildconfig.info"
                            puts $buildconfig $line
                    }
                }
            gets $fp data
        }
#set line "endmenu"
    puts $buildconfig $line
        close $fp
        close $buildconfig
}
