#!sh -x

##
# some sanity checks
if [[ ! -x $(which realpath) ]]; then
    echo "sonarbash: the realpath package is needed for the verify script to work"
fi

if [[ ! -x $(which cppcheck) ]]; then
    echo "sonarbash: the cppcheck package is needed for the verify script to work"
fi

##
# Where do we cache the workspaces
P4WORKSPACES="${HOME}/.workspaces"

##
# Update the workspace file
function sah-fetch-workspaces() {
    p4 clients -u $P4USER | awk '{print $2 "=" $5}' > "${P4WORKSPACES}"
}

##
# Setup P4CLIENT and P4ROOT when we are in a workspace
function sah-current-workspace() {
    local MATCH DIR WS
    if [[ ! -e "${P4WORKSPACES}" ]]; then
        sah-fetch-workspaces;
    fi

    MATCH=$(cat ${P4WORKSPACES} | while read WS; do
        DIR=${WS#*=}
        if [[ "${DIR}" == "${PWD:0:${#DIR}}" ]]; then
            echo "${WS}"
            break;
        fi
    done)
    if [ -z "${MATCH}" ]; then
        export P4CLIENT= P4ROOT=
        return 1;
    else
        export P4CLIENT="${MATCH%=*}" P4ROOT="${MATCH#*=}"
        return 0;
    fi
}

function sah-generate-sonar-env() {
    local MF="./Makefile.global" TOOLCHAIN CROSS_COMPILE
    # we need the toolchain to create the cross compile var
    TOOLCHAIN=$(grep HARDCO_TOOLCHAIN_DIR "${MF}" | cut -d= -f2)
    CROSS_COMPILE=$(grep CONFIG_CROSS_COMPILE= "./output/staging/components.config" | cut -d= -f2)
    CROSS_COMPILE=${CROSS_COMPILE/\$(HARDCO_TOOLCHAIN_DIR)/${TOOLCHAIN}}
    echo "export CROSS_COMPILE=${CROSS_COMPILE}" > "./sonar.env"
    cat "${MF}" |\
        sed -n -e '/export CROSS_COMPILE/ d' \
               -e '/export PATH/ d' \
               -e '/export TARGET_MACHINE/ d' \
               -e '/export DUMP_CMD/ d' \
               -e '/\+\+/ d' \
               -e '/^export / s/ *= *\(.*\)/="\1"/p' \
               -e "/:=/ s/ *:= *\(.*\)/='\1'/p" |\
        sed -e 's/\$(/${/g' -e 's/)/}/g' >> "./sonar.env"
}

function sah-configure-env() {
    if sah-current-workspace; then
        if test -d "${P4ROOT}/output/.."; then
            BUILDDIR=$(realpath "${P4ROOT}/output/..")
            export BUILDDIR
            if [ -e "${BUILDDIR}/Makefile.global" ]; then
                if [ "${BUILDDIR}/sonar.env" -ot "${BUILDDIR}/Makefile.global" ]; then
                    sah-generate-sonar-env
                fi
                . "${BUILDDIR}/sonar.env"
                . "${BUILDDIR}/output/staging/buildconfig.info"
                if [ -e "${BUILD_DEF}/sonar.config" ]; then
                  . "${BUILD_DEF}/sonar.config"
                elif [ -e "${BUILDDIR}/${BUILD_DEF}/sonar.config" ]; then
		  . "${BUILDDIR}/${BUILD_DEF}/sonar.config"
                fi
            fi
        else
            unset BUILDDIR
        fi
    else
        unset GLOBALMAKEFILE RELEASEDIR ROOT_FS_DIR PKG_CONFIG_LIBDIR \
            PKG_CONFIG_TOP_BUILD_DIR CCACHE_BASEDIR CROSS_COMPILE AS LD CC CXX \
            AR NM STRIP OBJCOPY OBJDUMP RANLIB INSTALL HARDCO_HAL_DIR \
            HARDCO_KERNEL_DIR HARDCO_TOOLCHAIN_DIR P4CLIENT P4ROOT
    fi
}

function sah-component-name() {
    egrep -i "${1}.*=" "${STAGINGDIR}/buildconfig.dep" | cut -d '=' -f 1 | sort | head -n 1
}

function sah-component-path() {
    if [[ "${1}" == "build" ]]; then
        echo "${BUILDDIR}"
    else
        egrep -i "${1}=" ${BUILDDIR}/buildconfig.default | cut -d '=' -f 2 | head -n 1
    fi
}

function sonar_properties_create() {
  mkdir -p output/staging/sonar
  
  project=$1
  component=$2
  
  echo "# requiered metadata" > ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  echo "sonar.projectKey=com.softathome:$project" >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  echo "sonar.projectName=/output/staging/sonar/${component^^}" >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  echo "sonar.projectVersion=DEV_LOCAL" >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  echo >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  
  echo "# The value of the property must be the key of the language." >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  echo "sonar.language=c" >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  echo >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties

  echo "# path to cppcheck tool." >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  echo "sonar.c.cppcheck.path=$(which cppcheck)" >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  echo >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
}

function sonar_properties_add_include_dirs() {
  toolchain_include_dirs=$1
  kernel_include_dirs=$2
  extra_include_dirs=$3

  includes="sonar.c.library.directories="
  separator=""
  
  if [ "$(echo $extra_include_dirs | grep "%KERNEL%")" != "" ]; then
      kernel_root=$(cat ${BUILDDIR}/buildconfig.default | grep "SAH_KERNEL=" | awk -F'=' '{print $2}')
      dirs=$(echo $kernel_include_dirs | tr "," " ")
      temp=""
      for i in $dirs; do
        temp=$temp$separator$(realpath $kernel_root$i)
        separator=","
      done
      separator=""
      extra_include_dirs=$(echo $extra_include_dirs | sed "s#%KERNEL%#$temp#g")
  fi
  
  if [ "$toolchain_include_dirs" != "" ]; then
      toolchain_dir=$(realpath $(dirname $CC)/../..)
      dirs=$(echo $toolchain_include_dirs | tr "," " ")
      for i in $dirs; do
        includes=$includes$separator$(realpath $toolchain_dir$i)
        separator=","
      done
  fi
  includes=$includes$separator${BUILDDIR}/output/staging/include
  if [ "$extra_include_dirs" != "" ]; then
      includes=$includes","$extra_include_dirs
  fi
  
  # on x86 make sure that /usr/include,/usr/include/i386-linux-gnu are added as extra include dirs
  echo "# Optional comma-separated list of additional libraries folders (such as /usr/include)" >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  echo "$includes"  >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
  echo >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
}

function sonar_properties_add_module() {
  component=$1
  path=$2
 
  if [ -e $path/sonar-project.properties ]; then
      echo "Component $component has sonar properties ($path/sonar-project.properties)"
      mkdir -p ${BUILDDIR}/output/staging/sonar/$component
      echo "sonar.sources=$path" >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
      cat $path/sonar-project.properties | grep "sonar\.exclusion" >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
      component_include_dirs=$(cat $path/sonar-project.properties | grep "sonar\.c\.library\.directories=")
      toolchain_dir=$(realpath $(dirname $CC)/../..)
      staging_dir=$(realpath ${BUILDDIR}/output/staging/include)
      if [ "$component_include_dirs" != "" ]; then
          sonar_properties_add_include_dirs $TOOLCHAIN_INCLUDE_DIRS $KERNEL_INCLUDE_DIRS $(echo $component_include_dirs | awk -F'=' '{print $2}' | \
                                                                                                                          sed "s#%COMPONENT_PATH%#$path#g" | \
                                                                                                                          sed "s#%TOOLCHAIN%#$toolchain_dir#g" | \
                                                                                                                          sed "s#%STAGING%#$staging_dir#g")
      else
          sonar_properties_add_include_dirs $TOOLCHAIN_INCLUDE_DIRS
      fi
  else
      echo "Component $component has no sonar properties file - please add one"
      return 1
  fi

  return 0
}

function sonar_properties_add_sources() {
  path=$1

  sources="sonar.inclusions=";
  separator=""
  for i in $(p4 opened $path/... | awk -F'#' '{ print $1 }' | grep ".*\.[c|h]"); do
      file=$(echo $i | sed "s#//depot#$P4ROOT#" | sed "s#$path/##")
      echo $file
      sources=$sources$separator$file
      separator=","
  done
  echo $sources >> ${BUILDDIR}/output/staging/sonar/sonar-project.properties
}

PROMPT_COMMAND=sah-configure-env

function verify-component() {
    if [ "$PROJECT" == "" ]; then
        echo "No sonar project defined - create sonar.config file"
    elif [[ -z "$(sah-component-path "${1}")" ]]; then
        echo "Invalid Component"
    else
        echo "Verifying component $(sah-component-name "${1}") for project $PROJECT"
        sonar_properties_create $PROJECT $(sah-component-name "${1}")
        if sonar_properties_add_module $(sah-component-name "${1}") $(sah-component-path "${1}"); then
            cd ${BUILDDIR}/output/staging/sonar
            ${SONAR_RUNNER_HOME}/bin/sonar-runner -Dsonar.dryRun=true -Dsonar.issuesReport.html.location=${BUILDDIR}/output/staging/sonar/$(sah-component-name "${1}").html -Dsonar.login=local_runner -Dsonar.password=local -Dsonar.build-stability.url=""
            cd -
            if [ "$BROWSER" != "" ]; then
                $BROWSER ${BUILDDIR}/output/staging/sonar/$(sah-component-name "${1}").html
            fi
        fi
    fi
}

function verify-opened() {
    if [ "$PROJECT" == "" ]; then
        echo "No sonar project defined - create sonar.config file"
    elif [[ -z "$(sah-component-path "${1}")" ]]; then
        echo "Invalid Component"
    else
        echo "Verifying component $(sah-component-name "${1}") for project $PROJECT"
        sonar_properties_create $PROJECT $(sah-component-name "${1}")
        if sonar_properties_add_module $(sah-component-name "${1}") $(sah-component-path "${1}"); then
            sonar_properties_add_sources $(sah-component-path "${1}")
            cd ${BUILDDIR}/output/staging/sonar
            ${SONAR_RUNNER_HOME}/bin/sonar-runner -Dsonar.dryRun=true -Dsonar.issuesReport.html.location=${BUILDDIR}/output/staging/sonar/$(sah-component-name "${1}").html -Dsonar.login=local_runner -Dsonar.password=local -Dsonar.build-stability.url=""
            cd -
            if [ "$BROWSER" != "" ]; then
                $BROWSER ${BUILDDIR}/output/staging/sonar/$(sah-component-name "${1}").html
            fi
        fi
    fi
}

function verify-complete() {
    local current previous commands
    current="${COMP_WORDS[COMP_CWORD]}"
    previous="${COMP_WORDS[COMP_CWORD-1]}"
    commands="component opened"
    if [[ "${previous}" == "verify" || "${previous}" == "help" ]]; then
        COMPREPLY=( $(compgen -W "${commands}" -- "${current}") )
    else
        COMPREPLY=( $(egrep -i "${current}.*=" "${STAGINGDIR}/buildconfig.dep" | cut -d '=' -f 1) )
    fi
}

complete -F verify-complete verify

##
# main command
function verify() {
    CMD=$1
    shift
    "verify-${CMD}" $@
}
