#!/usr/bin/perl
##
# p4 to Rel
# Quick integration script for integrating a branch to a new/latest rel
#
# Install
# In p4v select Tools -> Manage Custom Tools
# Select New -> Tool
#   Under 
#       Name:                               p4ToRel
#       Add to applicable context menus:    YES
#       Application:                        /home/sah5056/ws_w1/build/MAIN/tools/p4ToRel.pl
#       Arguments:                          --base %D (--update-latest || --update-major || --update-minor || --update-micro)
#       Start In:                           /home/sah5056/ws_w1/build/MAIN
#       Run tool in terminal:               this is up to you!
#       Refresh P4V upon completion:        YES
#       Press OK 
#       And Press OK again!
#
#       Right click on a BRANCH p4ToRel should now be visable!
#
#       Email: <david.cluytens_ext@softathome.com>
##
use strict;
use v5.10.1;
use Time::Piece;
use Getopt::Long;

my $p4From;
my $p4Dir;
my $p4To;
my $p4Rel;
my $rDate;
my $rMajor;
my $rMinor;
my $rMicro;
my $rNano;
my $chlist          = 0;
my $src_is_rel      = 0;
my $help            = 0;
my $update_mode     = 0;
my $cDate           = Time::Piece->new->strftime('%Y-%m-%d');

GetOptions('help'           => sub { $help = 1 },
        'base=s'            => \$p4From,
        'update-latest'     => sub { $update_mode = 1 },
        'update-nano'       => sub { $update_mode = 2 },
        'update-micro'      => sub { $update_mode = 3 },
        'update-minor'      => sub { $update_mode = 4 },
        'update-major'      => sub { $update_mode = 5 },
        ) or printSyntax();


if ($p4From =~ /^(.*)\/DEV\/.*\/\.\.\.$/) {
    $p4Dir = $1;
} elsif ($p4From =~ /^(.*)\/MAIN\/\.\.\.$/) {
    $p4Dir = $1;
} elsif ($p4From =~ /^(.*)\/REL\/(.*)_V([\d]+)\.([\d]+)\.([\d]+)\/\.\.\.$/) {
    $p4Dir  = $1;
    $rDate  = $2;
    $rMajor = $3;
    $rMinor = $4;
    $rMicro = $5;
    $src_is_rel = 1;
} elsif ($p4From =~ /^(.*)\/REL\/(.*)_V([\d]+)\.([\d]+)\.([\d]+)\.([\d]+)\/\.\.\.$/) {
    $p4Dir  = $1;
    $rDate  = $2;
    $rMajor = $3;
    $rMinor = $4;
    $rMicro = $5;
    $rNano  = $6;
    $src_is_rel = 1;
} else {
    printSyntax();
}

if (!$src_is_rel) {
    ($rDate, $rMajor, $rMinor, $rMicro, $rNano) = p4GetLatestComponent($p4Dir);
    my $p4Last;
    if ($update_mode == 2) {
        $p4Last = $p4Dir . "/REL/" . $rDate . "_V" . $rMajor . "." . $rMinor . "." . $rMicro . "." . $rNano . "/...";
    } else {
        $p4Last = $p4Dir . "/REL/" . $rDate . "_V" . $rMajor . "." . $rMinor . "." . $rMicro . "/...";
    }
    my $ch1     = p4GetLatestChangelistNo($p4Last);
    my $ch2     = p4GetLatestChangelistNo($p4From);
    my $desc    = p4GetLatestChangelists($p4From, $ch1, $ch2);
    print "Debug Chlist: $ch1, $ch2 From $p4From Last $p4Last\nDesc: $desc\n";
    $chlist     = p4CreateChangeList($desc);
    print "Changelist no: $chlist\n";
} else {
    $chlist     = p4CreateChangeList("\tBranched from $p4From");
}

for($update_mode) {
    when (/^1$/) {
    }
    when (/^2$/) {
        if (!$src_is_rel) {
            $rDate = $cDate;    
        }
        $rNano++;
    }
    when (/^3$/) {
        if (!$src_is_rel) {
            $rDate = $cDate;    
        }
        $rMicro++;
        $rNano=0;
    }
    when (/^4$/) {
        if (!$src_is_rel) {
            $rDate = $cDate;    
        }
        $rMinor++;
        $rMicro=0;
        $rNano=0;
    }
    when (/^5$/) {
        if (!$src_is_rel) {
            $rDate = $cDate;    
        }
        $rMajor++;
        $rMinor = 0;
        $rMicro = 0;
        $rNano=0;
    }
    default {
        printSyntax();
    }
}

if ($update_mode == 2) {
    $p4Rel = $rDate . "_V" . $rMajor . "." . $rMinor . "." . $rMicro . "." . $rNano;
} else {
    $p4Rel = $rDate . "_V" . $rMajor . "." . $rMinor . "." . $rMicro;
}
$p4To = $p4Dir . "/REL/" . $p4Rel . "/...";

if ($chlist) {
    print "p4 integrate -c $chlist -d -o $p4From $p4To\n";
    system("p4 integrate -c $chlist -d -o $p4From $p4To");
} else {
    print "p4 integrate -d -o $p4From $p4To\n";
    system("p4 integrate -d -o $p4From $p4To");

}
exit $?;

sub printSyntax {
    print "p4ToRel --base <//depot/URL/...> --update-latest | --update-major | --update-minor | --update-micro | --update-nano\n\n";
    exit 1;
}

sub trim {
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}

sub p4GetLatestComponent {
    my $location    = shift;
    my $fh;
    my $p4location;
    my $lver            = ""; 
    my $ldate           = "";
    my $cdate           = "";
    my $cmajor          = 0;
    my $cminor          = 0;
    my $cmicro          = 0;
    my $cnano           = 0;
    my $lmajor          = 0;
    my $lminor          = 0;
    my $lmicro          = 0;
    my $lnano           = 0;
    my $vbranch_major   = 0;
    my $vbranch_minor   = 0;
    my $vbranch_micro   = 0;

    $p4location = "${location}/REL/";
    if ($p4location =~ /^.*\/V([\d]+)\.([\d]+)\.([\d]+)\/REL\/$/) {
        $vbranch_major = $1; 
        $vbranch_minor = $2;
        $vbranch_micro = $3;
    }

    open($fh, "p4 -p sahwp4s01.be.softathome.com:1666 dirs ${p4location}* |");

    while (!eof($fh)) {
        my $line = <$fh>;
        my $len  = length $p4location;
        my $cver = substr $line, $len;
        my $update_ver = 0;
        $cver = trim($cver);
        if ($cver =~ /^([\d]+-[\d]+-[\d]+)_V([\d]+)\.([\d]+)\.([\d]+)$/) {
            $cdate   = $1;
            $cmajor  = $2;
            $cminor  = $3;
            $cmicro  = $4;
        }elsif ($cver =~ /^([\d]+-[\d]+-[\d]+)_V([\d]+)\.([\d]+)\.([\d]+)\.([\d]+)$/) {
            $cdate   = $1;
            $cmajor  = $2;
            $cminor  = $3;
            $cmicro  = $4;
            $cnano   = $5;
        } else {
            next;
        }

        if ($cmajor > $lmajor) {
            $update_ver = 1;
        } elsif (($cmajor == $lmajor) && ($cminor > $lminor)) {
            $update_ver = 1;
        } elsif (($cmajor == $lmajor) && ($cminor == $lminor) && ($cmicro > $lmicro)) {
            $update_ver = 1;
        } elsif (($cmajor == $lmajor) && ($cminor == $lminor) && ($cmicro == $lmicro) && ($cnano > $lnano)) {
            $update_ver = 1;
        }
        if ($update_ver) {
            $ldate  = $cdate; 
            $lmajor = $cmajor;
            $lminor = $cminor;
            $lmicro = $cmicro;
            $lnano  = $cnano;
        }
    }   
    close $fh;
    if (($ldate eq "") && (!$lmajor) && (!$lminor) && (!$lmicro) && (!$lnano)) {
        $lmajor = $vbranch_major;
        $lminor = $vbranch_minor;
        $lmicro = $vbranch_micro;
    }
    return($ldate, $lmajor, $lminor, $lmicro, $lnano);
}

sub p4GetLatestChangelistNo {
    my $fh;
    my $p4location  = shift;
    my $ch          = 0;

    open($fh, "p4 -p sahwp4s01.be.softathome.com:1666 changes -m 1 -s submitted ${p4location} |");
    while (!eof($fh)) {
        my $line = <$fh>;
        if ($line =~ /^Change\s+(\d+).*$/) {
            $ch = $1
        } 
    }   
    close $fh;
    return($ch);
}

sub p4GetLatestChangelists {
    my $fh;
    my $p4location  = shift;
    my $no1         = shift;
    my $no2         = shift;
    my $ch          = 0;
    my $list        = "";

    open($fh, "p4 -p sahwp4s01.be.softathome.com:1666 changes -l -s submitted ${p4location}\@${no1},${no2} |");
    while (!eof($fh)) {
        my $line = <$fh>;
        $line = trim($line);
        if ($line =~ /^Change \d+ on \d+\/\d+\/\d+ by \w+@\w+$/) {
            next;
        }
        $list = $list . "\t" . $line . "\n";
    }   
    close $fh;
    return($list);
}

sub p4CreateChangeList {
    my $description = shift;
    my @spec        = (); 
    my $ch_spec     = ""; 

    $description = escapeString($description);
    @spec = `p4 change -o | sed '/^Files:/q'`;
    foreach my $line (@spec) {
        $line =~ s#^\s+<enter description here>$#$description#g;
        $ch_spec = $ch_spec . $line;
    }   
    my  $chlist = `echo "$ch_spec" | p4 change -i | cut -d ' ' -f 2`;
    chomp $chlist;
    return $chlist; 
}

sub escapeString {
    my $string = shift;
    $string  =~ s/\(/\\(/g;
    $string  =~ s/\)/\\)/g;
    $string  =~ s/\'/\\'/g;
    $string  =~ s/\`/\\`/g;
    $string  =~ s/\"/\\"/g;
    $string  =~ s/\./\\./g;
    return $string;
}
