#!/usr/bin/perl
###############################################################################
#
# Tool for comparing the versions of components in the project list
# with the latest available version in perforce.
# Email <david.cluytens_ext@softathome.com>
#
###############################################################################
use strict;
use Getopt::Long;
use Cwd;
use File::Basename;
use File::Path qw(make_path remove_tree);
use File::Basename;

our $cplst          = {};
our $work_dir       = getcwd;
our $menuconfig     = 0;
my $cp_file         = "$work_dir/component.lst";
my $help            = 0;
my $proj_path;
my @files           = ();

our $GIT_CACHE_PATH = "/home/sah5056/gitcache";

GetOptions('help'           => sub { $help = 1 },
        'path=s'         => \$proj_path,
        'menuconfig'     => sub { $menuconfig = 1 }
        ) or printSyntax();

if (defined($proj_path)) {
    @files = `find $proj_path -name project.list`;
}

if ($help) {
    printSyntax();
} elsif ($proj_path ne "") {
    @files = `find $proj_path -name project.list`;
    syncKernels($cp_file, @files);
    foreach my $proj_file (@files) {
        doProjFile($cp_file, $proj_file);    
    }
} else {
    printSyntax();
}
exit 0;

sub doProjFile {
    my $cp_file     = shift;
    my $proj_file   = shift;

    print "Reading " . trim($proj_file) . "\n";
    parseComponentLst($cp_file);
    parseConfig($proj_file);
}

sub printSyntax {
    print "update_proj.pl   --path  <project_path>\
        \t--path        proj path\n\n";
    exit 1;
}

sub getP4WorkspaceDir {
    my $ws_dir = `p4 client -o | awk '/^Root:.*\$/ {print(\$2)}'`;
    $ws_dir = trim($ws_dir);
    return $ws_dir;
}

sub trim {
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}

sub parseComponentLst {
    my  $file   = shift;
    our $cplst;
    my  $cp;
    my  $location;
    my  $server;

    my @lines = readFile($file);
    foreach my $line (@lines) {
        if ($line =~ /^(.*)=(.*):(.*)$/) {
            $cp         = $1;
            $server     = $2;
            $location   = $3;
            $cplst->{$cp}->{"server"}   = $server;
            $cplst->{$cp}->{"location"} = $location;
        }
    }
}

sub getComponentServer {
    my $cp = shift;
    return $cplst->{$cp}->{"server"};
}

sub getComponentLocation {
    my $cp = shift;
    return $cplst->{$cp}->{"location"};
}

sub parseConfig {
    my $fh;
    my $file = shift;
    my @lines = readFile($file);
    foreach my $line (@lines) {
        parseLine($file, $line);
    }
}

sub updateGitCache {
    our $GIT_CACHE_PATH;
    my $location    = shift;

    my $git_component_dir = $GIT_CACHE_PATH . "$location";
    if (-e $git_component_dir and -d $git_component_dir) {
        chdir $git_component_dir;
        my $cmd = "git fetch origin +refs/heads/*:refs/heads/*";
        system($cmd);
    } else {
        $git_component_dir = dirname($git_component_dir);
        make_path  $git_component_dir;
        chdir "$git_component_dir";

        my $uri = "ssh://git.be.softathome.com/srv/git".$location;
        my $cmd = "git clone --bare $uri";
        system($cmd);
    }
}

sub getGitComponent {
    our $GIT_CACHE_PATH;
    my $location            = shift;
    my $ver                 = shift;
    my $clone_cmd           = "";
    my $git_component_dir   = $GIT_CACHE_PATH . "$location";
    my $ws_dir          = getP4WorkspaceDir();


    $ver = basename($ver);

    my $path = $ws_dir . dirname($location);
    my $repo = $path . "/" . basename($location);

    if (!(-e $repo and -d $repo)) {
        make_path($path);
        chdir($path);
        $clone_cmd       = "git clone $git_component_dir $repo";
        system($clone_cmd);
    } else {
        chdir($repo);
        $clone_cmd       = "git fetch --all";
        system($clone_cmd);
    }


    my $checkout_cmd    = "git checkout $ver";
    chdir($repo);
    system($checkout_cmd);
}

sub p4GetComponent {
    our $GIT_CACHE_PATH;
    my $server      = shift;
    my $location    = shift;
    my $ver         = shift;

    if ($server eq "w1") {
        my $p4location = "//depot${location}/${ver}/...";
        print "Syncing $p4location\n";
        system("p4 -p sahwp4s01.be.softathome.com:1666 sync -q -f ${p4location}");
    } elsif ($server eq "n1") {
        my $p4location = "/${location}/${ver}...";
        print "Syncing $p4location\n";
        system("p4 -p sahwp4s01.be.softathome.com:1999 sync -q -f ${p4location}");
    } elsif ($server eq "wg1") {
        updateGitCache($location);
        getGitComponent($location, $ver);
    }
}

sub syncKernels {
    my $cp_file = shift;
    my $projs = shift;

    parseComponentLst($cp_file);
    my %kernels = (); # hash

        foreach my $proj (@files) {
            my @lines = readFile($proj);

            foreach my $line (@lines) {
                if ($line =~ /^(sah_kernel.*)=(.*)$/) {
                    $kernels{$1}{$2} = '1';
                }
            }
        }

    for my $kernel ( keys %kernels ) {
#        print "DEBUG $kernel\n";
        for my $ver ( keys $kernels{$kernel} ) {
#            print "DEBUG $location\n";
            my $server = getComponentServer($kernel);
            my $location = getComponentLocation($kernel);
            p4GetComponent($server, $location,$ver);
        }
    }
}

sub parseLine {
    our $configs;
    our $menuconfig;
    our $work_dir;
    my $file    = shift;
    my $line    = shift;
    my $cp      = "";
    my $version;
    my $location;
    my $server;
    my $res;

    system("cd $work_dir");
    if ($line =~ /^(sah_kernel.*)=(.*)$/) {
        $cp         = $1;
        $version    = $2; 
        $server     = getComponentServer($cp);
        if ($server eq "wg1") {
            $location   = getComponentLocation($cp);
        } else {
            $location   = getComponentLocation($cp) . "/" . $version;
        }

        if ($server ne "" && $location ne "" ) {
            my $config_dir      = dirname($file);
            my $config_kernel   = $config_dir . "/kernel.config";
            my $makefile_global   = $work_dir . "/Makefile.global";

            print "cd $work_dir; $work_dir/build.sh $config_dir nosync nocleanstaging nobuild\n";
            system("cd $work_dir; $work_dir/build.sh $config_dir nosync nocleanstaging nobuild");

            my $kernel_dir=`grep HARDCO_KERNEL_DIR Makefile.global | sed 's/.*HARDCO_KERNEL_DIR=//g'`;
            chomp $kernel_dir;
            my $kernel_kernel   = $kernel_dir . "/.config";
            
#print "Sourcing Makefile.global\n";
#system("source Makefile.global");

            if (-f $config_kernel) {
                print "cp --no-preserve=mode $config_kernel $kernel_kernel\n";
                system("cp --no-preserve=mode $config_kernel $kernel_kernel");
            }


            if ($menuconfig) {
                print "Doing make -f Component.mk menuconfig\n";
                system("make -C $kernel_dir -f Component.mk menuconfig");
                my $ret = $? >> 8;
                if ($ret != 0) {
                    print "Error when handling $kernel_dir\n";
                    exit(1);
                }
            } else {
                print "Doing: source $makefile_global; make -C $kernel_dir -f Component.mk oldconfig\n";
                system("source $makefile_global; make -C $kernel_dir -f Component.mk oldconfig");
                my $ret = $? >> 8;
                if ($ret != 0) {
                    print "Error when handling $kernel_dir\n";
                    exit(1);
                }
            }

            chdir($work_dir);

# ignore it if only the date changed
            if (-f $config_kernel) {
                system("tail -n +5 $kernel_kernel > /tmp/diff1");
                system("tail -n +5 $config_kernel > /tmp/diff2");
                my $diffret = `diff /tmp/diff1 /tmp/diff2`;

                if( $diffret != '' ) {
                        print "Opening file for edit in perforce $config_kernel\n";
                        system("p4 edit $config_kernel");

                        print "cp $kernel_kernel $config_kernel\n";
                        system("cp $kernel_kernel $config_kernel");
                }
            }
        }
    } 
}

sub readFile {
    my $name = shift;
    open FILE, $name or die "File $name " . $!; 
    my @lines = <FILE>;
    close FILE;
    return @lines;
}

