#!/usr/bin/perl
###############################################################################
#
# Tool for comparing the versions of components in the project list
# with the latest available version in perforce.
# Email <david.cluytens_ext@softathome.com>
#
###############################################################################
use strict;
use lib './tools/lib';
use Getopt::Long;
use Term::ReadKey;
use File::Path qw(make_path remove_tree);
use File::Basename;
use Try::Tiny;

require 'tools/lib/core.pl';
require 'tools/lib/file.pl';
require 'tools/lib/string.pl';
require 'tools/lib/compare.pl';
require 'tools/lib/vcs.pl';
require 'tools/lib/bugzilla.pl';

our $GIT_CACHE_PATH = $ENV{'HOME'}."/gitcache";

our $chlist     = 0;
my $cp_file     = "component.lst";
my $help        = 0;
my $proj_file;
my $proj_path;

our @files              = ();
our $upgrade            = 0;
our $upgrade_method     = 5;
our $PRODUCT            = "MAINLINE_GW";
our $TARGET_MILESTONE   = "V6.1";
our $INTERACTIVE        = 0;
our $PRODUCT_ONLY       = 0;

GetOptions('help'               => sub { $help = 1 },
           'file=s'             => \$proj_file,
           'path=s'             => \$proj_path,
           'upgrade'            => sub { $upgrade = 1 },
           'upgrade_all'        => sub { $upgrade = 2 },
           'range_micro'        => sub { $upgrade_method = 3 },
           'range_minor'        => sub { $upgrade_method = 4 },
           'range_major'        => sub { $upgrade_method = 5 },
           'product=s'          => \$PRODUCT,
           'product_only'       => sub { $PRODUCT_ONLY = 1 },
           'target_milestone=s' => \$TARGET_MILESTONE,
           'interactive'        => sub {$INTERACTIVE = 1},
            ) or printSyntax();

if (defined($proj_path)) {
    @files = `find $proj_path -name project.list`;
}

parseComponentLst($cp_file);

if ($help) {
    printSyntax();
} elsif (defined($proj_file)) {
    @files = ($proj_file);
    doProjFile($proj_file);    
} elsif ($proj_path ne "") {
    @files = `find $proj_path -name project.list`;
    foreach $proj_file (@files) {
        doProjFile($proj_file);    
    }
} else {
    printSyntax();
}
exit 0;

sub doProjFile {
    my $proj_file   = shift;

    print "Comparing " . trim($proj_file) . "\n";
    ReadMode('cbreak');
    parseConfig($proj_file);
    ReadMode('normal');
}

sub printSyntax {
    print "update_proj.pl   --file  <project.list>\
\t--file                project.list\
\t--path                proj path\
\t--upgrade             upgrade only the project.list specified with --file\
\t--upgrade_all         upgrade all the project.list(s) located under --path\n\
\t--range_micro         upgrade only if the micro, pico-number and date are different\n\
\t--range_minor         upgrade only if the minor, micro, pico-number and date are different\n\
\t--range_major         upgrade only if the major, minor, micro, pico-number and date are different (default)\n\
\t--product             bugzilla project\n\
\t--target_milestone    bugzilla target_milestone\n\
\t--interactive         yes / no\n\
Example\
\t
\ttools/update_proj.pl --file ~/ws_w1/config/MAIN/sah/generic/askey-sg230/sah_multi_sip_normal/project.list --upgrade_all --path ~/ws_w1/config/MAIN/sah/generic/\
\tCompare askey-sg230/sah_multi_sip_normal/project.list with p4 and apply all the changes to the project.list(s) under  ~/ws_w1/config/MAIN/sah/generic/\n\n";
    exit 1;
}

sub parseConfig {
    my $fh;
    my $file = shift;
    my @lines = readFile($file);
    foreach my $line (@lines) {
        parseLine($file, $line);
    }
}

sub parseLine {
    our $configs;
    our $upgrade_method;
    my $file = shift;
    my $line = shift;
    my $cp;
    my $path;
    my $version;
    my $location;
    my $server;
    my $lver;
    my $cur;
    my $latest;
    my $b_type;
    my $p   = 1;
    my $ch_info = "";
    my $cp_ex   = 0;
    my $ver_result = 0;

#    printf("DEBUG LINE $line");
    if ($line =~ /^:?(.*)=(.*)\/(REL|DEV|MAIN)\/(.*)$/) {
        $cp         = $1;
        $path       = $2;
        $b_type     = $3;
        $version    = $4;
        if ($cp =~ /:(.*)/) {
            $cp = $1;
        }
        $server     = getComponentServer($cp);
        $location   = getComponentLocation($cp) . "/" . $path;
        $cp_ex      = cpIsExcluded($cp);
#        printf("DEBUG $cp $version\n");
        if (!$cp_ex && $server ne "" && $location ne "" ) {
            $lver       = getLatestComponent($server, $location, $version);
            if ($lver eq "") {
                return;
            }

            $ver_result = compareVersion($lver, $version);
            if ($upgrade_method >= $ver_result) {
#                print "DEBUG $version <-> $lver $upgrade_method >= $ver_result\n";
                $cur        = "$cp=$path/$b_type/$version";
                $latest     = "$cp=$path/REL/$lver";
                if ($cur ne $latest) {
                    $p = 0;
                    $ch_info = getChangeLists($server, $location, $version, $lver); 
                    printLine($file, $cur, $latest, $ch_info);
                }
            }
        }
    } elsif ($line =~ /^:?(.*)=(.*)\/(.*)/) {
        $cp         = $1;
        $path       = $2;
        $version    = $3; 
        if ($cp =~ /:(.*)/) {
            $cp = $1;
        }
        $server     = getComponentServer($cp);
        $location   = getComponentLocation($cp);
        $cp_ex      = cpIsExcluded($cp);
#        printf("DEBUG $cp $version\n");
        if (!$cp_ex && $server ne "" && $location ne "" ) {
            $lver       = getLatestComponent($server, $location, $version);
            if ($lver eq "") {
                return;
            }

            $ver_result = compareVersion($lver, $version);
            if ($upgrade_method >= $ver_result) {
#                print "DEBUG $version <-> $lver $upgrade_method >= $ver_result\n";
                $cur        = "$cp=$path/$version";
                if (($path eq "DEV") || ($path eq "MAIN")) {
                    $path = "REL";
                }
                $latest     = "$cp=$path/$lver";
                if ($cur ne $latest) {
                    $p = 0;
                    $ch_info = getChangeLists($server, $location, $version, $lver); 
                    printLine($file, $cur, $latest, $ch_info);
                }
            }
        }
    } 
    if ($p) {
#print $line
    }
}

sub printLine {
    our $upgrade;
    our $chlist;
    my $proj_file   = shift;
    my $cur         = shift;
    my $latest      = shift;
    my $ch_info     = shift;
    my $char;
    my $auto_accept  = 0;
    my $i_status    = 0;

    print "C $cur\nL $latest\n";
    while ($ch_info =~ /Bug\s+(\d+)\s+.*/g) {
        my $bug_no = $1;
        try {
            (my $bz_product, my $bz_milestone, my $bz_ver, my $bz_status) = getBugInfo($bug_no);
            if ($bz_status eq "COMPONENT_RELEASED") {
                $i_status = 1;
            } elsif ($bz_status eq "CONFIG_UPDATED") {
                $i_status = 2;
            } elsif ($bz_status eq "INTEGRATED") {
                $i_status = 3;
            } elsif ($bz_status eq "RESOLVED") {
                $i_status = 4;
            }
            if (($bz_product ne $PRODUCT) && (!$PRODUCT_ONLY)) {
                if ($i_status >= 1) {
                    $auto_accept |= 1;
                }
            } else {
                if (($bz_milestone eq $TARGET_MILESTONE) && !$INTERACTIVE && ($i_status >= 2)) {
                    print "Bug $bug_no\n";
                    print "Product: $PRODUCT\t\t---> Matched.\n";
                    print "Milestone: $TARGET_MILESTONE\t\t---> Matched.\n";
                    print "Interactive: not set\n";
                    $auto_accept |= 1;
                }
            }
        } catch {
            warn "caught error: $_";
        }
    }

    if ($upgrade) {
        while(1) {
            print "Update component (l/e/y/n)? ";
            if ($auto_accept) {
                print "Auto accepting release $latest for:\n $ch_info\n";
                $char = 'y';
            } else {
                if ($INTERACTIVE) {
                    $char = ReadKey(0);
                } else {
                    print "Interactive mode is not set.\nExcluding release $latest for update.\n";
                    $char = 'e';
                }
            }

            print "$char\n";
            if ($char eq "y" || $char eq "Y") {
                $chlist = createChangeList($ch_info);
                if ($upgrade == 1) {
                    updateComponent($chlist, $proj_file, $cur, $latest, $ch_info);
                } elsif ($upgrade == 2) {
                    updateComponents($chlist, $cur, $latest, $ch_info);
                }
                p4Submit($chlist);
                return;
            } elsif ($char eq "n" || $char eq "N") {
                return;
            } elsif ($char eq "l" || $char eq "L") {
                print "Changelist:\n$ch_info\n";

            } elsif ($char eq "e" || $char eq "E") {
                addCPToExcludeList($cur);
                return;
            }
        }
    }
}

sub updateComponents {
    my $chlist  = shift;
    my $from    = shift;
    my $to      = shift;
    my $ch_info = shift;
    our @files;
    foreach my $proj_file (@files) {
        updateComponent($chlist, $proj_file, $from, $to, $ch_info);
    }
}

sub updateComponent {
    my $chlist  = shift;
    my $name    = shift;
    my $from    = shift;
    my $to      = shift;
    my $ch_info = shift;

    chomp $name;
#    print "Scanning $name\n";
    my $touched = 0;

#read in file
    my @lines = readFile($name);

# check if we (need to) replace something
    my @newfile = (); 
    foreach my $line (@lines) {
        my $oldline = $line;
        $line =~ s#$from#$to#g;
        if ($line ne $oldline) {
#            print "$name : $from -> $to\n";
            $touched = 1;
        }   
        push(@newfile, $line);
    }   

# if we edit, p4 edit and rewrite
    if ($touched) {
#print ">>>>Editing $name\n";
        system "p4 sync $name";
        system "p4 edit -c $chlist $name";
        open FILE, "+>", $name or die $!; 
        print FILE @newfile;
        close FILE;
    } 
}

sub changeListIsPending {
    my $ch      = shift;
    if (!$ch) {
        return 0;
    }

    my @spec    = `p4 change -o $ch`;
    foreach my $line (@spec) {
        if ($line =~ /^Status:\s+pending.*$/) {
            return 1;
        }
    }
    return 0;
}

sub createChangeList {
    our $chlist;
    my $description = shift;
    my @spec        = ();
    my $ch_spec     = "";
    my $pending     = changeListIsPending($chlist); 

    $description = escapeString($description);
    if ($pending) { 
        @spec = `p4 change -o $chlist`;
    } else {
        @spec = `p4 change -o`;
    }

    foreach my $line (@spec) {
        if ($pending) { 
            $line =~ s#^Description:$#Description:\n$description#g;
        } else {
            $line =~ s#^\s+<enter description here>$#$description#g;
        }
        $ch_spec = $ch_spec . $line;
    }
    $chlist = `echo "$ch_spec" | p4 change -i | cut -d ' ' -f 2`;
    chomp $chlist;
    print "Returning chlist $chlist\n";
    return $chlist; 
}

sub escapeString {
    my $string = shift;
    $string  =~ s/\(/\\(/g;
                $string  =~ s/\)/\\)/g;
    $string  =~ s/\'/\\'/g;
    $string  =~ s/\`/\\`/g;
    $string  =~ s/\"/\\"/g;
    return $string;
}


