#include <linux/init.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/file.h>
#include <linux/sched.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Francois Beerten <francois.beerten_ext@softathome.com>");
MODULE_DESCRIPTION("SoftAtHome proxy between file descriptors.");
MODULE_VERSION("v1.0");

#define MAX_MONITORED_FILES (10*4)
#define MAX_PACKET_SIZE (1500)

static int fdproxy_major_number = 0;

// copied from splice.c
static ssize_t kernel_write(struct file *file, const char *buf, size_t count,
                            loff_t pos)
{
    mm_segment_t old_fs;
    ssize_t res;

    old_fs = get_fs();
    set_fs(get_ds());
    /* The cast to a user pointer is valid due to the set_fs() */
    res = vfs_write(file, (const char __user *)buf, count, &pos);
    set_fs(old_fs);

    return res;
}


static void forward_data(struct file *src, struct file *dst, char *buf)
{
    ssize_t len;
    len = kernel_read(src, 0, buf, MAX_PACKET_SIZE);
    if (len <= 0) {
        printk(KERN_ERR 
               "Could not get data from %p to forward to %p\n", src, dst);
    } else {
        len = kernel_write(dst, buf, len, 0);
        if (len <= 0) {
            printk(KERN_ERR "Could not write to %p\n", dst);
        }
    }
}

static int proxy_available_files(struct file **files, unsigned int count,
    char *buf)
{
    int i;
    int mask = 0;
    int processed = 0;
    struct poll_wqueues pwq;
    
    poll_initwait(&pwq);
    set_current_state(TASK_INTERRUPTIBLE);
    for (i = 0; i < count; i++) {
        struct file *f = files[i];
        if (f && f->f_op->poll) {
            mask = f->f_op->poll(f, &(pwq.pt));
            if (mask & POLLIN) {
                set_current_state(TASK_RUNNING);
                processed++;
                forward_data(files[i], files[i ^ 1], buf);
                set_current_state(TASK_INTERRUPTIBLE);
            }
        }
    }
    
    if (processed == 0) {
        // No file is ready, wait for one to (maybe) become available
        schedule();
    }
    set_current_state(TASK_RUNNING);
    poll_freewait(&pwq);
    
    return processed;
}

void proxy_between_files(struct file **files, unsigned int count, char *buf)
{
    /* wait for data on the files and send it to the other side */
    while (!signal_pending(current)) {
        proxy_available_files(files, count, buf);
    }
}

void convert_fd_to_files(int *fd, struct file **files, unsigned int count)
{
    unsigned int i;
    for (i = 0; i < count; i++) {
        files[i] = fget(fd[i]);
    }
}

void free_files(struct file **files, unsigned int count)
{
    unsigned int i;
    for (i = 0; i < count; i++) {
        fput(files[i]);
    }  
}

long fdproxy_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    int fd[MAX_MONITORED_FILES];
    struct file *files[MAX_MONITORED_FILES];
    unsigned int count, size;

    // the argument should be a table with an even number of fds
    size = _IOC_SIZE(cmd);
    if (size % (sizeof(int)*2) != 0) {
        return -ENOTTY;
    }
    count = _IOC_NR(cmd);
    if (count > MAX_MONITORED_FILES || count == 0 || 
            count * sizeof(int) > size || count % 2 != 0) {
        return -ENOTTY;
    }

    copy_from_user(fd, (char*)arg, count * sizeof(int));
    convert_fd_to_files(fd, files, count);
    proxy_between_files(files, count, filp->private_data);
    free_files(files, count);
    return -EINTR;
}

static int fdproxy_open(struct inode *inode, struct file *filp)
{
    filp->private_data = kmalloc(MAX_PACKET_SIZE, GFP_KERNEL);
    if (!filp->private_data) {
        printk(KERN_ERR "Not enough MEMORY \n");
        return -ENOMEM;
    }
    return 0;
}

static int fdproxy_release(struct inode *inode, struct file *filp)
{
    kfree(filp->private_data);
    return 0;
}

struct file_operations fdproxy_fops = {
    .owner = THIS_MODULE,
    .unlocked_ioctl = fdproxy_ioctl,
    .open = fdproxy_open,
    .release = fdproxy_release,
};

static void fdproxy_exit(void)
{
    unregister_chrdev(fdproxy_major_number, "fdproxy");
}

static int fdproxy_init(void)
{
    fdproxy_major_number = register_chrdev(0, "fdproxy", &fdproxy_fops);
    if (fdproxy_major_number < 0) {
        printk(KERN_ERR "Error %d registering a device number for fdproxy\n", fdproxy_major_number);
        return fdproxy_major_number;
    }

    printk(KERN_ALERT "Registered fdproxy with device numbers %d,%d\n", fdproxy_major_number, 0);

    return 0;
}

module_init(fdproxy_init);
module_exit(fdproxy_exit);

