/*
 * Copyright (c) 2009 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 * fw client - SoftAtHome firewall client
 *
 */

#include <stdlib.h>
#include <debug/sahtrace.h>
#include <pcb/utils.h>
#include <pcb/core.h>
#include <pcb/utils/variant.h>
#include "fw_client.h"

variant_t *service_enable_t2variant(const service_enable_t *data) {
	variant_t *ret = (variant_t *)calloc(1,sizeof(variant_t));
	if (!ret) {
		goto error;
	}

	if (!data) {
		if (!variant_initialize(ret,variant_type_unknown)) {
			goto error_free;
		}
		return ret;
	} else {
		if (!variant_initialize(ret,variant_type_map)) {
			goto error_free;
		}
	}

	if (!variant_map_addBool(variant_da_map(ret),"enableIPv4",data->enableIPv4)) {
		goto error_free;
	}

	if (!variant_map_addBool(variant_da_map(ret),"enableIPv6",data->enableIPv6)) {
		goto error_free;
	}
	return ret;

error_free:
	free(ret);
error:
	return NULL;
}

service_enable_t *variant2service_enable_t(const variant_t *var) {
	variant_map_iterator_t *it = NULL;
	variant_map_t *map = NULL;
	service_enable_t *ret = NULL;
	if (variant_type(var) != variant_type_map) {
		goto error;
	}
	map = variant_da_map(var);
	if (!map) {
		goto error;
	}
	ret = (service_enable_t *)calloc(1,sizeof(service_enable_t));
	if (!ret) {
		goto error;
	}
	it = variant_map_find(map,"enableIPv4");
	if (!it) {
		goto error_free;
	}
	variant_map_iterator_take(it);
	ret->enableIPv4 = variant_bool(variant_map_iterator_data(it));
	variant_map_iterator_destroy(it);
	it = variant_map_find(map,"enableIPv6");
	if (!it) {
		goto error_free;
	}
	variant_map_iterator_take(it);
	ret->enableIPv6 = variant_bool(variant_map_iterator_data(it));
	variant_map_iterator_destroy(it);
	return ret;
error_free:
	free(ret);
error:
	return NULL;
}

static service_result_t *variant2service_result_t(const variant_t *var) {
	variant_map_iterator_t *it = NULL;	// variant must be of type map
	variant_map_t *map = NULL;
	service_result_t *ret = NULL;
	if (variant_type(var) != variant_type_map) {
		goto error;
	}

	map = variant_da_map(var);
	if (!map) {
		goto error;
	}
	ret = (service_result_t *)calloc(1,sizeof(service_result_t));
	if (!ret) {
		goto error;
	}
	// get name
	it = variant_map_find(map, "name");
	if (!it) {
		goto error_free;
	}
	variant_map_iterator_take(it);
	ret->name = variant_char(variant_map_iterator_data(it));
	variant_map_iterator_destroy(it);

	// get port 
	it = variant_map_find(map, "port");
	if (!it) {
		goto error_free;
	}
	variant_map_iterator_take(it);
	ret->port = variant_char(variant_map_iterator_data(it));
	variant_map_iterator_destroy(it);

	return ret;

error_free:
	free(ret);
error:
	return NULL;
}

static request_t *build_save_request(const char *objectPath) {
    variant_t val;
    request_t *req = request_create_executeFunction(objectPath,"save",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

static bool parse_bool_reply(request_t *req, bool *retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        return false;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_bool(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;
}

bool save(peer_info_t *dest, const char *objectPath) {

    bool  retval = true;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_save_request(objectPath);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_bool_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_reset_request(const char *objectPath) {
    request_t *req = request_create_executeFunction(objectPath,"reset",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

bool reset(peer_info_t *dest, const char *objectPath) {

    bool  retval = true;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_reset_request(objectPath);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_bool_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_commit_request(const char *objectPath) {
    request_t *req = request_create_executeFunction(objectPath,"commit",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

bool commit(peer_info_t *dest, const char *objectPath) {

    bool  retval = true;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_commit_request(objectPath);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_bool_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_setService_request(const char *objectPath, const char *id, const char *sourceInterface, const char *destinationPort, const char *protocol, int32_t ipversion, bool enable) {
    request_t *req = request_create_executeFunction(objectPath,"setService",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,sourceInterface)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourceInterface",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,destinationPort)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"destinationPort",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,protocol)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"protocol",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setInt32(&val,ipversion)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"ipversion",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setBool(&val,enable)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"enable",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

static bool parse_setService_reply(request_t *req, service_result_t **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant2service_result_t(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
}

service_result_t* setService(peer_info_t *dest, const char *objectPath, const char *id, const char *sourceInterface, const char *destinationPort, const char *protocol, int32_t ipversion, bool enable) {

    service_result_t *retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_setService_request(objectPath, id, sourceInterface, destinationPort, protocol, ipversion, enable);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_setService_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_deleteService_request(const char *objectPath, const char *id) {
    request_t *req = request_create_executeFunction(objectPath,"deleteService",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

bool  deleteService(peer_info_t *dest, const char *objectPath, const char *id) {

    bool  retval = true;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_deleteService_request(objectPath, id);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_bool_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_getService_request(const char *objectPath, const char *id) {
    request_t *req = request_create_executeFunction(objectPath,"getService",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

static bool parse_getService_reply(request_t *req, variant_list_t **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_array(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
}

variant_list_t * getService(peer_info_t *dest, const char *objectPath, const char *id) {

    variant_list_t * retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_getService_request(objectPath, id);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_getService_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_setRespondToPing_request(const char *objectPath, const char *sourceInterface, const service_enable_t *service_enable) {
    request_t *req = request_create_executeFunction(objectPath,"setRespondToPing",0);
    variant_t val;
    variant_t *var_service_enable = NULL;
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,sourceInterface)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourceInterface",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    var_service_enable = service_enable_t2variant(service_enable);
    if (!var_service_enable) {
        SAH_TRACE_ERROR("Failed to create variant");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"service_enable",var_service_enable)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        variant_cleanup(var_service_enable);
        free(var_service_enable);
        goto error_cleanup_variant;
    }
    variant_cleanup(var_service_enable);
    free(var_service_enable);

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

bool setRespondToPing(peer_info_t *dest, const char *objectPath, const char *sourceInterface, const service_enable_t *service_enable) {
    bool retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_setRespondToPing_request(objectPath, sourceInterface, service_enable);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_bool_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_getRespondToPing_request(const char *objectPath, const char *sourceInterface) {
    request_t *req = request_create_executeFunction(objectPath,"getRespondToPing",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,sourceInterface)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourceInterface",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

static bool parse_getRespondToPing_reply(request_t *req, service_enable_t **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant2service_enable_t(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
}

service_enable_t *getRespondToPing(peer_info_t *dest, const char *objectPath, const char *sourceInterface) {

    service_enable_t *retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_getRespondToPing_request(objectPath, sourceInterface);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_getRespondToPing_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_setPortForwarding_request(const char *objectPath, const char *id, const char *origin, const char *sourceInterface, const char *externalPort, const char *internalPort, const char *destinationIPAddress, const char *sourcePrefix, const char *protocol, bool enable, bool persistent, const char *description) {
    request_t *req = request_create_executeFunction(objectPath,"setPortForwarding",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,origin)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"origin",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,sourceInterface)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourceInterface",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,externalPort)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"externalPort",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,internalPort)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"internalPort",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,destinationIPAddress)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"destinationIPAddress",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,sourcePrefix)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourcePrefix",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,protocol)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"protocol",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setBool(&val,enable)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"enable",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setBool(&val,persistent)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"persistent",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,description)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"description",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

static bool parse_setPortForwarding_reply(request_t *req, char **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_char(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
}

char * setPortForwarding(peer_info_t *dest, const char *objectPath, const char *id, const char *origin, const char *sourceInterface, const char *externalPort, const char *internalPort, const char *destinationIPAddress, const char *sourcePrefix, const char *protocol, bool enable, bool persistent, const char *description) {

    char * retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_setPortForwarding_request(objectPath, id, origin, sourceInterface, externalPort, internalPort, destinationIPAddress, sourcePrefix, protocol, enable, persistent, description);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_setPortForwarding_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_deletePortForwarding_request(const char *objectPath, const char *id, const char *origin, const char* destinationIPAddress) {
    request_t *req = request_create_executeFunction(objectPath,"deletePortForwarding",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,origin)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"origin",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,destinationIPAddress)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"destinationIPAddress",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

bool  deletePortForwarding(peer_info_t *dest, const char *objectPath, const char *id, const char* destinationIPAddress, const char *origin) {

    bool  retval = true;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_deletePortForwarding_request(objectPath, id, destinationIPAddress, origin);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_bool_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_getPortForwarding_request(const char *objectPath, const char *id, const char *origin) {
    request_t *req = request_create_executeFunction(objectPath,"getPortForwarding",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,origin)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"origin",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    //----------------------------------------------------

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

static bool parse_getPortForwarding_reply(request_t *req, variant_list_t **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_array(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
}

variant_list_t * getPortForwarding(peer_info_t *dest, const char *objectPath, const char *id, const char *origin) {

    variant_list_t * retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_getPortForwarding_request(objectPath, id, origin);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_getPortForwarding_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_setPinhole_request(const char *objectPath, const char *id, const char *origin, const char *sourceInterface, const char *sourcePort, const char *destinationPort, const char *destinationIPAddress, const char *sourcePrefix, const char *protocol, int32_t ipversion, bool enable, bool persistent, const char *description) {
    request_t *req = request_create_executeFunction(objectPath,"setPinhole",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,origin)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"origin",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,sourceInterface)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourceInterface",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,sourcePort)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourcePort",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,destinationPort)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"destinationPort",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,destinationIPAddress)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"destinationIPAddress",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,sourcePrefix)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourcePrefix",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,protocol)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"protocol",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setInt32(&val,ipversion)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"ipversion",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setBool(&val,enable)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"enable",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setBool(&val,persistent)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"persistent",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,description)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"description",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

static bool parse_setPinhole_reply(request_t *req, char **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_char(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
} 

char * setPinhole(peer_info_t *dest, const char *objectPath, const char *id, const char *origin, const char *sourceInterface, const char *sourcePort, const char *destinationPort, const char *destinationIPAddress, const char *sourcePrefix, const char *protocol, int32_t ipversion, bool enable, bool persistent, const char *description) {

    char * retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_setPinhole_request(objectPath, id, origin, sourceInterface, sourcePort, destinationPort, destinationIPAddress, sourcePrefix, protocol, ipversion, enable, persistent, description);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_setPinhole_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_deletePinhole_request(const char *objectPath, const char *id, const char *origin) {
    request_t *req = request_create_executeFunction(objectPath,"deletePinhole",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,origin)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"origin",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

bool  deletePinhole(peer_info_t *dest, const char *objectPath, const char *id, const char *origin) {

    bool  retval = true;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_deletePinhole_request(objectPath, id, origin);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_bool_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_getPinhole_request(const char *objectPath, const char *id, const char *origin) {
    request_t *req = request_create_executeFunction(objectPath,"getPinhole",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,origin)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"origin",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

static bool parse_getPinhole_reply(request_t *req, variant_list_t **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_array(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
}

variant_list_t * getPinhole(peer_info_t *dest, const char *objectPath, const char *id, const char *origin) {

    variant_list_t * retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_getPinhole_request(objectPath, id, origin);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_getPinhole_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_setDMZ_request(const char *objectPath, const char *id, const char *sourceInterface, const char *destinationIPAddress, const char *sourcePrefix, bool enable) {
    request_t *req = request_create_executeFunction(objectPath,"setDMZ",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,sourceInterface)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourceInterface",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,destinationIPAddress)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"destinationIPAddress",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,sourcePrefix)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourcePrefix",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setBool(&val,enable)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"enable",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    //----------------------------------------------------

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

static bool parse_setDMZ_reply(request_t *req, char **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_char(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
}

char * setDMZ(peer_info_t *dest, const char *objectPath, const char *id, const char *sourceInterface, const char *destinationIPAddress, const char *sourcePrefix, bool enable) {

    char * retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_setDMZ_request(objectPath, id, sourceInterface, destinationIPAddress, sourcePrefix, enable);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_setDMZ_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_deleteDMZ_request(const char *objectPath, const char *id) {
    request_t *req = request_create_executeFunction(objectPath,"deleteDMZ",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

bool  deleteDMZ(peer_info_t *dest, const char *objectPath, const char *id) {
    bool  retval = true;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_deleteDMZ_request(objectPath, id);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_bool_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_getDMZ_request(const char *objectPath, const char *id) {
    request_t *req = request_create_executeFunction(objectPath,"getDMZ",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

static bool parse_getDMZ_reply(request_t *req, variant_list_t **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_array(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
}

variant_list_t * getDMZ(peer_info_t *dest, const char *objectPath, const char *id) {

    variant_list_t * retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_getDMZ_request(objectPath, id);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_getDMZ_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_setCustomRule_request(const char *objectPath, const char *id, const char *chain, const char *action, const char *destinationPort, const char *sourcePort, const char *destinationPrefix, const char *sourcePrefix, const char *protocol, int32_t ipversion, bool enable, const char *description) {
    request_t *req = request_create_executeFunction(objectPath,"setCustomRule",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,chain)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"chain",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,action)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"action",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,destinationPort)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"destinationPort",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,sourcePort)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourcePort",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,destinationPrefix)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"destinationPrefix",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,sourcePrefix)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourcePrefix",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,protocol)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"protocol",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setInt32(&val,ipversion)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"ipversion",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setBool(&val,enable)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"enable",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,description)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"description",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    //----------------------------------------------------

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

static bool parse_setCustomRule_reply(request_t *req, char **retval) {
    reply_item_t *item = NULL;
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_char(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
}

char * setCustomRule(peer_info_t *dest, const char *objectPath, const char *id, const char *chain, const char *action, const char *destinationPort, const char *sourcePort, const char *destinationPrefix, const char *sourcePrefix, const char *protocol, int32_t ipversion, bool enable, const char* description) {

    char * retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_setCustomRule_request(objectPath, id, chain, action, destinationPort, sourcePort, destinationPrefix, sourcePrefix, protocol, ipversion, enable, description);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_setCustomRule_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_deleteCustomRule_request(const char *objectPath, const char *id, const char *chain) {
    request_t *req = request_create_executeFunction(objectPath,"deleteCustomRule",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,chain)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"chain",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

bool  deleteCustomRule(peer_info_t *dest, const char *objectPath, const char *id, const char *chain) {

    bool  retval = true;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_deleteCustomRule_request(objectPath, id, chain);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_bool_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_getCustomRule_request(const char *objectPath, const char *id, const char *chain) {
    request_t *req = request_create_executeFunction(objectPath,"getCustomRule",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"id",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }
    if (!variant_setChar(&val,chain)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"chain",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

static bool parse_getCustomRule_reply(request_t *req, variant_list_t **retval) {
    reply_item_t *item = NULL;
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        goto exit_error;
    }

    item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_array(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;

exit_error:
    return false;
}

variant_list_t * getCustomRule(peer_info_t *dest, const char *objectPath, const char *id, const char *chain) {

    variant_list_t * retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    request_t *req = build_getCustomRule_request(objectPath, id, chain);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_getCustomRule_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_setFirewallLevel_request(const char *objectPath, const char *level) {
    request_t *req = request_create_executeFunction(objectPath,"setFirewallLevel",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val,level)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"level",&val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
}

bool  setFirewallLevel(peer_info_t *dest, const char *objectPath, const char *level) {
    pcb_t *pcb = NULL;
    request_t *req = NULL;
    bool  retval = true;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }
    pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }
    req = build_setFirewallLevel_request(objectPath, level);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb,dest,req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb,req,NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_bool_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

    request_destroy(req);
    return retval;

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_setPolicyLevel_request(const char *objectPath, const char *id, const char *sourceInterface, const char *destinationInterface, const char *level, int32_t ipversion, bool enable) {
    (void)enable;
    request_t *req = request_create_executeFunction(objectPath, "setPolicyLevel",0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        goto error;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val, id)) {
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req, "id", &val)) {
        goto error_cleanup_variant;
    }

    if (!variant_setChar(&val, sourceInterface)) {
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req,"sourceInterface", &val)) {
        goto error_cleanup_variant;
    }

    if (!variant_setChar(&val, destinationInterface)) {
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req, "destinationInterface", &val)) {
        goto error_cleanup_variant;
    }

    if (!variant_setChar(&val, level)) {
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req, "level", &val)) {
        goto error_cleanup_variant;
    }

    if (!variant_setInt32(&val, ipversion)) {
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req, "ipversion", &val)) {
        goto error_cleanup_variant;
    }

    // Enable is not (yet) supported in setPolicyLevel()
    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    SAH_TRACE_ERROR("Failed to add/set parameter");
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
error:
    return NULL;
} 

static bool parse_setPolicyLevel_reply(request_t *req, variant_t **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        return false;
    }

    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                variant_initialize(*retval, variant_type_unknown);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s", reply_item_error(item), reply_item_errorDescription(item), reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;
}

bool setPolicyLevel(peer_info_t *dest, const char *objectPath, const char *id, const char *sourceInterface, const char *destinationInterface, const char *level, int32_t ipversion, bool enable) {
    variant_t *retval = NULL;
    request_t *req = NULL;
    pcb_t *pcb = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        goto error;
    }

    pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        goto error;
    }

    req = build_setPolicyLevel_request(objectPath, id, sourceInterface, destinationInterface, level, ipversion, enable);
    if (!req) {
        SAH_TRACE_ERROR("Failed to build request");
        goto error;
    }

    if (!pcb_sendRequest(pcb, dest, req)) {
        SAH_TRACE_ERROR("Failed to send request");
        goto error_cleanup_request;
    }

    SAH_TRACE_INFO("Function called, wait for reply");
    if (!pcb_waitForReply(pcb, req, NULL)) {
        SAH_TRACE_ERROR("Communication failed");
        goto error_cleanup_request;
    }

    if (!parse_setPolicyLevel_reply(req, &retval)) {
        SAH_TRACE_ERROR("Failed to parse reply");
        goto error_cleanup_request;
    }

error_cleanup_request:
    request_destroy(req);
error:
    return retval;
}

static request_t *build_deletePolicy_request(const char *objectPath, const char *id) {
    request_t *req = request_create_executeFunction(objectPath, "deletePolicy", 0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        return NULL;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val, id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req, "id", &val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
    return NULL;
}

bool deletePolicy(peer_info_t *dest, const char *objectPath, const char *id) {
    bool retval = true;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        return true;
    }

    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        return true;
    }

    request_t *req = build_deletePolicy_request(objectPath, id);
    if (req) {
        if (!pcb_sendRequest(pcb, dest, req)) {
            SAH_TRACE_ERROR("Failed to send request");
            goto error_cleanup_request;
        }
    
        SAH_TRACE_INFO("Function called, wait for reply");
        if (!pcb_waitForReply(pcb,req,NULL)) {
            SAH_TRACE_ERROR("Communication failed");
            goto error_cleanup_request;
        }

        if (!parse_bool_reply(req, &retval)) {
           SAH_TRACE_ERROR("Failed to parse reply");
        }
    }

error_cleanup_request:
    request_destroy(req);
    return true;
}

static bool parse_getPolicyLevel_reply(request_t *req, char **retval) {
    reply_t *reply = request_reply(req);
    if (!reply) {
        SAH_TRACE_ERROR("No reply available");
        return false;
    }

    // iterate the reply items
    reply_item_t *item = reply_firstItem(reply);
    while(item) {
        switch(reply_item_type(item)) {
            case reply_type_function_return: {
                const variant_t *val = reply_item_returnValue(item);
                *retval = variant_char(val);
            }
            break;
            case reply_type_error: {
                SAH_TRACE_WARNING("Function call error (%d) - %s - %s",reply_item_error(item),reply_item_errorDescription(item),reply_item_errorInfo(item));
            }
            break;
            default:
                SAH_TRACE_WARNING("Unexpected reply item");
            break;
        }
        item = reply_nextItem(item);
    }

    return true;
}

static request_t *build_getPolicyLevel_request(const char *objectPath, const char *id) {
    request_t *req = request_create_executeFunction(objectPath, "getPolicyLevel", 0);
    if (!req) {
        SAH_TRACE_ERROR("Failed to create request");
        return NULL;
    }

    variant_t val;
    if (!variant_initialize(&val, variant_type_unknown)) {
        SAH_TRACE_ERROR("Failed to initialize variant");
        goto error_cleanup_request;
    }

    if (!variant_setChar(&val, id)) {
        SAH_TRACE_ERROR("Failed to set value");
        goto error_cleanup_variant;
    }
    if (!request_addParameter(req, "id", &val)) {
        SAH_TRACE_ERROR("Failed to add parameter");
        goto error_cleanup_variant;
    }

    variant_cleanup(&val);
    return req;

error_cleanup_variant:
    variant_cleanup(&val);
error_cleanup_request:
    request_destroy(req);
    return NULL;
}

char * getPolicyLevel(peer_info_t *dest, const char *objectPath, const char *id) {
    char *retval = NULL;
    if (!dest || !objectPath) {
        SAH_TRACE_ERROR("Invalid parameters specified");
        return NULL;
    }

    pcb_t *pcb = pcb_fromConnection(peer_connection(dest));
    if (!pcb) {
        SAH_TRACE_ERROR("Invalid destination specified");
        return NULL;
    }

    request_t *req = build_getPolicyLevel_request(objectPath, id);
    if (req) {
        if (!pcb_sendRequest(pcb, dest, req)) {
            SAH_TRACE_ERROR("Failed to send request");
            goto error_cleanup_request;
        }
    
        SAH_TRACE_INFO("Function called, wait for reply");
        if (!pcb_waitForReply(pcb,req,NULL)) {
            SAH_TRACE_ERROR("Communication failed");
            goto error_cleanup_request;
        }

        if (!parse_getPolicyLevel_reply(req, &retval)) {
            SAH_TRACE_ERROR("Failed to parse reply");
            goto error_cleanup_request;
        }
        return retval;
    }
error_cleanup_request:
    request_destroy(req);
    return NULL;
}

