/*
 * Copyright (c) 2010 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 */
#ifndef __NEMO_CLIENT_H__
#define __NEMO_CLIENT_H__

/**
@defgroup client client.h - Client Library
@{

@brief
The NeMo Client Library is a collection of utilities to facilitate out-of-process data model access to NeMo.

@details
The NeMo Client Library contains an implementation of the NeMo Common API (see @ref common_api), to be used by client applications and
data model plugins. On top of that, the libary contains some other utilities that are not available in the NeMo Core Libary as well.

Before using the library, please read the following important preconditions:
- The application needs to have a PCB event loop. The library has functionality that registers callbacks to the PCB core library
  and it relies on getting called by the PCB event loop.
- The library has no support for multi-threading whatsoever.
- The library contains blocking I/O. The API functions send one or more requests to the NeMo core process and wait for the reply.

The functionality of the library is backed by PCB. In fact, the library is only a thin layer on top of PCB specifically for
NeMo data model access. So if one of the above preconditions do not hold for your application, consider having a look at the source
code and write your own version of it suited for your environment.

To use the NeMo Client Library, applications shall:
- Include <nemo/client.h>.
- Add the package nemo_client to the pkg-config command line, e.g. link with `pkg-config --libs nemo_client`.
- Depend on sah_lib_nemo in their Component.dep and on SAH_LIB_NEMO in their Component.in.

@author Dries De Winter <dries.dewinter@softathome.com>
*/

#include <stdbool.h>
#include <stdint.h>

#include <nemo/common_api.h>
#include <nemo/client/csi.h>
#include <nemo/client/dhcp.h>
#include <nemo/client/netdev.h>
#include <nemo/client/wlan.h>
#include <nemo/client/ra.h>

/**
@brief
Initialize the NeMo Client Library.

@details
A reference to the PCB instance needs to be passed to this function and a peer through which the NeMo core process is reachable
can be given to this initializer optionally.
These references are stored in global variables and used by almost all other functions of the client library, so this
function MUST be called BEFORE any other function of the library is called. This also implies that these objects MUST NOT be
destroyed until the library's references to them are erased. See also nemo_client_cleanup().

If no peer is given, the client library tries to setup its own connection using the following list of methods in descending order
of priority:
-# Connect to the PCB URI given by the environmental variable NEMO_URI.
-# Connect to pcb://ipc:[/var/run/nemo] which is the Unix socket where NeMo is listening on by default.
-# Connect to the PCB URI given by the environmental variable PCB_SYS_BUS. If a direct connection to NeMo is not possbile, fall
   back to the PCB system bus. Notice that this method comes with a performance penalty.
-# Connect to pcb://ipc:[/var/run/pcb_sys] which is the Unix socket where PCB system bus is listening on by default.

All syslog messages of the library are logged using sahtrace zone "nemo". If this zone is not defined when calling
nemo_client_initialize(), it's created automatically with default log level 300 (TRACE_LEVEL_NOTICE). This can be overruled by
defining the zone and the corresponding log level before calling nemo_client_initialize().

The library keeps a global reference counter which is incremented each time nemo_client_initialize() is called and decremented each
time nemo_client_cleanup() is called. If the reference counter is non-zero when this function is entered, it does nothing. Caution:
if this function failed the first time it was called, it will always fail until the reference counter reaches zero again because
incrementing the reference counter does not imply retrying to connect if it failed previously.

@param pcb The PCB instance
@param peer The peer through which the NeMo core process is reachable or NULL to let the library set up its own connection.
@return True if the connection is set up successfully. False if no of the connection methods succeeded. Notice that if the caller
        provided a non-NULL peer, this function always returns true.
*/
bool nemo_client_initialize(pcb_t *pcb, peer_info_t *peer);

/**
@brief
Cleanup the NeMo Client Library.

@details
This erases the references to the PCB instance and the peer passed to nemo_client_initialize().
Without these references, the library is not functional, so this function MUST be called AFTER the last call to any other function
of the library, but BEFORE the objects referenced by pcb and peer are destroyed.

If the connection to NeMo was set up by the library itself, using one the methods described in the documentation of
nemo_client_initialize(), this function will close the connection. If the caller provided a peer to the library, the peer is left
untouched by this function.

The library keeps a global reference counter which is incremented each time nemo_client_initialize() is called and decremented each
time nemo_client_cleanup() is called. Only if the reference counter reaches zero when calling this function, this function
does something effectively. Caution: this means that the references to the PCB instance and the peer passed to
the FIRST call to nemo_client_initialize() are kept until the LAST call to nemo_client_cleanup().
*/
void nemo_client_cleanup();

/**
@brief
Returns the PCB instance used by the library.

@return the PCB instance used by the library, passed to nemo_client_initialize().
*/
pcb_t *nemo_client_pcb();

/**
@brief
Returns the peer used by the library.

@return the peer used by the library, passed to or set up by nemo_client_initialize().
*/
peer_info_t *nemo_client_peer();

/**
@}
*/

#endif
