/*
 * Copyright (c) 2010 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 */

/**
@defgroup core core.h - Core Library
@{

@brief
The NeMo Core Library contains the core logic of NeMo and it's intended to have only one instance, namely inside the NeMo core
process.

@details
Although it is not intended to be used by other applications, the library has a publicly available set of header files anyway
because NeMo supports dynamic module loading. These modules are supposed to include one or more of these header files and to link
against the core library and not against the client library.
Compare it with dynamic loadable kernel modules that make use of the public available kernel API for example.

By linking with the core library and including its API, modules can directly query and manipulate the internal data structures of
NeMo. On top of that, the libary contains an implementation of the NeMo Common API like the client library does, see also @ref
common_api. This way, the common API is available to both modules running inside NeMo and applications running in other processes.

Before starting to write a NeMo module, please read the following important preconditions:
- NeMo is a critical component of the SoftAtHome Operating Platform and its availability/stability must never be compromised.
- The NeMo core process is single threaded. It is forbidden to create threads in modules.
- The NeMo core process is and should remain completely asynchronous, so modules MUST NOT contain blocking I/O. Modules are allowed
  to get input from external sources, but they should add their file descriptors to the PCB event loop and process the available
  input in a corresponding callback function which must also not block in that case. Notice that if this important precondition is
  not met, the responsiveness of NeMo is put at stake affecting all dependants.
- Given the heavy load that is put on NeMo, both in terms of memory (huge data model) and performance (data model is highly dynamic
  putting heavy load on the PCB serializer and the query recalucation mechanism), it is important to have the possibility to run it
  on a PC on which memory and performance analysis tools are available. Not only to run it, but also to get as close as possible to
  the real life situation while running it on the PC, implying that it should model the network from the HGW, not from the PC
  and that all dependent services that run on the HGW can connect seamlessly to the NeMo instance running on the PC. This is
  quite a challenge, but it has been done before and it should remain possible. This means that modules should be portable and that
  there should be no relation between the system that is modeled by the data model and the system on which NeMo runs. Examples:
  - Modules must never invoke system commands like ip, iptables etc.
  - Modules must never talk directly to the system, e.g. ioctls are forbidden.
  - Modules may connect to IPC server sockets of other plugins without going through the PCB system bus (performance optimization),
    as long as fallback over the network, e.g. via the system bus, is possible.
.
Notice that although your own code might meet these preconditions, the libraries against which your code links might not. So
extra prudence is needed when including libraries. An example pitfall is to use a third-party library that calls exit() if it does
not like the input it gets. Another mistake would be to load a hardco library inside the NeMo core process because it clearly
violates the last precondition.

Shortly, you are discouraged to write a NeMo module. However, some functionality simply can't be implemented elsewhere:
- Volatile parameters like uptime counters need a parameter read handler which can only be implemented inside NeMo.
- Functions with query support. Functions can be implemented out-of-process, but query support can't (e.g. getDHCPOption).
- To map a MIB to a dedicated plugin, some code needs to be written that runs inside NeMo to establish the mapping.
.
Another valid reason to implement functionality in a NeMo module is if the performance penalty of going out of process is not worth
the relaxation of preconditions.

To use the NeMo Core Library, modules shall:
- Include <nemo/core.h>.
- Add the package nemo_core to the pkg-config command line, e.g. link with `pkg-config --libs nemo_core`.
- Depend on sah_lib_nemo in their Component.dep and on SAH_LIB_NEMO in their Component.in.
- Compile into a shared object.
- Be MTK modules preferably.

@author Dries De Winter <dries.dewinter@softathome.com>
*/

#include <nemo/core/flags.h>
#include <nemo/core/intf.h>
#include <nemo/core/traverse.h>
#include <nemo/core/query.h>
#include <nemo/core/plugin.h>
#include <nemo/core/arg.h>
#include <nemo/core/map.h>

#include <nemo/common_api.h>

/**
@}
*/
