/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <string.h>
#include <stdlib.h>

#include "core_priv.h"

#include <components.h>

#ifdef CONFIG_PCB_ACL_USERMNGT
#include <usermngt/usermngt.h>
#endif

#include <debug/sahtrace.h>

#include <pcb/common/error.h>
#include <pcb/core/datamodel.h>

/**
@file
@brief
Implementation of core common functions
*/

/**
@brief
Checks that a given name (object or parameter) is valid

@details
The following names are not valid:
    - Names containing one of the following characters './$?*[]+@!='
    - Full numeric names, numbers are allowed as long as there is at least one non numeric character in the name.
*/

bool isValidName(string_t *name) {
    if (strpbrk(string_buffer(name),"./$?*[]+@!=#") != NULL ||
        string_isNumeric(name)) {
        pcb_error = pcb_error_invalid_name;
        SAH_TRACEZ_ERROR("pcb","Name %s is not valid",string_buffer(name));
        return false;
    }

    return true;
}


bool isValidNameChar(const char *name) {
    string_t buffer;
    string_initialize(&buffer,64);
    string_fromChar(&buffer,name);
    bool retval = isValidName(&buffer);
    string_cleanup(&buffer);

    return retval;
}

ACL_t *aclFind(llist_t *ACL, uint32_t id, uint16_t flags) {
    /* search an ACL for the user/group */
    llist_iterator_t *it = NULL;
    ACL_t *acl = NULL;
    llist_for_each(it, ACL) {
        acl = llist_item_data(it,ACL_t,it);
        if (acl->id == id) {
            if (((flags & acl_group_id) && (acl->acl_flags & acl_group_id)) ||
                (!(flags & acl_group_id) && !(acl->acl_flags & acl_group_id))) {
                break;
            }
        }
        acl = NULL;
    }

    return acl;
}

bool aclSet(llist_t *ACL, uint32_t id, uint16_t flags) {
    ACL_t *acl = aclFind(ACL, id, flags);

    if (!acl) {
        acl = (ACL_t *)calloc(1,sizeof(ACL_t));
        if (acl) {
            llist_iterator_initialize(&acl->it);
            acl->id = id;
        } else {
            pcb_error = pcb_error_no_memory;
            return false;
        }
        if (!llist_append(ACL,&acl->it)) {
            free(acl);
            return false;
        }
    }

    acl->acl_flags = flags;
    return true;
}

#ifdef CONFIG_PCB_IGNORE_ACL
bool aclCheck(llist_t *ACL, uint32_t uid, uint16_t flags) {
    (void)ACL;
    (void)uid;
    (void)flags;

    return true;
}
#else
#ifdef CONFIG_PCB_ACL_USERMNGT
bool aclCheck(llist_t *ACL, uint32_t uid, uint16_t flags) {
    if (!ACL) {
        SAH_TRACEZ_INFO("pcb_acl","Access denied, no acl available");
        return false;
    }

    bool userfound = false;
    bool groupfound = false;
    bool userallowed = false;
    bool groupallowed = false;
    bool worldallowed = false;

    SAH_TRACEZ_INFO("pcb_acl","User id = %d",uid);
    llist_iterator_t *it = NULL;
    ACL_t *acl = NULL;
    const usermngt_user_t *user = usermngt_userFindByID(uid);

    llist_for_each(it,ACL) {
        acl = llist_item_data(it,ACL_t,it);
        if (acl->acl_flags & acl_group_id) {
            if (user && usermngt_userIsInGroup(user,acl->id)) {
                SAH_TRACEZ_INFO("pcb_acl","User in group %d ACL flags = %X", acl->id,acl->acl_flags);
                groupfound = true;
                const usermngt_group_t *grp = usermngt_groupFindByID(acl->id);
                if (!grp) {
                    SAH_TRACEZ_WARNING("pcb_acl","Group ID %d not found - ignore", acl->id);
                    continue;
                }
                if((acl->acl_flags & flags) == flags && usermngt_groupEnable(grp)) {
                    groupallowed = true;
                }
            }
        } else {
            if (acl->id == uid && acl->id != UINT32_MAX) {
                SAH_TRACEZ_INFO("pcb_acl","User id in ACL flags = %X", acl->acl_flags);
                userfound = true;
                if ((acl->acl_flags & flags) ==flags && user && usermngt_userEnable(user)) {
                    userallowed = true;
                }
            }
        }
        if (acl->id == UINT32_MAX) {
            SAH_TRACEZ_INFO("pcb_acl","Every one in ACL, flags = %x",acl->acl_flags);
            if ((acl->acl_flags & flags) == flags) {
                worldallowed = true;
            }
        }
    }

    if (userfound) {
        return userallowed;
    }
    if (groupfound) {
        return groupallowed;
    }
    return worldallowed;
}
#else
bool aclCheck(llist_t *ACL, uint32_t uid, uint16_t flags) {
    if (!ACL) {
        SAH_TRACEZ_INFO("pcb","Access denied, no acl available");
        return false;
    }

    bool userfound = false;
    bool groupfound = false;
    bool userallowed = false;
    bool groupallowed = false;
    bool worldallowed = false;

    llist_iterator_t *it = NULL;
    ACL_t *acl = NULL;

    llist_for_each(it,ACL) {
        acl = llist_item_data(it,ACL_t,it);
        if (acl->acl_flags & acl_group_id) {
            /* should check the groups that the user is in */
        } else {
            if (acl->id == uid) {
                userfound = true;
                if (acl->acl_flags & flags) {
                    userallowed = true;
                }
            }
        }
        if (acl->id == UINT32_MAX && (acl->acl_flags & flags)) {
            worldallowed = true;
        }
    }

    if (userfound) {
        return userallowed;
    }
    if (groupfound) {
        return groupallowed;
    }
    return worldallowed;
}
#endif
#endif

void aclDestroy(llist_t *ACL) {
    if (!ACL) {
        return;
    }

    llist_iterator_t *it = llist_takeFirst(ACL);
    ACL_t *acl = NULL;
    while(it) {
        acl = llist_item_data(it,ACL_t,it);
        free(acl);
        it = llist_takeFirst(ACL);
    }
}
