/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <debug/sahtrace.h>

#include <pcb/core.h>

static bool default_set_parameter_value(function_call_t *fcall, object_t *object, const char *name, variant_t *value, uint32_t uid) {
    bool rv = false;
    parameter_t *param = object_getParameter(object, name);
    if (!param) {
        fcall_addErrorReply(fcall, pcb_error_not_found, error_string(pcb_error_not_found), name);
        goto exit;
    }
    if (request_userID(fcall_request(fcall)) != 0 && parameter_isReadOnly(param)) {
        fcall_addErrorReply(fcall, pcb_error_read_only, error_string(pcb_error_read_only), name);
        goto exit;
    }
    if (!parameter_canWrite(param, uid)) {
        fcall_addErrorReply(fcall, EACCES, error_string(EACCES), name);
        goto exit;
    }
    if (!parameter_setValue(param, value)) {
        parameter_rollback(param);
        fcall_addErrorReply(fcall, pcb_error, error_string(pcb_error), name);
        goto exit;
    }

    rv = true;

exit:
    return rv;
}

static bool default_set_from_arg_list(function_call_t *fcall, argument_value_list_t *args, uint32_t uid) {
    const char *name = NULL;
    variant_t *value = NULL;
    object_t *object = fcall_object(fcall);
    bool rv = true;

    argument_value_t *arg = argument_valueTakeArgumentFirst(args);
    while(arg) {
        name = argument_valueName(arg);
        value = argument_value(arg);

        if (!name || !(*name)) {
            fcall_addErrorReply(fcall, pcb_error_invalid_name, error_string(pcb_error_invalid_name), name);
            continue;
        }
        rv &= default_set_parameter_value(fcall, object, name, value, uid);

        argument_valueDestroy(arg);
        arg = argument_valueTakeArgumentFirst(args);
    }

    return rv;
}

static bool default_set_from_map(function_call_t *fcall, variant_map_t *parameters, uint32_t uid) {
    bool rv = true;
    variant_map_iterator_t *it = NULL;
    const char *name = NULL;
    variant_t *value = NULL;
    object_t *object = fcall_object(fcall);

    /* loop over all parameters in the variant map */
    variant_map_for_each(it, parameters) {
        name = variant_map_iterator_key(it);
        value = variant_map_iterator_data(it);

        rv &= default_set_parameter_value(fcall, object, name, value, uid);
    }

    return rv;
}

function_exec_state_t __get(function_call_t *fcall, argument_value_list_t *args, variant_t *retval) {
    /* clear the arghument list in case the caller has added arguments
       if the argument list is not cleared, all arguments will be sent back as out arguments
    */
    argument_valueClear(args);

    uint32_t uid = request_userID(fcall_request(fcall));
    variant_initialize(retval, variant_type_map);
    object_update(fcall_object(fcall));
    object_getValuesWithACL(fcall_object(fcall), variant_da_map(retval), uid);

    /* end the RPC call */
    return function_exec_done;
}

function_exec_state_t __set(function_call_t *fcall, argument_value_list_t *args, variant_t *retval) {
    uint32_t attr = request_attributes(fcall_request(fcall));
    variant_map_t *parameters = NULL;
    bool rv = false;
    variant_t *data = NULL;
    argument_value_t *arg = NULL;

    argument_get(&arg, args, attr, "parameters");
    if (!arg) {
        if (attr & request_function_args_by_name) {
            rv = default_set_from_arg_list(fcall, args, request_userID(fcall_request(fcall)));
            goto exit_commit;
        } else {
            SAH_TRACEZ_ERROR("pcb_def", "Missing argument 'parameters'' - or parameters not given by name");
            /* nothing to be done */
            fcall_addErrorReply(fcall, pcb_error_function_argument_missing, error_string(pcb_error_function_argument_missing), "parameters");
            goto exit;
        }
    }

    data = argument_value(arg);
    if (variant_type(data) != variant_type_map) {
        SAH_TRACEZ_ERROR("pcb_def", "Type mismatch");
        fcall_addErrorReply(fcall, pcb_error_type_mismatch, "Argument has wrong type", "parameters");
        /* nothing to be done */
        goto exit;
    }

    parameters = variant_da_map(data);

    /* clear the arghument list in case the caller has added arguments
       if the argument list is not cleared, all arguments will be sent back as out arguments
    */
    argument_valueClear(args);

    rv = default_set_from_map(fcall, parameters, request_userID(fcall_request(fcall)));

exit_commit:
    if (!object_commit(fcall_object(fcall))) {
        fcall_addErrorReply(fcall, pcb_error, error_string(pcb_error), "object");
        rv = false;
        object_rollback(fcall_object(fcall));
    }

exit:
    variant_setBool(retval, rv);

    /* cleanup memory */
    variant_map_cleanup(parameters);
    free(parameters);

    /* end the RPC call */
    return function_exec_done;
}

function_exec_state_t __getAclInfo(function_call_t *fcall, argument_value_list_t *args, variant_t *retval) {
    uint32_t attr = request_attributes(fcall_request(fcall));
    object_t *object = fcall_object(fcall);
    int32_t userId = 0;
    bool all = false;

    argument_getInt32(&userId, args, attr, "userID", UINT32_MAX);
    argument_getBool(&all, args, attr, "all", true);
    argument_valueClear(args);

    SAH_TRACE_INFO("USER ID = 0x%8.8X", userId);

    char ar[4] = "---";
    string_t path;
    string_initialize(&path, 64);
    object_path(object, &path, path_attr_key_notation);

    variant_initialize(retval, variant_type_map);
    variant_map_t *data = variant_da_map(retval);

    ar[0] = object_canRead(object, userId)?'R':'-';
    ar[1] = object_canWrite(object, userId)?'W':'-';
    ar[2] = object_canExecute(object, userId)?'X':'-';
    variant_map_addChar(data, string_buffer(&path), ar);

    variant_map_iterator_t *mpit = variant_map_iterator_createMapMove("Parameters", NULL);
    variant_map_t *mp = variant_da_map(variant_map_iterator_data(mpit));

    parameter_t *parameter = NULL;
    object_for_each_parameter(parameter, object) {
        ar[0] = parameter_canRead(parameter, userId)?'R':'-';
        ar[1] = parameter_canWrite(parameter, userId)?'W':'-';
        ar[2] = parameter_canReadValue(parameter, userId)?'X':'-';
        if (all || strcmp(ar, "---") != 0) {
            variant_map_addChar(mp, parameter_name(parameter), ar);
        }
    }
    if (variant_map_isEmpty(mp)) {
        variant_map_iterator_destroy(mpit);
        mpit = NULL;
    } else {
        variant_map_append(data ,mpit);
    }

    variant_map_iterator_t *mfit = variant_map_iterator_createMapMove("Functions", NULL);
    variant_map_t *mf = variant_da_map(variant_map_iterator_data(mfit));

    function_t *function = NULL;
    object_for_each_function(function, object) {
        ar[0] = function_canRead(function, userId)?'R':'-';
        ar[1] = '-';
        ar[2] = function_canExecute(function, userId)?'X':'-';
        if (all || strcmp(ar, "---") != 0) {
            variant_map_addChar(mf, function_name(function), ar);
        }
    }
    if (variant_map_isEmpty(mf)) {
        variant_map_iterator_destroy(mfit);
        mfit = NULL;
    } else {
        variant_map_append(data ,mfit);
    }

    if (!all && variant_map_size(data) == 1) {
        variant_initialize(retval, variant_type_unknown);
    }
    string_cleanup(&path);
    /* end the RPC call */
    return function_exec_done;
}
