/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <stdlib.h>
#include <string.h>

#include <debug/sahtrace.h>

#include <pcb/common/error.h>
#include <pcb/utils/connection.h>
#include <pcb/core/datamodel.h>
#include <pcb/core/pcb_main.h>
#include <pcb/core/reply.h>
#include <pcb/pcb_client.h>

#include "core_priv.h"
#include "object_priv.h"
#include "pcb_main_priv.h"

object_t *object_searchCache(object_t *object, string_t *fullpath, uint32_t pathAttr, bool *exactMatch) {
    pcb_t *pcb = datamodel_pcb(object_datamodel(object));
    if (!pcb) {
        pcb_error = pcb_error_not_connected;
        return NULL;
    }

    datamodel_t *cache = pcb_cache(pcb);
    object_t *cache_root = datamodel_root(cache);
    object_t *foundObject = NULL;
    object_t *child = NULL;
    string_t cache_path;
    string_t path;
    string_initialize(&cache_path,0);
    string_initialize(&path,0);
    object_for_each_child(child,cache_root) {
        SAH_TRACEZ_INFO("pcb_cache","Child %s",object_name(child,pathAttr));
        object_path(child,&cache_path,pathAttr);
        SAH_TRACEZ_INFO("pcb_cache","Compare: %s - %s",string_buffer(&cache_path),string_buffer(fullpath));
        if (strncmp(string_buffer(&cache_path),string_buffer(fullpath),string_length(&cache_path)) == 0) {
            if (string_length(fullpath) == string_length(&cache_path)) {
                foundObject = child;
                break;
            } else {
                string_mid(&path,fullpath,string_length(&cache_path),UINT32_MAX);
                foundObject = local_object_getObject(child,string_buffer(&path),pathAttr,exactMatch);
                if (foundObject) {
                    break;
                }
            }

        }
    }
    string_cleanup(&cache_path);
    string_cleanup(&path);

    return foundObject;
}

bool remote_object_delete(object_t *object) {
    if (!object_isInstance(object)) {
        pcb_error = pcb_error_not_instance;
        return false;
    }
    pcb_t *pcb = datamodel_pcb(object_datamodel(object));
    if (!pcb) {
        pcb_error = pcb_error_not_connected;
        return false;
    }

    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    peer_info_t *dest = connection_find(&pcb->connection,ri->fd);
    if (!dest) {
        pcb_error = pcb_error_not_connected;
        return false;
    }

    string_t path;
    string_initialize(&path,0);
    object_path(object,&path,pcb->client_data.attributes);
    bool retval = pcb_client_delete_instance(dest,string_buffer(&path));
    string_cleanup(&path);

    if (retval) {
        pcb_cacheRemoveObject(pcb,object);
    }

    return retval;
}

object_t *remote_object_getObject(object_t *object, const char *path, const uint32_t pathAttr, bool *exactMatch, remote_object_info_t *ri) {
    object_t *foundObject = NULL;
    peer_info_t *dest = NULL;

    pcb_t *pcb = datamodel_pcb(object_datamodel(object));
    if (!pcb) {
        pcb_error = pcb_error_not_connected;
        return NULL;
    }

    if (object_parent(object) != NULL) {
        if (ri == NULL) {
            ri = (remote_object_info_t *)object->userData;
            if (!ri) {
                pcb_error = pcb_error_not_connected;
                return NULL;
            }
        }
        dest = connection_find(&pcb->connection,ri->fd);
        if (!dest) {
            pcb_error = pcb_error_not_connected;
            return NULL;
        }
    }

    string_t fullpath;
    string_initialize(&fullpath,0);
    object_path(object,&fullpath,pathAttr);
    if (string_length(&fullpath)) {
        if (pathAttr & path_attr_slash_seperator) {
            string_appendChar(&fullpath,"/");
        } else {
            string_appendChar(&fullpath,".");
        }
    }
    string_appendChar(&fullpath,path);

    foundObject = object_searchCache(object,&fullpath,pathAttr,exactMatch);

    if (foundObject) {
        string_cleanup(&fullpath);
        return foundObject;
    }

    if (dest) {
        uint32_t store_attrib = pcb->client_data.attributes;
        pcb->client_data.attributes =  pathAttr;
        SAH_TRACEZ_INFO("pcb","request path = %s",string_buffer(&fullpath));
        foundObject = pcb_client_get_object(dest,string_buffer(&fullpath));
        pcb->client_data.attributes = store_attrib;
    }

    string_cleanup(&fullpath);
    return foundObject;
}

bool remote_object_hasChildren(const object_t *object) {
    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    return (ri->childCount > 0);
}

uint32_t remote_object_childCount(const object_t *object) {
    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    return ri->childCount;
}

void remote_object_path(const object_t *object, string_t *path, const uint32_t pathAttributes) {
    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    string_clear(path);

    if (!ri) {
        return;
    }

    if (pathAttributes & path_attr_include_namespace) {
        const char *dm_namespace = datamodel_namespace(object_datamodel(object));
        string_appendFormat(path, "%s", dm_namespace);
    }

    if (pathAttributes & path_attr_key_notation) {
        if (!string_isEmpty(path) && !string_isEmpty(&ri->parentPathKey)) {
            string_appendChar(path,".");
        }
        string_append(path,&ri->parentPathKey);
    } else {
        if (!string_isEmpty(path) && !string_isEmpty(&ri->parentPathIndex)) {
            string_appendChar(path,".");
        }
        string_append(path,&ri->parentPathIndex);
    }

    if (pathAttributes & path_attr_slash_seperator) {
        string_replaceChar(path,".","/");
    }

    if (!(pathAttributes & path_attr_parent)) {
        if (!string_isEmpty(path)) {
            if (pathAttributes & path_attr_slash_seperator) {
                string_appendChar(path,"/");
            } else {
                string_appendChar(path,".");
            }
        }

        if (pathAttributes & path_attr_key_notation) {
            string_append(path,&ri->key);
        } else {
            string_appendChar(path,object->oname);
        }
    }
}

object_t *remote_object_createInstance(object_t *object, uint32_t index, const char *key) {
    pcb_t *pcb = datamodel_pcb(object_datamodel(object));
    if (!pcb) {
        pcb_error = pcb_error_not_connected;
        return NULL;
    }

    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    peer_info_t *dest = connection_find(&pcb->connection,ri->fd);
    if (!dest) {
        pcb_error = pcb_error_not_connected;
        return NULL;
    }

    string_t path;
    string_initialize(&path,0);
    object_path(object,&path,pcb->client_data.attributes);
    object_t *instance = pcb_client_add_instance(dest,string_buffer(&path),index,key,NULL);
    string_cleanup(&path);

    return instance;
}

bool remote_object_hasInstances(const object_t *object) {
    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    return (ri->instanceCount > 0);
}

uint32_t remote_object_instanceCount(const object_t *object) {
    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    return ri->instanceCount;
}

parameter_t *remote_object_getParameter(const object_t *object, const char *parameterName) {
    pcb_t *pcb = datamodel_pcb(object_datamodel(object));
    if (!pcb) {
        pcb_error = pcb_error_not_connected;
        return NULL;
    }

    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    peer_info_t *dest = connection_find(&pcb->connection,ri->fd);
    if (!dest) {
        pcb_error = pcb_error_not_connected;
        return NULL;
    }

    string_t path;
    string_initialize(&path,0);
    object_path(object,&path,pcb->client_data.attributes);
    parameter_t *parameter = pcb_client_get_parameter(dest,string_buffer(&path),parameterName);
    string_cleanup(&path);

    return parameter;
}

bool remote_object_hasParameters(const object_t *object) {
    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    return (ri->parameterCount > 0);
}

uint32_t remote_object_parameterCount(const object_t *object) {
    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    return ri->parameterCount;
}

function_t *remote_object_getFunction(object_t *object, const char *functionName) {
    pcb_t *pcb = datamodel_pcb(object_datamodel(object));
    if (!pcb) {
        pcb_error = pcb_error_not_connected;
        return NULL;
    }

    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    peer_info_t *dest = connection_find(&pcb->connection,ri->fd);
    if (!dest) {
        pcb_error = pcb_error_not_connected;
        return NULL;
    }

    string_t path;
    string_initialize(&path,0);
    object_path(object,&path,pcb->client_data.attributes);

    function_t *func = pcb_client_get_function(dest,string_buffer(&path),functionName);
    string_cleanup(&path);
    return func;
}

bool remote_object_hasFunctions(const object_t *object) {
    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    return (ri->functionCount > 0);
}

uint32_t remote_object_functionCount(const object_t *object) {
    remote_object_info_t *ri = (remote_object_info_t *)object->userData;
    return ri->functionCount;
}

