/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <stdlib.h>

#include <debug/sahtrace.h>

#include <pcb/common/error.h>
#include <pcb/utils/linked_list.h>

/**
@file
@brief
Implementation of the linked list functions
*/

/**
@ingroup pcb_utils_linked_list
@brief
Initialize a linked list.

@details
This function <b>initializes</b> a <b>linked list</b>. \n
It is mandatory to initialize a linked list (llist_t) before using it.\n

@param list the list to be initialized

@return
    - true: Initialization was succesfull, the llist_t item is ready to be used
    - false: An error has occurred
        - the list item is a NULL pointer
*/
bool llist_initialize(llist_t *list) {
    if (!list) {
        return false;
    }

    list->head = NULL;
    list->tail = NULL;

    return true;
}

/**
@ingroup pcb_utils_linked_list
@brief
Initialize a <b>linked list iterator</b> (=list item)

@details
This function <b>initializes</b> a <b>linked list iterator</b>. \n
This function call initializes the next, previous and list pointer of a linked list iterator to NULL. \n

@param it the linked list iterator that has to be initialized

@return
    - true: Initialization was succesfull, the llist_t_operator_t item is ready to be used
    - false: An error has occurred
        - the linked list iterator is a NULL pointer
*/
bool llist_iterator_initialize(llist_iterator_t *it) {
    if (!it){
        return false;
    }

    it->next = NULL;
    it->prev = NULL;
    it->list = NULL;

    return true;
}

/**
@ingroup pcb_utils_linked_list
@brief
Cleanup a linked list.

@details
This funtion removes all iterators from a list.\n
After calling this function, the linked list will be empty and can be re-used.\n

@warning
The elements in the list are not deleted, freeing up the memory used by the iterators must be done
manually.\n
The memory taken by the linked list itself is not freed, freeing up the memory taken by the linked list
must be done manually.

@param list the list that has to be cleaned up
*/
void llist_cleanup(llist_t *list) {
    if (!list){
        return;
    }

    llist_iterator_t *it = llist_takeFirst(list);
    while(it) {
        it = llist_takeFirst(list);
    }
}

/**
@ingroup pcb_utils_linked_list
@brief
Get a pointer to the specified list element.

@details
This function returns a pointer to the list element at the specified index of the linked list. \n
Due to the nature of a linked list the items are not garanteed to be at the same place all the time.\n
Use this function with care, it can have performance impact, the list is iterated to find the correct element.\n

@warning
Do not free the returned pointer before removing it from the linked list. Doing so will corrupt the linked list
itself. Removing the iterator from the linked list can be done by using the function @ref llist_iterator_take.

@param list the linked list
@param index the item # in which you are interested (start counting from 0)

@return
    - a <b>pointer</b> to the specified list element in the linked list.
    - NULL if the index is bigger then the number of elements in the list
    - NULL if the list parameter is NULL.
*/
llist_iterator_t *llist_at(const llist_t *list, unsigned int index) {
    if (!list) {
        return NULL;
    }

    unsigned int pos = 0;
    llist_iterator_t *it = llist_first(list);

    while(it && pos < index) {
        pos++;
        it = llist_iterator_next(it);
    }
    if (pos == index) {
        return it;
    } else {
        pcb_error = pcb_error_out_of_boundaries;
        return NULL;
    }
}

/**
@ingroup pcb_utils_linked_list
@brief
Insert a list iterator before the reference iterator.

@details
This function inserts the specified list iterator before the reference iterator in the list. If the
element was part of another list then the element will be automatically removed from that list.\n
If the reference is a NULL pointer the insert iterator will be added before the first element in
the list, the same behavior can be achieved with @ref llist_prepend

@warning
The iterator pointer ownership is moved to the linked list. Never free the iterator before removing it from
the linked list. Removing an interator can be done by using @ref llist_iterator_take.

@param list the list the iterator must be added to
@param reference the reference iterator
@param insert the iterator that has to be inserted in the list

@return
    - true: The list element was inserted succesfully before the reference object
    - false: An error has occurred
        - list or insert is a NULL pointer
        - the reference iterator is not part of the list
*/
bool llist_insertBefore(llist_t *list, llist_iterator_t *reference, llist_iterator_t *insert) {
    if (!list || !insert) {
        return false;
    }

    if (reference && reference->list != list) {
        return false;
    }

    if (reference == insert) {
        return true;
    }

    insert = llist_iterator_take(insert);
    insert->list = list;

    if (!reference) {
        reference = list->head;
    }

    llist_iterator_t *before = NULL;
    llist_iterator_t *after = NULL;

    if (reference) {
        before = reference->prev;
        after = reference;
    }

    insert->next = after;
    insert->prev = before;
    if (after) {
        after->prev = insert;
    } else {
        list->tail = insert;
    }
    if (before) {
        before->next = insert;
    } else {
        list->head = insert;
    }

    return true;
}

/**
@ingroup pcb_utils_linked_list
@brief
Insert a list element after the reference element.

@details
This function inserts the specified list iterator after the reference iterator in the list. If the
element was part of another list then the element will be automatically removed from that list.\n
If the reference is a NULL pointer the insert iterator will be added after the last element in
the list, the same behavior can be achieved with @ref llist_append

@warning
The iterator pointer ownership is moved to the linked list. Never free the iterator before removing it from
the linked list. Removing an interator can be done by using @ref llist_iterator_take.

@param list the list the iterator must be added to
@param reference the reference iterator
@param insert the iterator that has to be inserted in the list

@return
    - true: The list element was inserted succesfully after the reference object
    - false: An error has occurred
        - list or insert is a NULL pointer
        - the reference iterator is not part of the list
*/
bool llist_insertAfter(llist_t *list, llist_iterator_t *reference,llist_iterator_t *insert) {
    if (!list || !insert) {
        return false;
    }

    if (reference && reference->list != list) {
        return false;
    }

    if (reference == insert) {
        return true;
    }

    insert = llist_iterator_take(insert);
    insert->list = list;

    if (!reference) {
        reference = list->tail;
    }

    llist_iterator_t *before = NULL;
    llist_iterator_t *after = NULL;

    if (reference) {
        before = reference;
        after = reference->next;
    }

    insert->next = after;
    insert->prev = before;
    if (after) {
        after->prev = insert;
    } else {
        list->tail = insert;
    }
    if (before) {
        before->next = insert;
    } else {
        list->head = insert;
    }

    return true;
}

/**
@ingroup pcb_utils_linked_list
@brief
Append the element to a list.

@details
This function appends the specified iterator to an existing list. If the element was part of
another list then the element will be automatically removed from that list.\n
The iterator will become the last element in the list.

@warning
The iterator pointer ownership is moved to the linked list. Never free the iterator before removing it from
the linked list. Removing an interator can be done by using @ref llist_iterator_take.

@param list the list to which the iterator has to be added
@param it the iterator that has to be attached to the list

@return
    - true: The list element was added succesfully appended to the list
    - false: An error has occurred
        - list or it is a NULL pointer
*/
bool llist_append(llist_t *list, llist_iterator_t *it) {
    if (!list || !it) {
        return false;
    }

    if (it->list) {
        llist_iterator_take(it);
    }

    llist_iterator_t *before = list->tail;
    it->list = list;
    it->next = NULL;
    it->prev = before;

    if (before) {
        before->next = it;
    } else {
        list->head = it;
    }
    list->tail = it;

    return true;
}

/**
@ingroup pcb_utils_linked_list
@brief
Prepend the element to a list.

@details
This function prepends the specified iterator to an existing list. If the element was part of
another list then the element will be automatically removed from that list.\n
The iterator will become the first element in the list.

@warning
The iterator pointer ownership is moved to the linked list. Never free the iterator before removing it from
the linked list. Removing an interator can be done by using @ref llist_iterator_take.

@param list the list to which the element has to be added
@param it the iterator that has to be attached to the list

@return
    - true: The list element was added succesfully prepended to the list
    - false: An error has occurred
        - list or it is a NULL pointer
*/
bool llist_prepend(llist_t *list, llist_iterator_t *it) {
    if (!list || !it) {
        return false;
    }

    if (it->list) {
        llist_iterator_take(it);
    }

    llist_iterator_t *after = list->head;
    it->list = list;
    it->next = after;
    it->prev = NULL;

    if (after) {
        after->prev = it;
    } else {
        list->tail = it;
    }
    list->head = it;

    return true;
}

/**
@ingroup pcb_utils_linked_list
@brief
Attach the element to a list at a certain index.

@details
This function attaches the specified iterator to an existing list at the given index.
If the element is part of another list then the element will be automatically removed from that list.\n
If the index is higher then the number of elements in the list, it will be appended after the last element.\n

@warning
The iterator pointer ownership is moved to the linked list. Never free the iterator before removing it from
the linked list. Removing an interator can be done by using @ref llist_iterator_take.

@param list the list to which the element has to be added
@param index the item # where to add the element to the list (start counting from 0)
@param it the list element that has to be attached to the list

@return
    - true: The list element was succesfully inserted into the list
    - false: An error has occurred
        - list or it is a NULL pointer
*/
bool llist_insertAt(llist_t *list, unsigned int index, llist_iterator_t *it) {
    llist_iterator_t *pos = llist_at(list,index);
    if (!pos) {
        return llist_insertAfter(list,NULL,it);
    }
    return llist_insertBefore(list,pos,it);
}

/**
@ingroup pcb_utils_linked_list
@brief
Detach the element from a list.

@details
This function detaches the specified list element from an existing list.\n

@param it the list element that has to be detached from the list

@return
    - a <b>pointer</b> to the detached list element.
    - NULL if an error occured
        - it is NULL
*/
llist_iterator_t *llist_iterator_take(llist_iterator_t *it) {
    if (!it) {
         return NULL;
    }

    if (!it->list) {
        return it;
    }

    llist_iterator_t *before = it->prev;
    llist_iterator_t *after = it->next;

    if (before) {
        before->next = after;
    } else {
        it->list->head = after;
    }
    if (after) {
        after->prev = before;
    } else {
        it->list->tail = before;
    }

    it->prev = NULL;
    it->next = NULL;
    it->list = NULL;

    return it;
}

/**
@ingroup pcb_utils_linked_list
@brief
Get the size of the list.

@details
This function returns the size of the list.\n
This function will iterate over all elements in the list and count them, this can have performance impact.
If you need to check if a list contains elements or is empty call @ref llist_isEmpty.

@param list the list in wich you are interested

@return
    - the number of elements in the list.
    - 0 if list is a NULL pointer ot the list is empty.
*/
unsigned int llist_size(const llist_t *list) {
    if (!list) {
        return 0;
    }
    unsigned int size = 0;
    llist_iterator_t *it;
    llist_for_each(it,list) {
        size++;
    }
    return size;
}

#ifndef CONFIG_PCB_USE_INLINE_FUNCS
/**
@ingroup pcb_utils_linked_list
@brief
Determine if the list is empty.

@details
This function checks wether the list is empty or not.\n

@param list the list in wich you are interested

@return
    - true:
        - The list is empty
        - list == NULL
    - false: The list is not empty

*/
bool llist_isEmpty(const llist_t *list) {
    return !(list && list->head);
}
#endif

#ifndef CONFIG_PCB_USE_INLINE_FUNCS
/**
@ingroup pcb_utils_linked_list
@brief
Get the first list element.

@details
This function returns the first list iterator of the linked list.\n

@warning
Do not free the returned pointer before removing it from the linked list. Doing so will corrupt the linked list
itself. Removing the iterator from the linked list can be done by using the function @ref llist_iterator_take.

@param list the linked list

@return
    - a <b>pointer</b> to the first list iterator of the linked list.
    - NULL if the list parameter is NULL or the list is empty.
*/

llist_iterator_t *llist_first(const llist_t *list) {
    return list?list->head:NULL;
}
#endif

#ifndef CONFIG_PCB_USE_INLINE_FUNCS
/**
@ingroup pcb_utils_linked_list
@brief
Get the last list element.

@details
This function returns the last list iterator of the linked list.\n

@warning
Do not free the returned pointer before removing it from the linked list. Doing so will corrupt the linked list
itself. Removing the iterator from the linked list can be done by using the function @ref llist_iterator_take.

@param list the linked list

@return
    - a <b>pointer</b> to the last list iterator of the linked list.
    - NULL if the list parameter is NULL or the list is empty.

*/
llist_iterator_t *llist_last(const llist_t *list) {
    return list?list->tail:NULL;
}
#endif

#ifndef CONFIG_PCB_USE_INLINE_FUNCS
/**
@ingroup pcb_utils_linked_list
@brief
Get a pointer to the next list iterator.

@details
This function returns a pointer to the next list iterator.\n

@warning
Do not free the returned pointer before removing it from the linked list. Doing so will corrupt the linked list
itself. Removing the iterator from the linked list can be done by using the function @ref llist_iterator_take.

@param it the reference iterator

@return
    - a <b>pointer</b> to the next list iterator in the linked list.
    - NULL if the reference iterator is the last in the list.
    - NULL if the reference iterator is NULL.
*/
llist_iterator_t *llist_iterator_next(const llist_iterator_t *it) {
    return (it)?it->next:NULL;
}
#endif

#ifndef CONFIG_PCB_USE_INLINE_FUNCS
/**
@ingroup pcb_utils_linked_list
@brief
Get a pointer to the previous list iterator.

@details
This function returns a pointer to the previous list iterator. \n

@warning
Do not free the returned pointer before removing it from the linked list. Doing so will corrupt the linked list
itself. Removing the iterator from the linked list can be done by using the function @ref llist_iterator_take.

@param it the reference iterator

@return
    - a <b>pointer</b> to the previous list iterator in the linked list.
    - NULL if the reference iterator is the last in the list.
    - NULL if the reference iterator is NULL.
*/
llist_iterator_t *llist_iterator_prev(const llist_iterator_t *it) {
    return (it)?it->prev:NULL;
}
#endif

#ifndef CONFIG_PCB_USE_INLINE_FUNCS
/**
@ingroup pcb_utils_linked_list
@brief
Detach the first element from a list.

@details
This function detaches the first list element from an existing list.\n

@param list the list from which the first element has to be detached

@return
    - a <b>pointer</b> to the detached list element.
    - NULL if an error occured
        - list is NULL
        - the list is empty
*/
llist_iterator_t *llist_takeFirst(llist_t *list) {
    return llist_iterator_take(llist_first(list));
}
#endif

#ifndef CONFIG_PCB_USE_INLINE_FUNCS
/**
@ingroup pcb_utils_linked_list
@brief
Detach the last element from a list.

@details
This function detaches the last list element from an existing list.\n

@param list the list from which the last element has to be detached

@return
    - a <b>pointer</b> to the detached list element.
    - NULL if an error occured
        - list is NULL
        - the list is empty
*/
llist_iterator_t *llist_takeLast(llist_t *list) {
    return llist_iterator_take(llist_last(list));
}
#endif
