/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>
#include <ctype.h>
#include <wctype.h>
#include <wchar.h>

#include <debug/sahtrace.h>

#include <pcb/common/error.h>
#include <pcb/utils/variant_list.h>
#include <pcb/utils/variant_map.h>
#include <pcb/utils/string.h>

/**
@file
@brief
Implementation of the variant functions
*/

static uint32_t utf8Length(const char *pos) {
    if ((pos[0] & 0xF0) == 0xF0) {
        return 4;
    }
    if ((pos[0] & 0xE0) == 0xE0) {
        return 3;
    }
    if ((pos[0] & 0xC0) == 0xC0) {
        return 2;
    }

    return 1;
}

static void variant_linefeed(string_t *data, int indent) {
    int i;
    string_appendChar(data, "\n");
    for (i = 0; i < indent; i++) {
        string_appendChar(data, "    ");
    }
}

static const char *variant_print_default_delimiters[] = { "{", ":", ",", "}", "[", ",", "]", NULL };

static bool variant_printPriv(string_t *data, const variant_t *variant, uint32_t attributes,
                              const char **delimiters, int indent);

static void variant_printPrivQuotedString(string_t *data, string_t *temp, const variant_t *variant) {
    string_initialize(temp, string_length(&variant->data.str) * 2);
    string_appendChar(data,"\"");
    /* copy string to temp and escape special characters */
    string_copy(temp,&variant->data.str);
    string_replaceChar(temp,"\\","\\\\");
    /* escape special characters with \uxxxx */
    char *pos = variant->data.str.buffer;
    char needle[2] = {0, 0};
    char escaped[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
    uint32_t flags = 0;
    for(;*pos;) {
        if (*pos == '"' && !(flags & 0x01)) {
            string_replaceChar(temp,"\"","\\\"");
            flags |= 0x01;
            pos++;
            continue;
        }
        if (*pos == '/' && !(flags & 0x02)) {
            string_replaceChar(temp,"/","\\/");
            flags |= 0x02;
            pos++;
            continue;
        }
        if (*pos == '\t' && !(flags & 0x04)) {
            string_replaceChar(temp,"\t","\\t");
            flags |= 0x04;
            pos++;
            continue;
        }
        if (*pos == '\r' && !(flags & 0x08)) {
            string_replaceChar(temp,"\r","\\r");
            flags |= 0x08;
            pos++;
            continue;
        }
        if (*pos == '\n' && !(flags & 0x10)) {
            string_replaceChar(temp,"\n","\\n");
            flags |= 0x10;
            pos++;
            continue;
        }
        if (*pos == '\f' && !(flags & 0x20)) {
            string_replaceChar(temp,"\f","\\f");
            flags |= 0x20;
            pos++;
            continue;
        }
        if (*pos == '\b' && !(flags & 0x40)) {
            string_replaceChar(temp,"\b","\\b");
            flags |= 0x40;
            pos++;
            continue;
        }

        uint32_t length = utf8Length(pos);
        if (length > 1) {
            pos += length;
            continue;
        }
        needle[0] = *pos;
        if (!isprint(*needle)) {
            snprintf(escaped,9,"\\u00%2.2x",(unsigned char)(*needle));
            string_replaceChar(temp,needle,escaped);
        }
        pos++;
    }
    string_append(data,temp);
    string_appendChar(data,"\"");
}

static void variant_printPrivList(string_t *data, const variant_t *variant, uint32_t attributes,
                                  const char **delimiters, int indent) {
    if ((attributes & variant_print_multiline) && (attributes & variant_print_need_newline) &&
            !variant_list_isEmpty(variant->data.vl)) {
        variant_linefeed(data, indent);
    }
    attributes &= ~variant_print_need_newline;
    string_appendChar(data, delimiters[variant_print_delimiter_list_open]);
    if ((attributes & variant_print_spaces) && !(attributes & variant_print_multiline)) {
        string_appendChar(data, " ");
    }
    indent++;
    variant_list_iterator_t *it = NULL;
    variant_list_for_each(it,variant->data.vl) {
        if (attributes & variant_print_multiline) {
            variant_linefeed(data, indent);
        }
        if (!variant_printPriv(data,&it->variant,attributes,delimiters,indent)) {
            string_appendChar(data,"\"#ERROR#\"");
        }
        if (variant_list_iterator_next(it)) {
            string_appendChar(data, delimiters[variant_print_delimiter_list_iterate]);
            if ((attributes & variant_print_spaces) && !(attributes & variant_print_multiline)) {
                string_appendChar(data, " ");
            }
        }
    }
    indent--;
    if ((attributes & variant_print_multiline) && !variant_list_isEmpty(variant->data.vl)) {
        variant_linefeed(data, indent);
    }
    if ((attributes & variant_print_spaces) && !(attributes & variant_print_multiline)) {
        string_appendChar(data, " ");
    }
    string_appendChar(data, delimiters[variant_print_delimiter_list_close]);
}

static void variant_printPrivMap(string_t *data, const variant_t *variant, uint32_t attributes,
                                  const char **delimiters, int indent) {
    if ((attributes & variant_print_multiline) && (attributes & variant_print_need_newline) &&
            !variant_map_isEmpty(variant->data.vm)) {
        variant_linefeed(data, indent);
    }
    attributes |= variant_print_need_newline;
    string_appendChar(data, delimiters[variant_print_delimiter_map_open]);
    if ((attributes & variant_print_spaces) && !(attributes & variant_print_multiline)) {
        string_appendChar(data, " ");
    }
    indent++;
    variant_map_iterator_t *it = NULL;
    int i,n=0;
    if (attributes & variant_print_outline_colons) {
        variant_map_for_each(it,variant->data.vm) {
            i = strlen(it->key?it->key:"");
            if (i > n) {
                n = i;
            }
        }
    }
    variant_map_for_each(it,variant->data.vm) {
        if (attributes & variant_print_multiline) {
            variant_linefeed(data, indent);
        }
        if (attributes & variant_print_quote_keys) {
            string_appendChar(data,"\"");
        }
        string_appendChar(data,it->key);
        if (attributes & variant_print_quote_keys) {
            string_appendChar(data,"\"");
        }
        if (attributes & variant_print_outline_colons) {
            for (i = n - strlen(it->key?it->key:""); i; i--) {
                string_appendChar(data, " ");
            }
        }
        if (attributes & variant_print_spaces) {
            string_appendChar(data, " ");
        }
        string_appendChar(data, delimiters[variant_print_delimiter_map_assign]);
        if (attributes & variant_print_spaces) {
            string_appendChar(data, " ");
        }
        if (!variant_printPriv(data,&it->variant,attributes,delimiters,indent)) {
            string_appendChar(data,"\"#ERROR#\"");
        }
        if (variant_map_iterator_next(it)) {
            string_appendChar(data, delimiters[variant_print_delimiter_map_iterate]);
            if ((attributes & variant_print_spaces) && !(attributes & variant_print_multiline)) {
                string_appendChar(data, " ");
            }
        }
    }
    indent--;
    if ((attributes & variant_print_multiline) && !variant_map_isEmpty(variant->data.vm)) {
        variant_linefeed(data, indent);
    }
    if ((attributes & variant_print_spaces) && !(attributes & variant_print_multiline)) {
        string_appendChar(data, " ");
    }
    string_appendChar(data, delimiters[variant_print_delimiter_map_close]);
}

static bool variant_printPriv(string_t *data, const variant_t *variant, uint32_t attributes,
                              const char **delimiters, int indent) {
    if (!variant || !data) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    string_t temp;
    string_initialize(&temp,0);
    switch(variant->type) {
        case variant_type_unknown:
            if (attributes & variant_print_quote_strings) {
                string_appendChar(data,"null");
            } else {
                return false;
            }
        break;
        case variant_type_string:
            if (attributes & variant_print_quote_strings) {
                if (string_isNull(&variant->data.str)) {
                    string_appendChar(data,"null");
                } else {
                    variant_printPrivQuotedString(data, &temp, variant);
                }
            } else {
                string_append(data,&variant->data.str);
            }
        break;
        case variant_type_int8:
            string_fromInt8(&temp,variant->data.i8);
            string_append(data,&temp);
        break;
        case variant_type_int16:
            string_fromInt16(&temp,variant->data.i16);
            string_append(data,&temp);
        break;
        case variant_type_int32:
            string_fromInt32(&temp,variant->data.i32);
            string_append(data,&temp);
        break;
        case variant_type_int64:
            string_fromInt64(&temp,variant->data.i64);
            string_append(data,&temp);
        break;
        case variant_type_uint8:
            string_fromUInt8(&temp,variant->data.i8);
            string_append(data,&temp);
        break;
        case variant_type_uint16:
            string_fromUInt16(&temp,variant->data.i16);
            string_append(data,&temp);
        break;
        case variant_type_uint32:
            string_fromUInt32(&temp,variant->data.i32);
            string_append(data,&temp);
        break;
        case variant_type_uint64:
            string_fromUInt64(&temp,variant->data.i64);
            string_append(data,&temp);
        break;
        case variant_type_bool:
            if (attributes & variant_print_bool_to_num) {
                if (variant->data.b) {
                    string_appendChar(data,"1");
                } else {
                    string_appendChar(data,"0");
                }
            } else {
                if (variant->data.b) {
                    string_appendChar(data,"true");
                } else {
                    string_appendChar(data,"false");
                }
            }
        break;
        case variant_type_double:
            string_fromDouble(&temp,variant->data.d);
            string_append(data,&temp);
        break;
        case variant_type_date_time: {
            struct tm tmValue;
            if (variant_toDateTime(&tmValue, variant)) {
                if (attributes & variant_print_quote_strings) {
                    string_appendFormat(data,"\"%04i-%02i-%02iT%02i:%02i:%02i\"",
                                            tmValue.tm_year + 1900, tmValue.tm_mon + 1,
                                            tmValue.tm_mday, tmValue.tm_hour, tmValue.tm_min, tmValue.tm_sec);
                } else {
                     string_fromTime(data,variant->data.dt,NULL);
                }
            } else {
                string_cleanup(&temp);
                return false;
            }
        }
        break;
        case variant_type_array_reference:
        case variant_type_array:
            variant_printPrivList(data, variant, attributes, delimiters, indent);
        break;
        case variant_type_map_reference:
        case variant_type_map:
            variant_printPrivMap(data, variant, attributes, delimiters, indent);
        break;
        case variant_type_reference:
            variant_printPriv(data, variant->data.v, attributes, delimiters, indent);
        break;
        case variant_type_file_descriptor:
            string_fromInt32(&temp,variant->data.fd);
            string_append(data,&temp);
        break;
        case variant_type_byte_array:
            pcb_error = pcb_error_invalid_variant_conversion;
            return false;
        break;
        default:
        break;
    }

    string_cleanup(&temp);
    return true;
}

/**
@ingroup pcb_utils_variant
@brief
Create a string variant using the specified character array as initialization data.

@details
This function cleans up the existing variant value, (re-)initializes the variant to the string type,
and initializes the string value with the provided char array.\n

@param variant the variant we want to set
@param data the initial value

@return
    - true: the variant_t is ready to be used
    - false: An error has occurred, pcb_error contains the reason why the set failed
        - an invalid variant was provided
        - an invalid character array was provided
*/
bool variant_setChar(variant_t *variant, const char *data) {
    if (!variant) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }
    variant_cleanup(variant);

    if (variant_initialize(variant,variant_type_string)) {
        string_fromChar(&variant->data.str,data);
        return true;
    }

    return false;
}

/**
@ingroup pcb_utils_variant
@brief
Create a string variant using the specified string as initialization data.

@details
This function cleans up the existing variant value, (re-)initializes the variant to the string type,
and initializes the string value with the provided data string.\n

@param variant the variant we want to set
@param data the initial value

@return
    - true: the variant_t is ready to be used
    - false: An error has occurred, pcb_error contains the reason why the set failed
        - an invalid variant was provided
        - an invalid character array was provided
*/
bool variant_setString(variant_t *variant, const string_t *data) {
    if (!variant) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }
    variant_cleanup(variant);

    if (variant_initialize(variant,variant_type_string)) {
        if (data) {
            string_copy(&variant->data.str,data);
        }
        return true;
    }

    return false;
}

/**
@ingroup pcb_utils_variant
@brief
Create a variant containing a string list

@details
This function cleans up the existing variant value, (re-)initializes the variant to the string type
and initializes the value with the provided string list joined using the provided separator.\n

@param variant the variant we want to set
@param list pointer to the string list
@param separator the separator use to join the string list

@return
    - true: the variant_t is ready to be used
    - false: An error has occurred, pcb_error contains the reason why the set failed
*/
bool variant_setStringList(variant_t *variant, const string_list_t *list, const char *separator) {
    if (!variant || !list || !separator) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    string_t data;
    string_initialize(&data, 256);
    string_join(&data, list, separator);
    bool rc = variant_setString(variant, &data);
    string_cleanup(&data);
    return rc;
}

/**
@ingroup pcb_utils_variant
@brief
Get the value of the specified variant and copy it into the char array.

@details
This function converts the variant value to a string and copies it into the a newly allocated
char array. The conversion type depends on the variant type. \n
The char pointer must be freed.
All variant types can be converted to a string_t.\n

@param data pointer to a char array
@param variant the variant of interest

@return
    - true: the resulting string_t is ready to be used
    - false: An error has occurred, pcb_error contains the reason why the set failed
        - an invalid variant was provided
        - an invalid/un-initialized string_t was provided
        - memory allocation failed
*/
bool variant_toChar(char **data, const variant_t *variant) {
    if (!variant || !data) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    string_t string;
    string_initialize(&string,0);
    if (!variant_toString(&string,variant)) {
        string_cleanup(&string);
        return false;
    }

    if (string.buffer) {
        *data = string.buffer;
    } else {
        *data = strdup("");
    }
    return true;
}

/**
@ingroup pcb_utils_variant
@brief
Get the value of the specified variant and copy it into the specified string.

@details
This function converts the variant value to the string_t type and copies it into the provided
string_t. The conversion type depends on the variant type. \n

The string_t data must be initialized before calling this function. The user is responsible for
cleaning up the resulting string.\n
All variant types can be converted to a string_t.\n

@param data the initialized sting_t that will contain the result
@param variant the variant of interest

@return
    - true: the resulting string_t is ready to be used
    - false: An error has occurred, pcb_error contains the reason why the set failed
        - an invalid variant was provided
        - an invalid/un-initialized string_t was provided
        - memory allocation failed
*/
bool variant_toString(string_t *data, const variant_t *variant) {
    return variant_print(data, variant, variant_print_bool_to_num, NULL);
}


/**
@ingroup pcb_utils_variant
@brief
Get the value of the specified variant and copy it into the specified string as a JSON string.

@details
This function converts the variant value to the string_t type and copies it into the provided
string_t. The conversion type depends on the variant type. \n

The string_t data must be initialized before calling this function. The user is responsible for
cleaning up the resulting string.\n
All variant types can be converted to a string_t.\n

@param data the initialized sting_t that will contain the result
@param variant the variant of interest

@return
    - true: the resulting string_t is ready to be used
    - false: An error has occurred, pcb_error contains the reason why the set failed
        - an invalid variant was provided
        - an invalid/un-initialized string_t was provided
        - memory allocation failed
*/
bool variant_toJSON(string_t *data, const variant_t *variant) {
    return variant_print(data, variant, variant_print_json, NULL);
}

/**
@ingroup pcb_utils_variant
@brief
Customized conversion of a variant to a string.

@details
This function converts the variant value to the string_t type and copies it into the provided
string_t. The format of the resulting string is customizable through the function arguments
attributes and delimiters.\n

The string_t data must be initialized before calling this function. The user is responsible for
cleaning up the resulting string.\n
All variant types can be converted to a string_t.\n

@param data the initialized sting_t that will contain the result
@param variant the variant of interest
@param attributes Attributes that affect the format of the resulting string. A bitwise OR of
                  zero or more of the following:
                  - @ref variant_print_multiline
                  - @ref variant_print_spaces
                  - @ref variant_print_outline_colons
                  - @ref variant_print_quote_keys
                  - @ref variant_print_quote_strings
                  - @ref variant_print_bool_to_num
                  - @ref variant_print_need_newline
                  - @ref variant_print_json
                  - @ref variant_print_beautify
@param delimiters Array of strings that specify the delimiters to be used for maps and lists.
                  The exact order in which delimiter types should appear in the array,
                  is defined by the enumerator @ref variant_print_delimiter_t.\n
                  If a NULL pointer is provided, default delimiters are used.\n
                  Remark: if this argument is provided, it MUST be a NULL-terminated array.
@return
    - true: the resulting string_t is ready to be used
    - false: An error has occurred, pcb_error contains the reason why the set failed
        - an invalid variant was provided
        - an invalid/un-initialized string_t was provided
        - memory allocation failed
*/
bool variant_print(string_t *data, const variant_t *variant, uint32_t attributes, const char **delimiters) {
    string_clear(data);
    if (delimiters) {
        const char **delimiter = delimiters;
        int i;
        for (i = 0; i < variant_print_delimiter_count && *delimiter; i++) {
            delimiter++;
        }
        if (i < variant_print_delimiter_count) {
            pcb_error = pcb_error_invalid_value;
            return false;
        }
    } else {
        delimiters = variant_print_default_delimiters;
    }
    return variant_printPriv(data, variant, attributes, delimiters, 0);
}

/**
@ingroup pcb_utils_variant
@brief
Get a string pointer directly to the string data of a string variant.

@details
This function returns a string pointer directly to the string data contained within the variant.\n
The resulting string may only be used to READ, NOT WRITE. This function only works when
the variant type is a string.\n
The returned pointer MUST not be freed.\n

@param variant the variant of interest

@return
The resulting string pointer, NULL if the variant is invalid or not of the string type.
*/
const string_t *variant_da_string(const variant_t *variant) {
    if (!variant) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }

    if (variant->type == variant_type_reference) {
        variant = variant->data.v;
    }

    if (variant->type != variant_type_string) {
        pcb_error = pcb_error_invalid_variant_type;
        return NULL;
    }

    return &variant->data.str;
}

/**
@ingroup pcb_utils_variant
@brief
Converts a variant to a string list.

@details
This function converts the variant value to a string list.\n
The conversion depends on the type of the variant given.\n
Simple types are added to the string list.\n
String is split using the separator.\n
If the given variant is a variant list or a variant map, all elements of the given
list or map are added the to the string list.

The string list provided must be initialized.\n

@param data the initialized string list that will contain the result
@param variant the variant of interest
@param separator the separator to use if the variant type is a string

@return
    - true: the resulting struct tm is ready to be used
    - false: An error has occurred, pcb_error contains the reason why the set failed
        - an invalid variant was provided
        - an invalid variant list was provided
        - memory allocation failed
*/
bool variant_toStringList(string_list_t *data, const variant_t *variant, const char *separator) {
    if (!variant || !data) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    string_list_clear(data);

    string_t str;
    string_initialize(&str, 64);

    switch(variant->type) {
        case variant_type_string:
          variant_toString(&str, variant);
          string_list_split(data, &str, separator, strlist_skip_empty_parts, string_case_sensitive);
        break;
        case variant_type_int8:
        case variant_type_int16:
        case variant_type_int32:
        case variant_type_int64:
        case variant_type_uint8:
        case variant_type_uint16:
        case variant_type_uint32:
        case variant_type_uint64:
        case variant_type_bool:
        case variant_type_double:
        case variant_type_date_time:
        case variant_type_file_descriptor:
        case variant_type_byte_array:
            variant_toString(&str, variant);
            string_list_append(data, string_list_iterator_create(&str));
        break;
        case variant_type_array_reference:
        case variant_type_array: {
            variant_list_iterator_t *it = NULL;
            variant_list_for_each(it,variant->data.vl) {
                variant_toString(&str, &it->variant);
                string_list_append(data, string_list_iterator_create(&str));
            }
        }
        break;
        case variant_type_map_reference:
        case variant_type_map: {
            variant_map_iterator_t *it = NULL;
            variant_map_for_each(it,variant->data.vm) {
                variant_toString(&str, &it->variant);
                string_list_append(data, string_list_iterator_create(&str));
            }
        }
        break;
        case variant_type_reference:
            return variant_toStringList(data, variant->data.v, separator);
        break;
        default:
        break;
    }

    string_cleanup(&str);

    return true;
}

/**
@ingroup pcb_utils_variant
@brief
Converts a variant to a string.

@details
This function is provided for convience and is doing the same as @ref variant_toString.
The returned pointer must be freed.\n

@param variant the variant of interest

@return
    - The converted value
*/
string_t *variant_string(const variant_t *variant) {
    if (!variant) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }

    string_t *data = (string_t *)calloc(1,sizeof(string_t));
    if (!data) {
        return NULL;
    }

    if (!variant_toString(data,variant)) {
        string_cleanup(data);
        free(data);
        return NULL;
    }

    return data;
}

/**
@ingroup pcb_utils_variant
@brief
Converts a variant to a string list.

@details
This function is provided for convience and is doing the same as @ref variant_toStringList.
The returned pointer must be freed.\n

@param variant the variant of interest
@param separator separator string to insert in the string

@return
    - The converted value
*/
string_list_t *variant_stringList(const variant_t *variant, const char *separator) {
    if (!variant) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }

    string_list_t *data = (string_list_t *)calloc(1,sizeof(string_list_t));
    if (!data) {
        return NULL;
    }

    if (!string_list_initialize(data)) {
        free(data);
        return NULL;
    }

    if (!variant_toStringList(data, variant, separator)) {
        string_list_cleanup(data);
        free(data);
        return NULL;
    }

    return data;
}
