/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sahtest/sahtest.h>
#include <pcb/core/datamodel.h>
#include "datamodel_priv.h"
#include "mock.h"

#define TEST_FORMAT         10
#define TEST_FORMAT_UNKNOWN 11
#define TEST_FORMAT_EMPTY   12

mock_struct_declare(test_load,50);
mock_struct_declare(test_verify,50);
mock_struct_declare(test_save,50);

bool test_load(int fd, object_t *object, const char *filename) {
    assert_bool(mock_get_expectation(test_load)->fd == fd);
    assert_bool(mock_get_expectation(test_load)->object == object);
    assert_string_equals(mock_get_expectation(test_load)->filename,filename);

    return mock_get_expectation(test_load)->retval;
}

bool test_verify(int socketfd, const char *filename, variant_map_t *info, bool *hasDefinition) {
    assert_bool(mock_get_expectation(test_verify)->fd == socketfd);
    assert_string_equals(mock_get_expectation(test_verify)->filename,filename);
    assert_bool(mock_get_expectation(test_verify)->info == info);
    assert_bool(mock_get_expectation(test_verify)->hasDefinition == hasDefinition);

    return mock_get_expectation(test_verify)->retval;
}

bool test_save(int fd, object_t *object, uint32_t depth, const char *filename) {
    assert_bool(mock_get_expectation(test_save)->fd == fd);
    assert_bool(mock_get_expectation(test_save)->object == object);
    assert_bool(mock_get_expectation(test_save)->depth == depth);
    assert_string_equals(mock_get_expectation(test_save)->filename,filename);

    return mock_get_expectation(test_save)->retval;
}

static datamodel_t dm;

void dm_load_save_setup() {
    mock_reset(getenv);
    mock_reset(open);
    mock_reset(rename);
    mock_reset(dirname);
    mock_reset(fsync);
    mock_reset(close);
    mock_reset(test_load);
    mock_reset(test_verify);
    mock_reset(test_save);

    datamodel_initialize(&dm);
    serialization_initialize();

    serialize_handlers_t *handlers = (serialize_handlers_t *)calloc(1,sizeof(serialize_handlers_t));
    if (!handlers)
        return;

    handlers->load = test_load;
    handlers->verify = test_verify;
    handlers->save = test_save;
    serialization_addSerializer(handlers,SERIALIZE_FORMAT(TEST_FORMAT,0,0));

    handlers = (serialize_handlers_t *)calloc(1,sizeof(serialize_handlers_t));
    if (!handlers)
        return;

    handlers->load = test_load;
    handlers->verify = test_verify;
    handlers->save = test_save;
    serialization_addSerializer(handlers,SERIALIZE_FORMAT(serialize_format_odl,0,0));

    handlers = (serialize_handlers_t *)calloc(1,sizeof(serialize_handlers_t));
    if (!handlers)
        return;
    serialization_addSerializer(handlers,SERIALIZE_FORMAT(TEST_FORMAT_EMPTY,0,0));
}

void dm_load_save_teardown() {
    serialization_unregister();
    datamodel_cleanup(&dm);
}

int dm_load_null() {
    assert_bool(datamodel_load(NULL,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load_unknown_format() {
    assert_bool(datamodel_load(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT_UNKNOWN,0,0)) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load_no_handler() {
    assert_bool(datamodel_load(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT_EMPTY,0,0)) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;
    mock_add_expectation(getenv)->varname = "PCB_AUTO_RESOLVE";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_load)->fd = 15;
    mock_add_expectation(test_load)->object = datamodel_root(&dm);
    mock_add_expectation(test_load)->filename = "test_load.txt";
    mock_add_expectation(test_load)->retval = true;

    assert_bool(datamodel_load(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == true);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load_prefix_path() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = "/usr/lib/test/";
    mock_add_expectation(getenv)->varname = "PCB_AUTO_RESOLVE";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "/usr/lib/test/test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_load)->fd = 15;
    mock_add_expectation(test_load)->object = datamodel_root(&dm);
    mock_add_expectation(test_load)->filename = "/usr/lib/test/test_load.txt";
    mock_add_expectation(test_load)->retval = true;

    assert_bool(datamodel_load(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == true);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load_auto_resolve_yes() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;
    mock_add_expectation(getenv)->varname = "PCB_AUTO_RESOLVE";
    mock_add_expectation(getenv)->value = "yes";

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_load)->fd = 15;
    mock_add_expectation(test_load)->object = datamodel_root(&dm);
    mock_add_expectation(test_load)->filename = "test_load.txt";
    mock_add_expectation(test_load)->retval = true;

    assert_bool(datamodel_load(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == true);
    assert_bool(datamodel_isAutoResolving(&dm) == true);

    assert_bool(datamodel_autoResolve(NULL,true) == false);
    assert_bool(datamodel_isAutoResolving(NULL) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load_auto_resolve_no() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;
    mock_add_expectation(getenv)->varname = "PCB_AUTO_RESOLVE";
    mock_add_expectation(getenv)->value = "no";

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_load)->fd = 15;
    mock_add_expectation(test_load)->object = datamodel_root(&dm);
    mock_add_expectation(test_load)->filename = "test_load.txt";
    mock_add_expectation(test_load)->retval = true;

    assert_bool(datamodel_load(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == true);
    assert_bool(datamodel_isAutoResolving(&dm) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load_auto_resolve_invalid_value() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;
    mock_add_expectation(getenv)->varname = "PCB_AUTO_RESOLVE";
    mock_add_expectation(getenv)->value = "foo";

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_load)->fd = 15;
    mock_add_expectation(test_load)->object = datamodel_root(&dm);
    mock_add_expectation(test_load)->filename = "test_load.txt";
    mock_add_expectation(test_load)->retval = true;

    assert_bool(datamodel_load(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == true);
    assert_bool(datamodel_isAutoResolving(&dm) == true);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load_no_auto_resolve_set_resolve_to_false() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;
    mock_add_expectation(getenv)->varname = "PCB_AUTO_RESOLVE";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_load)->fd = 15;
    mock_add_expectation(test_load)->object = datamodel_root(&dm);
    mock_add_expectation(test_load)->filename = "test_load.txt";
    mock_add_expectation(test_load)->retval = true;

    assert_bool(datamodel_autoResolve(&dm,false));
    assert_bool(datamodel_load(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == true);
    assert_bool(datamodel_isAutoResolving(&dm) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load_open_failing() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;
    mock_add_expectation(getenv)->varname = "PCB_AUTO_RESOLVE";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = -1;

    assert_bool(datamodel_load(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load_failing() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;
    mock_add_expectation(getenv)->varname = "PCB_AUTO_RESOLVE";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_load)->fd = 15;
    mock_add_expectation(test_load)->object = datamodel_root(&dm);
    mock_add_expectation(test_load)->filename = "test_load.txt";
    mock_add_expectation(test_load)->retval = false;

    assert_bool(datamodel_load(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_verify_null() {
    bool hasDefinition = false;
    variant_map_t info;
    variant_map_initialize(&info);

    assert_bool(datamodel_verify(NULL,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0),&info,&hasDefinition) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_verify));

    variant_map_cleanup(&info);
    sahtest_success();
}

int dm_verify_unknown_format() {
    bool hasDefinition = false;
    variant_map_t info;
    variant_map_initialize(&info);

    assert_bool(datamodel_verify(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT_UNKNOWN,0,0),&info,&hasDefinition) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_verify));

    variant_map_cleanup(&info);
    sahtest_success();
}

int dm_verify_no_handler() {
    bool hasDefinition = false;
    variant_map_t info;
    variant_map_initialize(&info);


    assert_bool(datamodel_verify(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT_EMPTY,0,0),&info,&hasDefinition) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_verify));

    variant_map_cleanup(&info);
    sahtest_success();
}

int dm_verify() {
    bool hasDefinition = false;
    variant_map_t info;
    variant_map_initialize(&info);

    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_verify)->fd = 15;
    mock_add_expectation(test_verify)->filename = "test_load.txt";
    mock_add_expectation(test_verify)->info = &info;
    mock_add_expectation(test_verify)->hasDefinition = &hasDefinition;
    mock_add_expectation(test_verify)->retval = true;

    assert_bool(datamodel_verify(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0),&info,&hasDefinition) == true);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_verify));

    variant_map_cleanup(&info);
    sahtest_success();
}

int dm_verify_prefix_path() {
    bool hasDefinition = false;
    variant_map_t info;
    variant_map_initialize(&info);

    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = "/usr/lib/test/";

    mock_add_expectation(open)->pathname = "/usr/lib/test/test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_verify)->fd = 15;
    mock_add_expectation(test_verify)->filename = "/usr/lib/test/test_load.txt";
    mock_add_expectation(test_verify)->info = &info;
    mock_add_expectation(test_verify)->hasDefinition = &hasDefinition;
    mock_add_expectation(test_verify)->retval = true;

    assert_bool(datamodel_verify(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0),&info,&hasDefinition) == true);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_verify));

    variant_map_cleanup(&info);
    sahtest_success();
}

int dm_verify_open_failing() {
    bool hasDefinition = false;
    variant_map_t info;
    variant_map_initialize(&info);


    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = -1;

    assert_bool(datamodel_verify(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0),&info,&hasDefinition) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_verify));

    variant_map_cleanup(&info);
    sahtest_success();
}

int dm_verify_failing() {
    bool hasDefinition = false;
    variant_map_t info;
    variant_map_initialize(&info);


    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_verify)->fd = 15;
    mock_add_expectation(test_verify)->filename = "test_load.txt";
    mock_add_expectation(test_verify)->info = &info;
    mock_add_expectation(test_verify)->hasDefinition = &hasDefinition;
    mock_add_expectation(test_verify)->retval = false;

    assert_bool(datamodel_verify(&dm,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0),&info,&hasDefinition) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_verify));

    variant_map_cleanup(&info);
    sahtest_success();
}

int dm_save_null() {
    assert_bool(datamodel_save(NULL,UINT32_MAX,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_save));

    sahtest_success();
}

int dm_save_unknown_format() {
    assert_bool(datamodel_save(datamodel_root(&dm),UINT32_MAX,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT_UNKNOWN,0,0)) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_save));

    sahtest_success();
}

int dm_save_no_handler() {
    assert_bool(datamodel_save(datamodel_root(&dm),UINT32_MAX,"test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT_EMPTY,0,0)) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_save));

    sahtest_success();
}

int dm_save() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "/usr/lib/test/test_load.txt.temp";
    mock_add_expectation(open)->flags = O_WRONLY | O_CREAT | O_TRUNC;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_save)->fd = 15;
    mock_add_expectation(test_save)->object = datamodel_root(&dm);
    mock_add_expectation(test_save)->depth = UINT32_MAX;
    mock_add_expectation(test_save)->filename = "/usr/lib/test/test_load.txt.temp";
    mock_add_expectation(test_save)->retval = true;

    mock_add_expectation(rename)->oldpath = "/usr/lib/test/test_load.txt.temp";
    mock_add_expectation(rename)->newpath = "/usr/lib/test/test_load.txt";
    mock_add_expectation(rename)->retval = 0;

    mock_add_expectation(dirname)->path = "/usr/lib/test/test_load.txt.temp";
    mock_add_expectation(dirname)->dirname = "/usr/lib/test";

    mock_add_expectation(open)->pathname = "/usr/lib/test";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 16;

    mock_add_expectation(fsync)->fd = 16;
    mock_add_expectation(fsync)->retval = 0;

    mock_add_expectation(close)->fd = 16;
    mock_add_expectation(close)->retval = 0;

    assert_bool(datamodel_save(datamodel_root(&dm),UINT32_MAX,"/usr/lib/test/test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == true);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(rename));
    assert_bool(mock_verify_expectations(dirname));
    assert_bool(mock_verify_expectations(fsync));
    assert_bool(mock_verify_expectations(close));
    assert_bool(mock_verify_expectations(test_save));

    sahtest_success();
}

int dm_save_open_failing() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "/usr/lib/test/test_load.txt.temp";
    mock_add_expectation(open)->flags = O_WRONLY | O_CREAT | O_TRUNC;
    mock_add_expectation(open)->fd = -1;

    assert_bool(datamodel_save(datamodel_root(&dm),UINT32_MAX,"/usr/lib/test/test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(rename));
    assert_bool(mock_verify_expectations(dirname));
    assert_bool(mock_verify_expectations(fsync));
    assert_bool(mock_verify_expectations(close));
    assert_bool(mock_verify_expectations(test_save));

    sahtest_success();
}

int dm_save_failing() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "/usr/lib/test/test_load.txt.temp";
    mock_add_expectation(open)->flags = O_WRONLY | O_CREAT | O_TRUNC;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_save)->fd = 15;
    mock_add_expectation(test_save)->object = datamodel_root(&dm);
    mock_add_expectation(test_save)->depth = UINT32_MAX;
    mock_add_expectation(test_save)->filename = "/usr/lib/test/test_load.txt.temp";
    mock_add_expectation(test_save)->retval = false;

    assert_bool(datamodel_save(datamodel_root(&dm),UINT32_MAX,"/usr/lib/test/test_load.txt",SERIALIZE_FORMAT(TEST_FORMAT,0,0)) == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(rename));
    assert_bool(mock_verify_expectations(dirname));
    assert_bool(mock_verify_expectations(fsync));
    assert_bool(mock_verify_expectations(close));
    assert_bool(mock_verify_expectations(test_save));

    sahtest_success();
}

int dm_load_definition_null() {
    assert_bool(datamodel_loadDefinition(NULL,"test_load.txt") == false);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}

int dm_load_definition() {
    mock_add_expectation(getenv)->varname = "PCB_PREFIX_PATH";
    mock_add_expectation(getenv)->value = NULL;
    mock_add_expectation(getenv)->varname = "PCB_AUTO_RESOLVE";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "test_load.txt";
    mock_add_expectation(open)->flags = O_RDONLY;
    mock_add_expectation(open)->fd = 15;

    mock_add_expectation(test_load)->fd = 15;
    mock_add_expectation(test_load)->object = datamodel_root(&dm);
    mock_add_expectation(test_load)->filename = "test_load.txt";
    mock_add_expectation(test_load)->retval = true;

    assert_bool(datamodel_loadDefinition(&dm,"test_load.txt") == true);

    assert_bool(mock_verify_expectations(getenv));
    assert_bool(mock_verify_expectations(open));
    assert_bool(mock_verify_expectations(test_load));

    sahtest_success();
}
