/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <sahtest/sahtest.h>
#include <pcb/core/datamodel.h>
#include <pcb/core/object.h>
#include <pcb/utils.h>
#include "datamodel_priv.h"
#include "object_priv.h"

static datamodel_t dm;

void obj_create_setup() {
    datamodel_initialize(&dm);
}

void obj_create_teardown() {
    datamodel_cleanup(&dm);
}

int obj_create() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);

    object_t *obj = object_create(root,"TestRoot1",0);
    assert_not_null(obj);
    assert_bool(obj->state == object_state_created);
    assert_bool(obj->attributes == object_attr_default);
    assert_string_equals(obj->oname,"TestRoot1");
    assert_bool(llist_isEmpty(&obj->parameters));
    assert_bool(llist_isEmpty(&obj->functions));
    assert_bool(llist_isEmpty(&obj->notifyRequests));
    assert_bool(llist_isEmpty(&obj->mapRequests));
    assert_bool(llist_isEmpty(&obj->ACL));
    assert_bool(llist_isEmpty(&obj->destinations));
    assert_bool(object_parent(obj) == root);

    assert_bool(object_commit(obj));
    assert_bool(obj->state == object_state_ready);

    obj = object_create(root,"TestRoot2",object_attr_read_only);
    assert_not_null(obj);
    assert_bool(obj->attributes == object_attr_read_only);

    obj = object_create(root,"TestRoot3",object_attr_template);
    assert_not_null(obj);
    assert_bool(obj->attributes == object_attr_template);

    obj = object_create(root,"TestRoot4",object_attr_accept_parameters);
    assert_not_null(obj);
    assert_bool(obj->attributes == object_attr_accept_parameters);

    obj = object_create(root,"TestRoot5",object_attr_persistent);
    assert_not_null(obj);
    assert_bool(obj->attributes == object_attr_persistent);

    sahtest_success();
}

int obj_create_null_parent() {
    object_t *obj = object_create(NULL,"TestRoot1",0);
    assert_null(obj);

    sahtest_success();
}

int obj_create_null_name_empty_name() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);

    object_t *obj = object_create(root,"",0);
    assert_null(obj);

    obj = object_create(root,NULL,0);
    assert_null(obj);

    sahtest_success();
}

int obj_create_invalid_name() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);

    object_t *obj = object_create(root,"/a/b.c",0);
    assert_null(obj);

    sahtest_success();
}

int obj_create_two_object_same_name_same_parent() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);

    object_t *obj = object_create(root,"TestRoot1",0);
    assert_not_null(obj);

    obj = object_create(root,"TestRoot1",0);
    assert_null(obj);

    sahtest_success();
}

void obj_create_instance_setup() {
    datamodel_initialize(&dm);
    object_t *root = datamodel_root(&dm);
    object_t *obj = object_create(root,"TestRoot1",object_attr_template);
    obj = object_create(obj,"Child",0);
    obj = object_create(root,"TestRoot2",0);
    object_commit(root);
}

void obj_create_instance_teardown() {
    datamodel_cleanup(&dm);
}

int obj_create_instance() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);
    object_t *templ = object_getObject(root,"TestRoot1",0,NULL);
    assert_not_null(templ);

    object_t *inst = object_createInstance(templ,0,NULL);
    assert_not_null(inst);
    assert_bool(inst->state == object_state_created);
    assert_bool(inst->attributes & object_attr_instance);
    assert_string_equals(inst->oname,"1");
    assert_null(string_buffer(&inst->object_info.instance_info.key));
    assert_bool(llist_isEmpty(&inst->parameters));
    assert_bool(llist_isEmpty(&inst->functions));
    assert_bool(llist_isEmpty(&inst->notifyRequests));
    assert_bool(llist_isEmpty(&inst->mapRequests));
    assert_bool(llist_isEmpty(&inst->ACL));
    assert_bool(llist_isEmpty(&inst->destinations));
    assert_bool(object_parent(inst) == templ);
    assert_bool(object_hasChildren(inst));

    assert_bool(object_commit(inst));
    assert_bool(inst->state == object_state_ready);

    inst = object_createInstance(templ,5,NULL);
    assert_not_null(inst);
    assert_string_equals(inst->oname,"5");
    assert_null(string_buffer(&inst->object_info.instance_info.key));

    inst = object_createInstance(templ,0,"Key");
    assert_not_null(inst);
    assert_string_equals(inst->oname,"6");
    assert_string_equals(string_buffer(&inst->object_info.instance_info.key),"Key");

    sahtest_success();
}

int obj_create_instance_null() {
    object_t *inst = object_createInstance(NULL,0,NULL);
    assert_null(inst);

    sahtest_success();
}

int obj_create_instance_not_template() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);
    object_t *templ = object_getObject(root,"TestRoot2",0,NULL);
    assert_not_null(templ);
    object_t *inst = object_createInstance(templ,0,NULL);
    assert_null(inst);

    sahtest_success();
}

int obj_create_instance_deleted_template() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);
    object_t *templ = object_getObject(root,"TestRoot1",0,NULL);
    assert_not_null(templ);

    assert_bool(object_delete(templ));
    object_t *inst = object_createInstance(templ,0,NULL);
    assert_null(inst);

    sahtest_success();
}

int obj_create_instance_duplicate_index() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);
    object_t *templ = object_getObject(root,"TestRoot1",0,NULL);
    assert_not_null(templ);

    object_t *inst = object_createInstance(templ,2,NULL);
    assert_not_null(inst);
    inst = object_createInstance(templ,2,NULL);
    assert_null(inst);

    sahtest_success();
}

int obj_create_instance_duplicate_key() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);
    object_t *templ = object_getObject(root,"TestRoot1",0,NULL);
    assert_not_null(templ);

    object_t *inst = object_createInstance(templ,0,"Key");
    assert_not_null(inst);
    inst = object_createInstance(templ,0,"Key");
    assert_null(inst);

    sahtest_success();
}

void obj_delete_setup() {
    datamodel_initialize(&dm);
    object_t *root = datamodel_root(&dm);
    object_t *obj = object_create(root,"TestRoot1",object_attr_template);
    obj = object_create(obj,"Child",0);
    obj = object_create(root,"TestRoot2",0);
    object_commit(root);
}

void obj_delete_teardown() {
    datamodel_cleanup(&dm);
}

int obj_delete() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);
    object_t *obj = object_getObject(root,"TestRoot2",0,NULL);
    assert_not_null(obj);

    assert_bool(object_delete(obj));
    assert_bool(obj->state == object_state_deleted);

    sahtest_success();
}

int obj_delete_twice() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);
    object_t *obj = object_getObject(root,"TestRoot2",0,NULL);
    assert_not_null(obj);

    assert_bool(object_delete(obj));
    assert_bool(obj->state == object_state_deleted);
    assert_bool(object_delete(obj));

    sahtest_success();
}

int obj_delete_null() {
    assert_bool(object_delete(NULL) == false);

    sahtest_success();
}

int obj_delete_wrong_state() {
    object_t *root = datamodel_root(&dm);
    assert_not_null(root);

    object_t *obj = object_create(root,"TestRoot3",0);
    assert_not_null(obj);
    assert_bool(obj->state == object_state_created);

    assert_bool(object_delete(obj) == false);

    sahtest_success();
}
