/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <sahtest/sahtest.h>
#include <pcb/core/datamodel.h>
#include <pcb/core/object.h>
#include <pcb/core/parameter.h>
#include <pcb/utils.h>
#include "datamodel_priv.h"
#include "object_priv.h"
#include "parameter_priv.h"

static datamodel_t dm;
static object_t *root = NULL;
static object_t *templ = NULL;
static object_t *inst = NULL;
static parameter_t *root_param = NULL;
static parameter_t *templ_param = NULL;
static parameter_t *inst_param = NULL;

void param_create_setup() {
    datamodel_initialize(&dm);
    root = datamodel_root(&dm);
    templ = object_create(root,"TestRoot1",object_attr_template);
}

void param_create_teardown() {
    datamodel_cleanup(&dm);
}

void param_delete_setup() {
    param_create_setup();
    root_param = parameter_create(root,"Foo",parameter_type_string,parameter_attr_default);
    templ_param = parameter_create(root,"Foo",parameter_type_string,parameter_attr_default);
    inst = object_createInstance(templ,5,"Foo");
    inst_param = object_getParameter(inst,"Foo");

    object_commit(root);
}

void param_delete_teardown() {
    datamodel_cleanup(&dm);
}

int param_create() {
    parameter_t *param = parameter_create(root,"Foo",parameter_type_string,parameter_attr_default);
    assert_not_null(param);
    assert_bool(parameter_owner(param) == root);
    assert_bool(parameter_state(param) == parameter_state_created);
    assert_bool(object_commit(root));
    assert_bool(parameter_state(param) == parameter_state_ready);

    param = parameter_create(templ,"Foo",parameter_type_string,parameter_attr_default);
    assert_not_null(param);
    assert_bool(parameter_owner(param) == templ);
    assert_bool(parameter_state(param) == parameter_state_created);
    object_t *instance = object_createInstance(templ,5,"Foo");
    assert_bool(object_hasParameters(instance));
    parameter_t *instance_param = object_getParameter(instance,"Foo");
    assert_not_null(instance_param);
    assert_bool(parameter_state(instance_param) == parameter_state_validate_created);
    assert_bool(instance_param->definition == param->definition);
    assert_bool(parameter_owner(instance_param) == instance);

    assert_bool(object_commit(root));
    assert_bool(parameter_state(param) == parameter_state_ready);
    assert_bool(parameter_state(instance_param) == parameter_state_ready);

    sahtest_success();
}

int param_create_null_object() {
    parameter_t *param = parameter_create(NULL,"Foo",parameter_type_string,parameter_attr_default);
    assert_null(param);

    assert_null(parameter_owner(NULL));

    sahtest_success();
}

int param_create_empty_name() {
    parameter_t *param = parameter_create(root,"",parameter_type_string,parameter_attr_default);
    assert_null(param);

    sahtest_success();
}

int param_create_invalid_name() {
    parameter_t *param = parameter_create(root,"My.String/Param.",parameter_type_string,parameter_attr_default);
    assert_null(param);

    sahtest_success();
}

int param_create_null_name() {
    parameter_t *param = parameter_create(root,NULL,parameter_type_string,parameter_attr_default);
    assert_null(param);

    sahtest_success();
}

int param_create_duplicate_name() {
    parameter_t *param = parameter_create(root,"Foo",parameter_type_string,parameter_attr_default);
    assert_not_null(param);
    assert_bool(parameter_owner(param) == root);
    param = parameter_create(root,"Foo",parameter_type_string,parameter_attr_default);
    assert_null(param);

    sahtest_success();
}

int param_create_template_only_param() {
    parameter_t *param = parameter_create(templ,"Foo",parameter_type_string,parameter_attr_template_only);
    assert_not_null(param);
    assert_bool(parameter_owner(param) == templ);
    object_t *instance = object_createInstance(templ,5,"Foo");
    assert_bool(object_hasParameters(instance) == false);

    assert_bool(object_commit(root));

    sahtest_success();
}

int param_copy() {
    parameter_t *param = parameter_create(root,"Foo",parameter_type_string,parameter_attr_default);
    assert_not_null(param);
    assert_bool(parameter_owner(param) == root);

    param = parameter_copy(templ,param);
    assert_not_null(param);
    assert_bool(parameter_owner(param) == templ);

    assert_bool(object_commit(root));

    sahtest_success();
}

int param_copy_parameter_name_already_used() {
    parameter_t *param = parameter_create(root,"Foo",parameter_type_string,parameter_attr_default);
    assert_not_null(param);
    assert_bool(parameter_owner(param) == root);

    param = parameter_copy(root,param);
    assert_null(param);

    sahtest_success();
}

int param_copy_null_source() {
    parameter_t *param = parameter_create(root,"Foo",parameter_type_string,parameter_attr_default);
    assert_not_null(param);
    assert_bool(parameter_owner(param) == root);

    param = parameter_copy(templ,NULL);
    assert_null(param);

    sahtest_success();
}

int param_copy_null_dest() {
    parameter_t *param = parameter_create(root,"Foo",parameter_type_string,parameter_attr_default);
    assert_not_null(param);
    assert_bool(parameter_owner(param) == root);

    param = parameter_copy(NULL,param);
    assert_null(param);

    sahtest_success();
}

int param_delete() {
    assert_bool(parameter_delete(root_param));
    assert_bool(parameter_state(root_param) == parameter_state_deleted);
    assert_bool(object_parameterCount(root) == 1);
    assert_bool(object_commit(root));
    assert_bool(object_parameterCount(root) == 0);

    sahtest_success();
}

int param_delete_wrong_state() {
    parameter_t *param = parameter_create(root,"Bar",parameter_type_string,parameter_attr_default);
    assert_bool(parameter_delete(param) == false);
    sahtest_success();
}

int param_delete_twice() {
    assert_bool(parameter_delete(root_param));
    assert_bool(parameter_state(root_param) == parameter_state_deleted);
    assert_bool(object_parameterCount(root) == 1);
    assert_bool(parameter_delete(root_param));
    assert_bool(parameter_state(root_param) == parameter_state_deleted);
    assert_bool(object_parameterCount(root) == 1);

    sahtest_success();
}

int param_delete_null() {
    assert_bool(parameter_delete(NULL) == false);

    sahtest_success();
}
