/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <sahtest/sahtest.h>
#include <pcb/utils/circular_buffer.h>

#define LOOPS 100

static circbuf_t cb;

void test_circular_buffer_setup() {
    if (!circbuf_initialize(&cb, 100)) {
        SAHTEST_ERROR("Failed to initialize circular buffer");
    }
}

void test_circular_buffer_teardown() {
    circbuf_cleanup(&cb);
}

int test_open_close() {
    FILE *stream = circbuf_fopen(&cb, "w+");
    T_NOT_NULL(stream);
    T_SUCCEED(fclose(stream));

    return 0;
}

int test_rw() {
    FILE *stream = circbuf_fopen(&cb, "w+");
    char buffer[256];

    T_EQUAL(int, fwrite("123456", sizeof(char), 7, stream), 7);
    fflush(stream);
    T_EQUAL(int, fread(buffer, sizeof(char), 7, stream), 7);
    T_EQUAL(int, strncmp(buffer, "123456", 7), 0);
    T_SUCCEED(fprintf(stream, "123456"));
    fclose(stream);

    return 0;
}

int test_write_off_by_one_bug() {
    char out[1024];
    FILE *stream = circbuf_fopen(&cb, "w+");
    T_SUCCEED(feof(stream));
    char *str = (char *) "123";
    fwrite(str, sizeof(char), 3, stream);
    fflush(stream);

    fwrite(str, sizeof(char), 3, stream);
    fflush(stream);

    ssize_t size = circbuf_availableForRead(&cb);

    T_EQUAL(int, size, 6);
    strncpy(out, (char *) circbuf_readBuffer(&cb), 6);
    circbuf_markAsRead(&cb, 6);
    T_EQUAL(int, strncmp(out, "123123", 6), 0);

    fclose(stream);

    return 0;
}

int test_reset() {
    FILE *stream = circbuf_fopen(&cb, "w+");

    T_EQUAL(int, fwrite("123456", sizeof(char), 7, stream), 7);
    fflush(stream);
    circbuf_reset(&cb,0);
    T_EQUAL(int, feof(stream), 0);
    fclose(stream);

    return 0;
}


int test_rw_over_boundary() {
    FILE *stream = circbuf_fopen(&cb, "w+");
    char buffer[256];
    int i = 0;
    char *str[2] = { (char *) "123456", (char *) "abcdef" };
    for (i = 0; i < LOOPS; i++) {
        char *test = str[i%2];
        T_EQUAL(int, fwrite(test, sizeof(char), 7, stream), 7);
        fflush(stream);
        T_EQUAL(int, fread(buffer, sizeof(char), 7, stream), 7);
        T_EQUAL(int, strncmp(buffer, test, 7), 0);
    }
    fclose(stream);

    return 0;
}

int test_rw_cornercase() {
    FILE *stream = circbuf_fopen(&cb, "w+");
    char buffer[256];
    int i = 0;
    char *str[2] = { (char *) "12345", (char *) "abcde" };
    for (i = 0; i < LOOPS; i++) {
        char *test = str[i%2];
        T_EQUAL(int,fwrite(test, sizeof(char), 5, stream), 5);
        fflush(stream);
        T_EQUAL(int,fread(buffer, sizeof(char), 5, stream), 5);
        T_EQUAL(int,strncmp(buffer, test, 5), 0);
    }
    fclose(stream);

    return 0;
}

int test_direct_api() {
    FILE *stream = circbuf_fopen(&cb, "w+");

    T_NOT_NULL(stream);
    T_VERIFY(circbuf_stream(&cb) == stream);
    T_VERIFY(circbuf_size(&cb) == 100);
    T_VERIFY(circbuf_availableForRead(&cb) == 0);

    int i = 0;
    char *str[2] = { (char *) "123456", (char *) "abcdef" };
    ssize_t toRead = 0;
    for (i = 0; i < LOOPS; i++) {
        char *test = str[i%2];
        T_VERIFY(fwrite(test, sizeof(char), 7, stream) == 7);
        fflush(stream);
        toRead = circbuf_availableForRead(&cb);
        T_VERIFY(strncmp((char *) circbuf_readBuffer(&cb), test, toRead) == 0);
        T_VERIFY(circbuf_markAsRead(&cb, toRead) == toRead);
        if (toRead < 7) {
            T_VERIFY(strncmp((char *) circbuf_readBuffer(&cb), test + toRead, 7 - toRead) == 0);
            T_VERIFY(circbuf_markAsRead(&cb, 7 - toRead) == 7 - toRead);
        }
    }
    fclose(stream);

    return 0;
}

int test_eof() {
    char buffer[256];
    FILE *stream = circbuf_fopen(&cb, "w+");
    T_VERIFY(feof(stream));
    T_VERIFY(fwrite("123456", sizeof(char), 7, stream) == 7);
    fflush(stream);
    T_VERIFY(!feof(stream));
    T_VERIFY(fread(buffer, sizeof(char), 10, stream) == 7);
    T_VERIFY(feof(stream));
    T_VERIFY(fread(buffer, sizeof(char), 10, stream) == 0);
    T_VERIFY(feof(stream));
    fclose(stream);

    return 0;
}

int test_rw_increase_size() {
    FILE *stream = circbuf_fopen(&cb, "w+");
    char buffer[256];
    int i = 0;
    char *str[2] = { (char *) "123456", (char *) "abcdef" };
    for (i = 0; i < LOOPS; i++) {
        char *test = str[i%2];
        T_VERIFY(fwrite(test, sizeof(char), 7, stream) == 7);
        fflush(stream);
    }
    for (i = 0; i < LOOPS; i++) {
        char *test = str[i%2];
        T_VERIFY(fread(buffer, sizeof(char), 7, stream) == 7);
        T_VERIFY(strncmp(buffer, test, 7) == 0);
    }

    fclose(stream);

    return 0;
}

int test_resize_with_read_after_write() {
    FILE *stream = circbuf_fopen(&cb, "w+");
    char buffer[256];

    /* fill the buffer */
    T_VERIFY(fwrite("0123456789", sizeof(char), 10, stream) == 10);
    fflush(stream);
    /* read a bit */
    T_VERIFY(fread(buffer, sizeof(char), 7, stream) == 7);

    /* trigger a resize */
    T_VERIFY(fwrite("0123456789", sizeof(char), 5, stream) == 5);
    fflush(stream);
    T_VERIFY(fwrite("0123456789", sizeof(char), 5, stream) == 5);
    fflush(stream);
    fclose(stream);

    return 0;
}

/*
int test_under_constraint() {
    int limit = 5 * 1024 * 1024, oldLimit;
    int i;
    struct rlimit l;
    if (getenv("VALGRIND_OPTS") != NULL) {
        fprintf(stderr, "WARN   : PCBTestCircBuf::test_under_constraint() skipped, valgrind detected\n");
    return 0;
    }

    T_VERIFY(getrlimit(RLIMIT_AS, &l) == 0);
    oldLimit = l.rlim_cur;
    l.rlim_cur = limit;
    T_VERIFY(setrlimit(RLIMIT_AS, &l) == 0);

    circbuf_t cb;
    T_VERIFY(!circbuf_initialize(&cb, limit * 2));
    T_VERIFY(errno == ENOMEM);

    circbuf_initialize(&cb, 100);
    FILE *stream = circbuf_fopen(&cb, "w+");
    T_VERIFY(stream);

    for (i = 0; i < limit; i++) {
        errno = 0;
        if (fwrite("0123456789", sizeof(char), 10, stream) < 10)
            break;
        errno = 0;
        fflush(stream);
        if (errno) {
            break;
        }
        T_VERIFY(errno == 0);
    }
    T_VERIFY(errno == ENOMEM);

    fclose(stream);

    circbuf_cleanup(&cb);

    l.rlim_cur = oldLimit;
    T_VERIFY(setrlimit(RLIMIT_AS, &l) == 0);
    return 0;
}
*/

int test_seek() {
    FILE *stream = circbuf_fopen(&cb, "w+");
    char result[100];

    /* simple case */
    char *str = (char *) "foo is to be replaced by bar";
    size_t len = strlen(str);

    T_VERIFY(fwrite(str, sizeof(char), len, stream) == len);
    T_VERIFY(fseek(stream, 0 - len, SEEK_CUR) == 0);
    T_VERIFY(fwrite("bar", sizeof(char), 3, stream) == 3);
    T_VERIFY(fseek(stream, len - 3, SEEK_CUR) == 0);
    T_VERIFY(fflush(stream) == 0);
    T_VERIFY(fread(result, sizeof(char), len, stream) == len);
    T_VERIFY(strncmp(result, "bar is to be replaced by bar", len) == 0);

    /*  cross boundaries */
    int i;
    for (i = 0; i < LOOPS; i++) {
        T_VERIFY(fwrite(str, sizeof(char), len, stream) == len);
        T_VERIFY(fseek(stream, 0 - len, SEEK_CUR) == 0);
        T_VERIFY(fwrite("bar", sizeof(char), 3, stream) == 3);
        T_VERIFY(fseek(stream, len - 3, SEEK_CUR) == 0);
        T_VERIFY(fflush(stream) == 0);
        T_VERIFY(fread(result, sizeof(char), len, stream) == len);
        T_VERIFY(strncmp(result, "bar is to be replaced by bar", len) == 0);
    }

    return 0;
}
