#include "cups.h"
#include "config.h"

#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <debug/sahtrace.h>

/**
 * send standard output/error to /dev/null
 */
int loc_cups_redirect_output(struct cups *c)
{
  (void)c;
  if (freopen("/dev/null", "w", stdout) == NULL) {
    SAH_TRACE_WARNING("unable to redirect stdout to /dev/null: %s", strerror(errno));
    return -1;
  }
  if (freopen("/dev/null", "w", stderr) == NULL) {
    SAH_TRACE_WARNING("unable to redirect stderr to /dev/null: %s", strerror(errno));
    return -1;
  }
  return 0;
}

/**
 * initialize the cups module
 */
int cups_initialize(struct cups *c)
{
  /* sanity checks */
  struct stat statbuf;
  if (stat(c->cups, &statbuf) != 0) {
    SAH_TRACE_FATAL("unable to find cups binary %s: %s", c->cups, strerror(errno));
    goto error;
  }

  /* check if we can execute those binaries */
  if (access(c->cups, R_OK | X_OK) != 0) {
    SAH_TRACE_FATAL("unable to execute cups binary %s: %s", c->cups, strerror(errno));
    goto error;
  }

  c->start_count = 0;

  /* setup signals */
  return 0;

 error:
  return -1;
}

/**
 * release cups module
 */
int cups_release(struct cups *c)
{
  return cups_stop(c);
}

/**
 * check if cups is running
 */
int cups_is_running(struct cups *c)
{
  return c->cupsPid != 0;
}

/**
 * check if cups is stopped
 */
int cups_is_disabled(struct cups *c)
{
  return c->cupsPid == 0;
}

/**
 * starts cups
 */
int cups_start(struct cups *c, const char *config_path)
{
  /* make sure that config file is there */
  struct stat statb;

  char *path_cupsd_conf = (char *)calloc(strlen(config_path) + strlen("cupsd.conf") + 1, sizeof(char));
  if(!path_cupsd_conf) {
    SAH_TRACE_ERROR("Memory full");
    return -1;
  }
  strcpy(path_cupsd_conf, config_path);
  strcat(path_cupsd_conf, "cupsd.conf");

  if (stat(path_cupsd_conf, &statb) != 0) {
    SAH_TRACE_ERROR("cups cannot start, config file is not there: %s", strerror(errno));
    free(path_cupsd_conf);
    return -1;
  }

  if (c->start_count >= CUPS_RESTART_COUNT) {
    SAH_TRACE_ERROR("cups has already been restarted %i time, assume it is not working", CUPS_RESTART_COUNT);
    free(path_cupsd_conf);
    return -1;
  }

  /* start cups */
  c->cupsPid = vfork();
  if (c->cupsPid < 0) {
    SAH_TRACE_ERROR("unable to fork");
    goto error;
  } else if (c->cupsPid == 0) {
    dup2(fileno(stderr), fileno(stdout));

    SAH_TRACE_INFO("starting %s with config path %s", c->cups,path_cupsd_conf);
    loc_cups_redirect_output(c);
    execl(c->cups,"cupsd","-f","-c",path_cupsd_conf,NULL); // run the daemon in the foreground
    exit(1);
    SAH_TRACE_ERROR("unable to start the cups process: %s", strerror(errno));
  } else {
    SAH_TRACE_APP_INFO("cups started");
  }

  c->started_at = time(NULL);
  c->start_count++;
  free(path_cupsd_conf);

  return 0;

 error:
  return -1;
}

/**
 * stops cups
 */
int cups_stop(struct cups *c)
{
  int rc = 0;

  SAH_TRACE_INFO("stopping cups");

  if (c->cupsPid != 0) {
    int status = 0;
    SAH_TRACE_APP_INFO("sending SIGTERM to %u", c->cupsPid);
    if (kill(c->cupsPid, SIGKILL) != 0) {
      SAH_TRACE_ERROR("unable to stop cups %u: %s", c->cupsPid, strerror(errno));
      rc = - 1;
    } else {
      SAH_TRACE_APP_INFO("waiting for cups to stop");
      waitpid(c->cupsPid, &status, 0);
      c->cupsPid = 0;
      SAH_TRACE_APP_INFO("cups has stopped");
    }
  }

  return rc;
}

/**
 * send a HUP to cups proces
 */
int cups_reload(struct cups *c)
{
  int rc = 0;
  SAH_TRACE_APP_INFO("restarting cups");
  if (c->cupsPid != 0) {
    if (kill(c->cupsPid, SIGHUP) != 0) {
      SAH_TRACE_ERROR("unable to reload cups: %s", strerror(errno));
      rc = -1;
    }
  }
  return rc;
}

