#include <stdlib.h>
#include <debug/sahtrace.h>
#include <pcb/utils.h>
#include <pcb/core.h>
#include <string.h>
#include <nemo/client.h>
#include <netdev/common_api.h>
#include <fw/fw_client.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <components.h>

#include "cfgfile.h"
#include "config.h"
#include "cups.h"

#define CONFIG_FILE CONFIG_RWDATAPATH"/cups.odl"

bool enableTriggered();

static plugin_config_t *plugin_config = NULL;
static pcb_timer_t *retry_timer = NULL;
static struct cups cups_process;
static struct cfg_file_writer cw;
static int objectsLoaded = 0;
static nemo_query_t *get_addresses = NULL;
static variant_list_t ip_addresses;
static object_t *cups_object = NULL;

/**
 * signal handler for SIGCHLD, called when cups proces exits
 */
void loc_main_signal_child(int sig __attribute__ ((unused)))
{
  SAH_TRACE_WARNING("SIGCHLD received %p", &waitpid);

#ifdef HAS_WAITID
  SAH_TRACE_INFO("SIGCHLD received %p", &waitpid);
  siginfo_t infop;
  memset(&infop, 0, sizeof(infop));
  int rc = 0;
  SAH_TRACE_INFO("SIGCHLD received %p", &waitpid);
  if ((rc = waitid(P_ALL, 0, &infop, WNOHANG)) != 0) {
    SAH_TRACE_WARNING("unable to wait for child: %s", strerror(errno));
    return;
  }
  SAH_TRACE_INFO("SIGCHLD received %p", &waitpid);
  /* if no child died */
  if (rc == 0 && infop.si_pid == 0)
    return;


  SAH_TRACE_WARNING("samba process %u exited", infop.si_pid);
  cups_process.cupsPid = 0;
#else
  int rc = 0;
  int status = 0;
  if (cups_process.cupsPid > 0 && (rc = waitpid(cups_process.cupsPid, &status, WNOHANG)) == cups_process.cupsPid) {
    SAH_TRACE_WARNING("cups process exited with %u", WEXITSTATUS(status));
    cups_process.cupsPid = 0;
  }
#endif

  pcb_timer_start(retry_timer, CUPS_RESTART_WAITTIME*1000);

}

static void loc_retry_timer_event(pcb_timer_t *timer __attribute__ ((unused)),
                                    void *userdata __attribute__ ((unused)))
{
    SAH_TRACE_INFO("retry timer event");
    enableTriggered();
}

static void loc_get_ip_address(const variant_t *result, void *userdata __attribute__ ((unused)))
{
  variant_toArray(&ip_addresses, result);
  enableTriggered();
}

static int loc_netdev_loaded()
{
  int rc = -1;

  nemo_client_initialize(plugin_config->pcb, plugin_config->system_bus);

  /* add handler for ip address */
  char *netdev = object_parameterCharValue(cups_object, "NetworkDevice");
  SAH_TRACE_INFO("fetching ip address for %s", netdev);
  if ((get_addresses = netdev_openQuery_getAddrs(netdev, "cups", "ipv4", "down", loc_get_ip_address, NULL)) == NULL) {
    SAH_TRACE_FATAL("unable to fetch ip addresses: %s", error_string(pcb_error));
    free(netdev);
    goto error;
  }

  rc = 0;

  free(netdev);
 error:
  return rc;
}

static bool loc_object_loaded(request_t *rq,
                              pcb_t *pcb __attribute__ ((unused)),
                              peer_info_t *from __attribute__ ((unused)),
                              void *userdata __attribute__ ((unused)))
{
  reply_t *reply = request_reply(rq);
  reply_item_t *item;

  for (item = reply_firstItem(reply); item; item = reply_nextItem(item)) {
    if (reply_item_type(item) == reply_type_error) {
      SAH_TRACE_WARNING("unable to wait for object: %s (%s)", reply_item_errorDescription(item), reply_item_errorInfo(item));
      continue;
    }
    if (reply_item_type(item) == reply_type_object) {
      const char *name = object_name(reply_item_object(item), path_attr_key_notation);
      if (name == NULL) {
      //  SAH_TRACE_WARNING("unable to fetch object name: %s", error_string(pcb_error));
        continue;
      }

      if (strcmp(name, "NetDev") == 0) {
        SAH_TRACE_INFO("NetDev object is there");
        if (loc_netdev_loaded() == 0) {
          objectsLoaded ++;
        }
      } else if (strcmp(name, "Firewall") == 0) {
        SAH_TRACE_INFO("Firewall object is there");
        objectsLoaded ++;
      }
    }
  }

  if (objectsLoaded == 2) {
    enableTriggered();
    request_destroy(rq);
  }
  return true;
}

bool pcb_plugin_initialize(pcb_t *pcb, int argc, char *argv[]) {
    (void)argc;
    (void)argv;
    SAH_TRACE_INFO("Initializing cups plugin");

    plugin_config = pcb_pluginConfig(pcb);

    datamodel_t *dm = pcb_datamodel(pcb);
    cups_object = object_getObject(datamodel_root(dm), "CupsService", 0, NULL);

    // load the data
    if(!datamodel_load(dm,CONFIG_FILE,SERIALIZE_FORMAT(serialize_format_odl,0,0))) {
      SAH_TRACE_WARNING("Failed to load configuration from file %s",CONFIG_FILE);
    }

    cups_process.cups = CUPS_DEFAULT_CUPSD_PATH;
    cw.config_path    = CUPS_DEFAULT_CONFIG_PATH;
    cw.log_level      = CUPS_DEFAULT_DEBUG_LEVEL;
    cw.request_root   = CUPS_DEFAULT_REQUESTROOT;
    cw.printer_name   = CUPS_DEFAULT_PRINTERNAME;
    cw.printer_uri    = CUPS_DEFAULT_PRINTERURI;

    if(cups_initialize(&cups_process)) {
      SAH_TRACE_WARNING("Failed to initialize cups context");
      return false;
    }
    connection_setSignalEventHandler(pcb_connection(pcb), SIGCHLD, loc_main_signal_child);
    retry_timer = pcb_timer_create();
    pcb_timer_setHandler(retry_timer, loc_retry_timer_event);


    /* wait for NetDev, Firewall */
    request_t *rq = request_create_getObject("", 1, /*request_no_object_caching |*/ request_notify_object_added);
    if (rq == NULL) {
      SAH_TRACE_FATAL("unable to wait for objects");
      return false;
    }
    request_setReplyHandler(rq, loc_object_loaded);
    if (!pcb_sendRequest(pcb, plugin_config->system_bus, rq)) {
      SAH_TRACE_FATAL("unable to wait for objects");
      request_destroy(rq);
      return false;
    }

    return true;
}



void pcb_plugin_cleanup() {
    SAH_TRACE_INFO("Cleaning up cups plugin");

    if(cups_release(&cups_process)) {
        SAH_TRACE_WARNING("Failed to release cups context");
    }
    deleteService(plugin_config->system_bus, "Firewall", "cups");
    if (get_addresses) {
      nemo_closeQuery(get_addresses);
      nemo_client_cleanup();
      variant_list_cleanup(&ip_addresses);
    }
}

bool saveHandler() {
    // save settings
    if (datamodel_save(cups_object,UINT32_MAX,CONFIG_FILE,SERIALIZE_FORMAT(serialize_format_odl,0,0))==false) {
        SAH_TRACE_WARNING("Failed to save configuration to file %s",CONFIG_FILE);
    }

    return true;
}


int loc_main_regenerate_config()
{
  int rc = 0;
  SAH_TRACE_APP_INFO("regenerating config files");
  free(cw.interface_name);
  cw.interface_name = NULL;
  free(cw.ip_address);
  cw.ip_address = NULL;
  free(cw.hostname);
  cw.hostname = NULL;

  variant_list_iterator_t *it;
  variant_list_for_each(it, &ip_addresses) {
    variant_map_t *v = variant_da_map(variant_list_iterator_data(it));
    const char *interfacename = string_buffer(variant_da_string(variant_map_iterator_data(variant_map_find(v, "NetDevName"))));
    if(interfacename) cw.interface_name = strdup(interfacename);
    const char *ip_address = string_buffer(variant_da_string(variant_map_iterator_data(variant_map_find(v, "Address"))));
    if(ip_address) cw.ip_address = strdup(ip_address);
  }

  cw.hostname = object_parameterCharValue(cups_object, "ServerName");

  /* initialize it */
  if ((rc = cfg_initialize(&cw)) != 0) {
    SAH_TRACE_ERROR("unable to initialize configuration writer");
    goto error;
  }

  /* generate cupsd.conf */
  if ((rc = cfg_write_cupsd_conf(&cw)) != 0) {
    SAH_TRACE_ERROR("unable to write the cupsd.conf configuration file");
    goto stoprelease;
  }

  /* generate printers.conf */
  if ((rc = cfg_write_printers_conf(&cw)) != 0) {
    SAH_TRACE_ERROR("unable to write the printers.conf configuration file");
    goto stoprelease;
  }

 stoprelease:
  cfg_release(&cw);

 error:
  return rc;
}

bool enableHandler(parameter_t *parameter) {
    (void)parameter;

    cups_process.start_count = 0;
    return enableTriggered();
}


bool enableTriggered() {
    if(cups_object==NULL) return true;

    bool enable = object_parameterBoolValue(cups_object, "Enable");

    if(objectsLoaded < 2) {
      SAH_TRACE_INFO("Netdev and firewall plugins not yet loaded");
    }
    /* update the service */
    char *netdev = object_parameterCharValue(cups_object, "NetworkDevice");
    if (netdev) {
      setService(plugin_config->system_bus, "Firewall", "cups", netdev, "631", "6", 4, object_parameterBoolValue(cups_object, "Enable"));
      free(netdev);
    } else {
      SAH_TRACE_WARNING("unable to fetch NetworkDevice parameter");
    }

    if(enable==true) { // start the service
      // regenerate the config file
      if(loc_main_regenerate_config()<0) {
        SAH_TRACE_ERROR("Cannot regenerate config");
        return false;
      }

      if(cups_is_running(&cups_process)) {
          SAH_TRACE_WARNING("Reload cups");
          if(cups_reload(&cups_process)<0) {
            SAH_TRACE_ERROR("Could not reload");
            return false;
          }
      } else {
          SAH_TRACE_WARNING("Start cups");
          if(cups_start(&cups_process, cw.config_path)<0) {
            SAH_TRACE_ERROR("Could not start");
            return false;
          }
      }
    } else {
        // stop the service
        if(cups_is_running(&cups_process)) {
            SAH_TRACE_WARNING("Stop cups");
            if(cups_stop(&cups_process)<0) {
              SAH_TRACE_ERROR("Could not start");
              return false;
            }
        }
    }
    saveHandler();
    return true;
}
