/* Shared library add-on to iptables to add TCPMSS target support.
 *
 * Copyright (c) 2000 Marc Boucher
*/
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <stdint.h>
#include "../include/ebtables_u.h"
#include <ebt_tcpmss/ebt_tcpmss.h>

struct mssinfo {
    struct xt_entry_target t;
    struct xt_tcpmss_info mss;
};


static int xtables_strtoul(const char *s, char **end, unsigned long *value,
                     unsigned long min, unsigned long max)
{
    unsigned long v;
    char *my_end;

    v = strtoul(s, &my_end, 0);

    if (my_end == s)
        return 0;
    if (end != NULL)
        *end = my_end;

    if (min <= v && (max == 0 || v <= max)) {
        if (value != NULL)
            *value = v;
        if (end == NULL)
            return *my_end == '\0';
        return 1;
    }

    return 0;
}

static int xtables_strtoui(const char *s, char **end, unsigned int *value,
                     unsigned int min, unsigned int max)
{
    unsigned long v = 0;
    bool ret;

    ret = xtables_strtoul(s, end, &v, min, max);
    if (value != NULL)
        *value = v;
    return ret;
}

static void init(struct ebt_entry_target *target)
{

}

static void __TCPMSS_help(int hdrsize)
{
    printf(
"TCPMSS target mutually-exclusive options:\n"
"  --set-mss value               explicitly set MSS option to specified value\n"
"  --clamp-mss-to-pmtu           automatically clamp MSS value to (path_MTU - %d)\n",
hdrsize);
}

static void TCPMSS_help(void)
{
    __TCPMSS_help(40);
}

static const struct option TCPMSS_opts[] = {
    {.name = "set-mss",           .has_arg = true,  .val = '1'},
    {.name = "clamp-mss-to-pmtu", .has_arg = false, .val = '2'}
};

static int __TCPMSS_parse(int c, char **argv, int invert, unsigned int *flags,
        const struct ebt_u_entry *entry, struct ebt_entry_target **target,
                          int hdrsize)
{
    struct xt_tcpmss_info *mssinfo
        = (struct xt_tcpmss_info *)(*target)->data;

    switch (c) {
        unsigned int mssval;

    case '1':
        if (*flags)
            printf("TCPMSS target: Only one option may be specified");
        if (!xtables_strtoui(optarg, NULL, &mssval,
            0, UINT16_MAX - hdrsize))
            printf("Bad TCPMSS value \"%s\"", optarg);
        
        mssinfo->mss = mssval;
        *flags = 1;
        break;

    case '2':
        if (*flags)
            printf("TCPMSS target: Only one option may be specified");
        mssinfo->mss = XT_TCPMSS_CLAMP_PMTU;
        *flags = 1;
        break;

    default:
        return 0;
    }

    return 1;
}

static int TCPMSS_parse(int c, char **argv, int invert, 
                        const struct ebt_u_entry *entry, unsigned int *flags, struct ebt_entry_target **target)
{
    return __TCPMSS_parse(c, argv, invert, flags, entry, target, 40);
}

static void TCPMSS_check(const struct ebt_u_entry *entry,
        const struct ebt_entry_target *target, const char *name,
        unsigned int hookmask, unsigned int time)
{
    if (!target)
        printf("TCPMSS target: At least one parameter is required");
}

static void TCPMSS_print(const struct ebt_u_entry *entry,
        const struct ebt_entry_target *target)
{
    const struct xt_tcpmss_info *mssinfo =
        (const struct xt_tcpmss_info *)target->data;
    if(mssinfo->mss == XT_TCPMSS_CLAMP_PMTU)
        printf("TCPMSS clamp to PMTU ");
    else
        printf("TCPMSS set %u ", mssinfo->mss);
}


static struct ebt_u_target tcpmss_target = {
    .name       = "TCPMSS",
    .init       = init,
    .size       = sizeof(struct xt_tcpmss_info),
    .help       = TCPMSS_help,
    .parse      = TCPMSS_parse,
    .final_check    = TCPMSS_check,
    .print      = TCPMSS_print,
    .extra_ops = TCPMSS_opts,
};

static void _init(void) __attribute__ ((constructor));
void _init(void)
{
    register_target(&tcpmss_target);
}
