/*
#############################################################################
#
# Copyright (c) Orange / France Telecom
# December 2012
# All rights reserved
#
# Author:
#	Frederic LEROY <frederic1.leroy@orange.com> - FT/SCE/G2S/NRS/IT LABS/UOGO/SFP/ARMOR
#	For Orange Labs - FT/OLNC/OLN/WTC/WHS/MAG
#
# Name:		DHCP L2 relay
# Release:	1
# Date:		26/12/2012
#
#############################################################################
*/

 /* Shared library add-on to iptables to add DHCP_RELAY target support. */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <linux/netfilter/x_tables.h>
#include "../include/ebtables_u.h"
#include <linux/types.h>
#include <linux/pkt_sched.h>
#include <l2dhcprelay/ebt_dhcp_relay.h>

#define OPT_CID 0x01
#define OPT_RID 0x02
#define OPT_SID 0x04
#define OPT_MODE 0x08
static void
DHCP_RELAY_help(void)
{
	printf(
"DHCP_RELAY target options: \n"
"	--customer-id <CID>	define a agent circuit id\n"
"	--remote-id <RID>	define a agent remote id\n"
"	--subscriber-id <SID> define a agent subscriber id\n"
"	--mode <forward|replace|discard|strip> mode this target is used in, default is forward\n"
"				forward: if packets already contain option282, don't change them\n"
"				replace: if packets contain option82, replace it with the configuredv values\n"
"				discard: if packets contain option82, discard the packet\n"
"				strip:   if packets contain option82, remove it. This option can't be used with other options.\n"
"\n"
"The DHCP_RELAY target option add/removes option82\n"
"\n"
		);
}

static void init(struct ebt_entry_target *target)
{

}

static const struct option DHCP_RELAY_opts[] = {
	{ "customer-id", required_argument, NULL, '2' },
	{ "remote-id", required_argument, NULL, '3' },
	{ "subscriber-id", required_argument, NULL, '4' },
	{ "mode", required_argument, NULL, '5' },
	{ .name = NULL }
};

static int xtables_check_inverse(const char option[],
			  int *my_optind, int argc, char **argv)
{
	if (option == NULL || strcmp(option, "!") != 0)
		return 0;

	printf( "Using intrapositioned negation "
			"(`--option ! this`) is deprecated in favor of "
			"extrapositioned (`! --option this`).\n");

	if (my_optind != NULL) {
		optarg = argv[*my_optind];
		++*my_optind;
		if (argc && *my_optind > argc)
			printf("no argument following `!'");
	}

	return 1;
}

static char char2hex (char c)
{
	char res;

	if ((c >= '0') && (c <= '9')) {
		res = c - '0';
	} else if ((c >= 'A') && (c <= 'F')) {
		res = c - 'A' + 0xa;
	} else if ((c >= 'a') && (c <= 'f')) {
		res = c - 'a' + 0xa;
	} else {
		res =  -1;
	}

	return res;
}

static int
parse_string(const char *s, uint8_t *dst)
{
	int i;
	int slen = strlen(s);

	if (slen >= OPT82_MAX * 2) {
		printf("string must be shorter than %i characters\n", OPT82_MAX * 2);
		return -1;
	}
	if (slen % 2 != 0) {
		printf("rid,cid string should have an even number of characters\n");
		return -1;
	}

	for (i = 0; i < slen / 2; i++) {
		char c = char2hex(s[i*2]);

		if (c < 0) {
			printf("Invalid char: %c\n", s[i]);
			return -1;
		}
		dst[i] = c << 4;

		c = char2hex(s[(i * 2) + 1]);
		if (c < 0) {
			printf("Invalid char: %c\n", s[(i * 2) + 1]);
			return -1;
		}
		dst[i] |= c & 0xf;
	}


	return i;
}

int dhcp_relay_check_option(struct xt_dhcp_relay_target_info *tginfo, unsigned int *flags, unsigned int mask)
{
	check_option(flags, mask);

	if ((tginfo->mode == STRIP) && (*flags & ~OPT_MODE)) {
		printf("--mode=strip must be used without other options\n");
		return -1;
	}
	return 0;
}

static int
DHCP_RELAY_parse(int c, char **argv, int argc, const struct ebt_u_entry *entry, unsigned int *flags,
	  struct ebt_entry_target **target)
{
	struct xt_dhcp_relay_target_info *tginfo = (struct xt_dhcp_relay_target_info *)(*target)->data;
	int len;

	switch (c) {
		case '2':
			if (dhcp_relay_check_option(tginfo, flags, OPT_CID) != 0)
				goto error;
			xtables_check_inverse(argv[optind-1], &optind, 0, argv);
			len = parse_string(argv[optind-1], tginfo->cid);
			if (len <  0)
				goto error;
			tginfo->cid_len = (uint8_t)len;
			tginfo->bitmask |= EBT_DHCP_RELAY_CID;
			break;

		case '3':
			if (dhcp_relay_check_option(tginfo, flags, OPT_RID) != 0)
				goto error;
			xtables_check_inverse(argv[optind-1], &optind, 0, argv);
			len = parse_string(argv[optind-1], tginfo->rid);
			if (len <  0)
				goto error;
			tginfo->rid_len = (uint8_t)len;
			tginfo->bitmask |= EBT_DHCP_RELAY_RID;
			break;
		case '4':
			if (dhcp_relay_check_option(tginfo, flags, OPT_SID) != 0)
				goto error;
			xtables_check_inverse(argv[optind-1], &optind, 0, argv);
			len = parse_string(argv[optind-1], tginfo->sid);
			if (len <  0)
				goto error;
			tginfo->sid_len = (uint8_t)len;
			tginfo->bitmask |= EBT_DHCP_RELAY_SID;
			break;
		case '5':
			if (dhcp_relay_check_option(tginfo, flags, OPT_MODE) != 0)
				goto error;
			if(strcmp("forward", optarg) == 0) {
				tginfo->mode = FORWARD;
			} else if (strcmp("replace", optarg) == 0) {
				tginfo->mode = REPLACE;
			} else if (strcmp("discard", optarg) == 0) {
				tginfo->mode = DISCARD;
			} else if (strcmp("strip", optarg) == 0) {
				tginfo->mode = STRIP;
			} else {
				printf("Invalid value for on-present option: %s\n", optarg);
				goto error;
			}
			tginfo->bitmask |= EBT_DHCP_RELAY_MODE;
			break;

		default:
			break;
	}
	return 1;
error:
	return 0;
}

static void
DHCP_RELAY_final_check(const struct ebt_u_entry *entry,
		const struct ebt_entry_target *target, const char *name,
		unsigned int hookmask, unsigned int time)
{
}

static void
DHCP_RELAY_print(const struct ebt_u_entry *entry,
		const struct ebt_entry_target *target)
{
	struct xt_dhcp_relay_target_info *tginfo = (struct xt_dhcp_relay_target_info *)target->data;

	const char *m;
	if ( tginfo->cid_len > 0 )
		printf(" --cid %s", tginfo->cid);
	if ( tginfo->rid_len > 0 )
		printf(" --rid %s", tginfo->rid);

	switch (tginfo->mode) {
		case FORWARD: m = "forward"; break;
		case REPLACE: m = "replace"; break;
		case DISCARD: m = "discard"; break;
		case STRIP: m = "strip"; break;
		default: m = ""; break;
	}
	printf(" --mode=%s", m);
}

static int compare(const struct ebt_entry_target *t1, const struct ebt_entry_target *t2) {
	struct xt_dhcp_relay_target_info *ti1 = (struct xt_dhcp_relay_target_info *)t1->data;
	struct xt_dhcp_relay_target_info *ti2 = (struct xt_dhcp_relay_target_info *)t2->data;

	if ((ti1->bitmask & EBT_DHCP_RELAY_MODE) && (ti1->mode != ti2->mode))
		return 0;

	if ((ti1->bitmask & EBT_DHCP_RELAY_CID) &&
		((ti1->cid_len != ti2->cid_len) ||
		(memcmp(ti1->cid, ti2->cid, ti1->cid_len) != 0)))
		return 0;

	if ((ti1->bitmask & EBT_DHCP_RELAY_RID) &&
		((ti1->rid_len != ti2->rid_len) ||
		(memcmp(ti1->rid, ti2->rid, ti1->rid_len) != 0)))
		return 0;

	if ((ti1->bitmask & EBT_DHCP_RELAY_SID) &&
		((ti1->sid_len != ti2->sid_len) ||
		(memcmp(ti1->sid, ti2->sid, ti1->sid_len) != 0)))
		return 0;

	return 1;
}

static struct ebt_u_target dhcp_relay_target = {

	.name		= "DHCP_RELAY",
	.size		= sizeof(struct xt_dhcp_relay_target_info),
	.help		= DHCP_RELAY_help,
	.init		= init,
	.parse		= DHCP_RELAY_parse,
	.final_check	= DHCP_RELAY_final_check,
	.print		= DHCP_RELAY_print,
	.extra_ops	= DHCP_RELAY_opts,
	.compare	= compare,
};

static void _init(void) __attribute__ ((constructor));
void _init(void)
{
	register_target(&dhcp_relay_target);
}
