/*
#############################################################################
#
# Copyright (c) Orange / France Telecom
# December 2012
# All rights reserved
#
# Author:
#	Frederic LEROY <frederic1.leroy@orange.com> - FT/SCE/G2S/NRS/IT LABS/UOGO/SFP/ARMOR
#	For Orange Labs - FT/OLNC/OLN/WTC/WHS/MAG
#
# Name:		DHCP L2 relay
# Release:	1
# Date:		26/12/2012
#
#############################################################################
*/

#include <linux/version.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/ip.h>
#include <linux/in.h>
#include <linux/udp.h>
#include <linux/tcp.h>
#include <linux/jiffies.h>
#include <linux/spinlock.h>
#include <linux/jhash.h>

#include <linux/netfilter/x_tables.h>
#include <linux/netfilter_bridge/ebtables.h>
#include <ebt_dhcp_relay.h>


MODULE_ALIAS("ebt_DHCP_RELAY");

#define DHCS_PORT 67
#define DHCC_PORT 68
#define DHCP_RELAY_AGENT_INFO_OPTION 82
#define END_OPTION 255
#define PAD_OPTION 0

static inline unsigned char *next_option(unsigned char *option)
{
	unsigned char *ret;
	if (*option == PAD_OPTION)
		ret = option + 1;
	else
		ret = option + option[1] + 2;
	return ret;
}

/* IPv4 */
static unsigned int
dhcp_relay_tg(struct sk_buff *skb, const struct xt_target_param *par)
{
	struct udphdr *uh = NULL;
	struct dhcp_packet *dh = NULL;
	struct iphdr *iph = ip_hdr(skb);
	struct xt_dhcp_relay_target_info *tginfo=(struct xt_dhcp_relay_target_info *)par->targinfo;
	int len_payload;
	int rid_len, cid_len, sid_len;
	int off;
	char *rid_data = NULL, *cid_data = NULL, *sid_data = NULL;
	unsigned char *option, *option82 = NULL;
	uint16_t old_len;

	/* Skip unmanaged protocols */
	if (! (iph->protocol == IPPROTO_UDP))
		return EBT_CONTINUE;

	uh = (struct udphdr *) ((void *)iph + (iph->ihl * 4));

	/* port check */
	if (! (
	 (ntohs(uh->source) == DHCS_PORT && ntohs(uh->dest) == DHCC_PORT) ||
	 (ntohs(uh->source) == DHCC_PORT && ntohs(uh->dest) == DHCS_PORT)))
		return EBT_CONTINUE;

	/* check size */
	if (skb->len < (sizeof(struct udphdr) + sizeof (struct dhcp_packet)))
		return EBT_CONTINUE;

	dh = (struct dhcp_packet *) ((void*)uh + sizeof(struct udphdr));

	/* check magic cookie */
	if (dh->magic_cookie != htonl(0x63825363)) {
		printk(KERN_DEBUG "%s:%d wrong magic cookie, dropping packet\n", __FUNCTION__, __LINE__);
		return EBT_DROP;
	}
	if (skb_linearize(skb) != 0)
		return EBT_DROP;

	for (option = dh->data ; *option != END_OPTION && option <
		skb_tail_pointer(skb); option = next_option(option)) {
		if (*option == DHCP_RELAY_AGENT_INFO_OPTION ) {
			option82 = option;
			break;
		}
	}

	if (option82 != NULL) {
		if (tginfo->mode == DISCARD) {
			return EBT_DROP;
		} if ((tginfo->mode == REPLACE) || (tginfo->mode == STRIP)) {
			/* remove option82 */
			unsigned char *noption = next_option(option82);
			unsigned int o82_len = noption - option82;

			memmove(option82, noption, skb_tail_pointer(skb) -
					noption);
			skb_trim(skb, skb->len - o82_len);

			if ((tginfo->mode == STRIP)) {
				iph = ip_hdr(skb);
				uh = (struct udphdr *) ((void *)iph + (iph->ihl * 4));

				/* need to update length in ip header and checksum*/
				old_len = ntohs(iph->tot_len);
				iph->tot_len = htons(old_len - o82_len);
				csum_replace2(&iph->check, htons(old_len), htons(iph->tot_len));
				/* update also udp length */
				old_len = ntohs(uh->len);
				uh->len = htons(old_len - o82_len);
				/* compute full udp checksum */
				uh->check = 0;
				skb->csum = skb_checksum(skb, ip_hdrlen(skb), skb->len - ip_hdrlen(skb), 0);
				uh->check = csum_tcpudp_magic(iph->saddr, iph->daddr, skb->len - ip_hdrlen(skb), IPPROTO_UDP, skb->csum);
				if (uh->check == 0)
					uh->check = CSUM_MANGLED_0;
			}
		} else {
			return EBT_CONTINUE;
		}
	}

	if (tginfo->mode == STRIP)
		return EBT_CONTINUE;

	/* look further for end*/
	for (; *option != END_OPTION && option <
		skb_tail_pointer(skb); option = next_option(option));

	printk(KERN_INFO "option(%p) tail(%p) *option(%u)\n", option,
			skb_tail_pointer(skb), *option);

	if (*option == END_OPTION) {	/* Add option 82 */
		int end_option_offset = option - skb->data;
		int8_t *buf;
		int padding_size = skb_tail_pointer(skb) - option - 1;

		if (!skb_make_writable(skb, skb->len))
		{
			printk(KERN_DEBUG "%s:%d can't write on skb\n", __FUNCTION__, __LINE__);
			return EBT_DROP;
		}

		len_payload = 2; /* option 82 header */

		/* rid */
		rid_len = 0;
		if (tginfo->rid_len != 0) {
			rid_len = tginfo->rid_len;
			rid_data = tginfo->rid;
			len_payload += 2 + rid_len;
		}

		/* cid */
		cid_len = 0;
		if (tginfo->cid_len != 0) {
			cid_len = tginfo->cid_len;
			cid_data = tginfo->cid;
			len_payload += 2 + cid_len;
		}

		/* sid */
		sid_len = 0;
		if (tginfo->sid_len != 0) {
			sid_len = tginfo->sid_len;
			sid_data = tginfo->sid;
			len_payload += 2 + sid_len;
		}


		/* Check size and reallocate if not enough room */
		if (padding_size < len_payload) {
            		int missing_bytes = len_payload - padding_size;

			if (skb_tailroom(skb) < missing_bytes) {
				if (pskb_expand_head(skb, 0, missing_bytes, GFP_ATOMIC)) {
					printk(KERN_DEBUG "%s:%d can't expand skb\n", __FUNCTION__, __LINE__);
					return EBT_DROP;
				}
			}

			if (!skb_put(skb, missing_bytes)) {
				printk(KERN_DEBUG "%s:%d can't expand skb\n", __FUNCTION__, __LINE__);
				return EBT_DROP;
			}
		}

		/* Fill packet with new bootp file */
		buf = skb->data + end_option_offset;
		buf[0] = DHCP_RELAY_AGENT_INFO_OPTION;
		buf[1] = len_payload - 2;
		off = 2;
		if (cid_len) {
			buf[off] = 1;
			buf[off+1] = cid_len;
			memcpy(buf + off + 2, cid_data, cid_len);
			off += 2 + cid_len;
		}
		if (rid_len) {
			buf[off] = 2;
			buf[off+1] = rid_len;
			memcpy(buf + off + 2, rid_data, rid_len);
			off += 2 + rid_len;
		}
		if (sid_len) {
			buf[off] = 6;
			buf[off+1] = sid_len;
			memcpy(buf + off + 2, sid_data, sid_len);
		}
		buf[len_payload] = 255;

		iph = ip_hdr(skb);
		uh = (struct udphdr *) ((void *)iph + (iph->ihl * 4));
		/* need to update length in ip header and checksum*/
		old_len = ntohs(iph->tot_len);
		iph->tot_len = htons(skb_tail_pointer(skb) - skb_network_header(skb));
		csum_replace2(&iph->check, htons(old_len), htons(iph->tot_len));

		/* update also udp length */
		uh->len = htons(skb_tail_pointer(skb) - (unsigned char *)uh);
		/* compute full udp checksum */
		uh->check = 0;
		skb->csum = skb_checksum(skb, ip_hdrlen(skb), skb->len - ip_hdrlen(skb), 0);
		uh->check = csum_tcpudp_magic(iph->saddr, iph->daddr, skb->len - ip_hdrlen(skb), IPPROTO_UDP, skb->csum);
		if (uh->check == 0)
			uh->check = CSUM_MANGLED_0;
	}
	return EBT_CONTINUE;
}

bool dhcp_relay_tg_check (const struct xt_tgchk_param *par)
{
	return true;
}

void dhcp_relay_tg_destroy (const struct xt_tgdtor_param *par)
{
}

static struct xt_target dhcp_relay_tg_reg[] __read_mostly = {
	{
	.name		= "DHCP_RELAY",
	.target		= dhcp_relay_tg,
	.targetsize = XT_ALIGN(sizeof(struct xt_dhcp_relay_target_info)),
	.checkentry = dhcp_relay_tg_check,
	.destroy	= dhcp_relay_tg_destroy,
	.me	= THIS_MODULE,
	.family		= NFPROTO_BRIDGE,
	.revision	= 0,
	},
};

static int __init dhcp_relay_tg_init(void)
{
	return xt_register_targets(dhcp_relay_tg_reg, ARRAY_SIZE(dhcp_relay_tg_reg));
}

static void __exit dhcp_relay_tg_fini(void)
{
	xt_unregister_targets(dhcp_relay_tg_reg, ARRAY_SIZE(dhcp_relay_tg_reg));
}

module_init(dhcp_relay_tg_init);
module_exit(dhcp_relay_tg_fini);

