#include "nemo_wrapper.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"
#include "pcb_eventloop.h"

#include <pcb/pcb_client.h>

/* static variables */
static nemo_interface s_nemo_wan_intf_status;
static const char NeMoDslObject[] = "NeMo.Intf.dsl0";

// MUTEX for locking nemo_wan_interface
ithread_mutex_t NeMoWanInterfaceMutex = PTHREAD_MUTEX_INITIALIZER;

static bool NeMoRequestInvoke(pcb_t *client, peer_info_t *peer, request_t *req) {
  string_t s;
  int n = 0;
  string_initialize(&s, 64);
  parameter_iterator_t *it;
  for (it = request_firstParameter(req); it; it = request_nextParameter(it)) {
    char *charvalue = request_parameterValue(it)?variant_char(request_parameterValue(it)):NULL;
    string_appendFormat(&s, "%s%s%s%s", n++?", ": "", request_parameterName(it), charvalue?"=":"", charvalue?charvalue:"");
    free(charvalue);
  }
  trace(4, "send request %s.%s(%s)", request_path(req), request_functionName(req), string_buffer(&s)?string_buffer(&s):"");
  string_cleanup(&s);

  if (!pcb_sendRequest(client, peer, req)) {
    trace(1, "%s : Failed to send request: %d (%s)", pcb_error, error_string(pcb_error), __FUNCTION__);
    return false;
  }

  if (!pcb_waitForReply(client, req, NULL)) {
    trace(1, "%s: Communication failed", __FUNCTION__);
    return false;
  }

  reply_t *reply = request_reply(req);
  if (!reply) {
    trace(1, "%s: No reply available", __FUNCTION__);
    return false;
  }

  reply_item_t *item;
  for (item = reply_firstItem(reply); item; item = reply_nextItem(item)) {
    if (reply_item_type(item) == reply_type_error) {
      trace(1, "%s: Function call error (%d) - %s - %s",
          __FUNCTION__, reply_item_error(item), reply_item_errorDescription(item), reply_item_errorInfo(item));
      return false;
    }
  }
  return true;
}

static const char* nemoInterfacef2path(const char *intf) {
  static char pathbuf[64];
  snprintf(pathbuf, 64, "NeMo.Intf.%s", intf?intf:"(null)");
  return pathbuf;
}

static bool NeMoRequestAddCharParameter(request_t *req, const char *name, const char *value) {
  bool retval = false;
  variant_t val;
  if (!variant_initialize(&val, variant_type_string)) {
    trace(1, "%s: variant_initialize failed", __FUNCTION__);
    goto leave;
  }
  if (!variant_setChar(&val, value)) {
    trace(1, "variant_setChar(%s) failed", __FUNCTION__ , value?value:"(null)");
    goto leave;
  }
  if (!request_addParameter(req, name, &val)) {
    trace(1, "request_addParameter(%s=%s) failed", __FUNCTION__, name?name:"(null)", value?value:"(null)");
    goto leave;
  }
  retval = true;
leave:
  variant_cleanup(&val);
  return retval;
}

const variant_t* NeMoRequestReturnValue(request_t *req) {
  reply_t *reply = request_reply(req);
  if (!reply) {
    trace(1, "%s: No reply available", __FUNCTION__);
    return NULL;
  }
  reply_item_t *item;
  for (item = reply_firstItem(reply); item; item = reply_nextItem(item)) {
    if (reply_item_type(item) == reply_type_function_return) {
      return reply_item_returnValue(item);
    }
  }
  return NULL;
}

static bool NeMoGetLuckyAddrAddress(pcb_t *client, peer_info_t *peer, const char *intf, char *flag, char *luckyAdr, int maxResultlength) {
  const char *retval = NULL;
  request_t *req = NULL;
  bool rc = false;

  if (!intf) {
    trace(1, "%s: Invalid parameter: intf is NULL!", __FUNCTION__);
    goto leave;
  }

  req = request_create_executeFunction(nemoInterfacef2path(intf), "luckyAddrAddress", request_common_path_key_notation | request_function_args_by_name);
  if (!req)
    goto leave;
  if (!NeMoRequestAddCharParameter(req, "flag", flag))
    goto leave;
  if (!NeMoRequestAddCharParameter(req, "traverse", "down"))
    goto leave;
  if (!NeMoRequestInvoke(client, peer, req))
    goto leave;

  if (NULL != (retval = variant_da_char(NeMoRequestReturnValue(req))) && strlen(retval)) {
    strncpy(luckyAdr, retval, maxResultlength);
    rc = true;
  }
leave:
  request_destroy(req);
  return rc;
}

static bool NeMoGetLuckyIntf(pcb_t *client, peer_info_t *peer, const char *intf, char *luckyInterface, int maxResultlength) {
  const char *retval = NULL;
  request_t *req = NULL;
  bool rc = false;

  if (!intf) {
    trace(1, "%s: Invalid parameter: intf is NULL!", __FUNCTION__);
    goto leave;
  }
  req = request_create_executeFunction(nemoInterfacef2path(intf), "luckyIntf", request_common_path_key_notation | request_function_args_by_name);
  if (!req)
    goto leave;
  if (!NeMoRequestAddCharParameter(req, "flag", "netdev"))
    goto leave;
  if (!NeMoRequestAddCharParameter(req, "traverse", "down"))
    goto leave;
  if (!NeMoRequestInvoke(client, peer, req))
    goto leave;

  if (NULL != (retval = variant_da_char(NeMoRequestReturnValue(req))) && strlen(retval)) {
    strncpy(luckyInterface, retval, maxResultlength);
    rc = true;
  }

leave:
  request_destroy(req);
  return rc;
}

static bool NeMoIntfHasFlag(pcb_t *client, peer_info_t *peer, const char *intf, const char *flag, bool *hasFlag) {
  request_t *req = NULL;
  bool rc = false;

  if (!intf || !flag || !hasFlag) {
    trace(1, "%s: Invalid input parameters", __FUNCTION__);
    goto leave;
  }
  req = request_create_executeFunction(nemoInterfacef2path(intf), "hasFlag", request_common_path_key_notation | request_function_args_by_name);
  if (!req)
    goto leave;
  if (!NeMoRequestAddCharParameter(req, "flag", flag))
    goto leave;
  if (!NeMoRequestInvoke(client, peer, req))
    goto leave;

  *hasFlag = variant_bool(NeMoRequestReturnValue(req));
  rc = true;

leave:
  request_destroy(req);
  return rc;
}

static bool NeMoIntfGetFirstParameter(pcb_t *client, peer_info_t *peer, const char *intf, const char *name, uint32_t *result) {
  request_t *req = NULL;
  bool rc = false;

  if (!intf || !name || !result) {
    trace(1, "%s: Invalid input parameters", __FUNCTION__);
    goto leave;
  }
  req = request_create_executeFunction(nemoInterfacef2path(intf), "getFirstParameter", request_common_path_key_notation | request_function_args_by_name);
  if (!req)
    goto leave;
  if (!NeMoRequestAddCharParameter(req, "name", name))
    goto leave;
  if (!NeMoRequestInvoke(client, peer, req))
    goto leave;

  if (variant_toUInt32(result, NeMoRequestReturnValue(req)))
  {
    rc = true;
  }

leave:
  request_destroy(req);
  return rc;
}

bool NeMoInitWrapper(const char *nemoWanInterfaceName) {
  ithread_mutex_lock(&NeMoWanInterfaceMutex);

  s_nemo_wan_intf_status.linkIsUp = false;
  s_nemo_wan_intf_status.isConnected = false;
#ifdef UPNP_ENABLE_IPV6
  s_nemo_wan_intf_status.ipv6Address[0] = '\0';
#endif // UPNP_ENABLE_IPV6
  s_nemo_wan_intf_status.ipv4Address[0] = '\0';
  s_nemo_wan_intf_status.luckyIntfName[0] = '\0';

  strncpy( s_nemo_wan_intf_status.nemoInterface, nemoWanInterfaceName ? nemoWanInterfaceName : "", IFNAMSIZ);

  ithread_mutex_unlock(&NeMoWanInterfaceMutex);

  return true;
}

bool NeMoUpdateWanStatus(pcb_t *client, peer_info_t *peer, bool physicalIsUp, bool logicalIsUp) {
  bool PCPEnabled = false;

  if (!client || !peer)
    return false;

  ithread_mutex_lock(&NeMoWanInterfaceMutex);

  s_nemo_wan_intf_status.linkIsUp = physicalIsUp;
  s_nemo_wan_intf_status.isConnected = physicalIsUp && logicalIsUp;

  if (s_nemo_wan_intf_status.isConnected) {
    char *interfaceName = s_nemo_wan_intf_status.nemoInterface;

    NeMoGetLuckyIntf(client, peer, interfaceName, s_nemo_wan_intf_status.luckyIntfName, IFNAMSIZ);

    NeMoGetLuckyAddrAddress(client, peer, interfaceName, "ipv4", s_nemo_wan_intf_status.ipv4Address, INET6_ADDRSTRLEN);
#ifdef UPNP_ENABLE_IPV6
    NeMoGetLuckyAddrAddress(client, peer, interfaceName, "ipv6", s_nemo_wan_intf_status.ipv6Address, INET6_ADDRSTRLEN);
#endif // UPNP_ENABLE_IPV6
  }
  else {
#ifdef UPNP_ENABLE_IPV6
    s_nemo_wan_intf_status.ipv6Address[0] = '\0';
#endif // UPNP_ENABLE_IPV6
    s_nemo_wan_intf_status.ipv4Address[0] = '\0';
    s_nemo_wan_intf_status.luckyIntfName[0] = '\0';
  }  

  if ((PCPEnabled = PCB_firewallPCPIsEnabled()) == true) {
    // DSLITE is overruling the NeMo.Intf.data external IP address
    const char *PCPWanIP = PCB_firewallPCPExternalIP();

    if (PCPWanIP && strlen(PCPWanIP)) {
      strncpy(s_nemo_wan_intf_status.ipv4Address, PCPWanIP, INET6_ADDRSTRLEN);
    }
    else {
      s_nemo_wan_intf_status.ipv4Address[0] = '\0';
    }

#ifdef UPNP_ENABLE_IPV6
    s_nemo_wan_intf_status.ipv6Address[0] = '\0';
#endif // UPNP_ENABLE_IPV6
  }

  trace(3, "%s: PCP.Enable(%d) - physicalIsUp(%d) - logicalIsUp(%d) - linkIsUp(%d) - isConnected(%d) - luckyIntfName(%s) - ipv6Address(%s) - ipv4Address(%s)",
        __FUNCTION__,
        PCPEnabled ? 1 : 0,
        physicalIsUp,
        logicalIsUp,
        s_nemo_wan_intf_status.linkIsUp,
        s_nemo_wan_intf_status.isConnected,
        s_nemo_wan_intf_status.luckyIntfName,
#ifdef UPNP_ENABLE_IPV6
        s_nemo_wan_intf_status.ipv6Address,
#else
        "NONE",
#endif //  UPNP_ENABLE_IPV6
        s_nemo_wan_intf_status.ipv4Address);

  ithread_mutex_unlock(&NeMoWanInterfaceMutex);

  return true;
}

bool NeMoGetWanStatus(nemo_interface *nemo_intf) {
  bool rc =false;

  if (nemo_intf) {

    ithread_mutex_lock(&NeMoWanInterfaceMutex);
    memcpy(nemo_intf, &s_nemo_wan_intf_status, sizeof(nemo_interface));
    ithread_mutex_unlock(&NeMoWanInterfaceMutex);

    rc = true;
  }

  return rc;
}

char* NeMoGetEthernetLinkStatus(nemo_interface *nemo_intf) {
  if (nemo_intf && nemo_intf->linkIsUp)
    return  "Up";
  else
    return "Down";
}

char* NeMoGetConnectionStatus(nemo_interface *nemo_intf) {
  if (nemo_intf && nemo_intf->isConnected)
    return  "Connected";
  else
    return "Disconnected";
}

bool NeMoGetLinkLayerMaxBitRates(nemo_link_layer_max_bit_rates *bitRates)
{
  bool rc = false;
  bool isDSL = false;

  if (!bitRates) {
    trace(1, "%s: Invalid parameter: bitRates is NULL!", __FUNCTION__);
    goto leave;
  }

  if (NeMoIntfHasFlag(PCB_getClient(), PCB_getPeer(), "data", "dsl", &isDSL)) {

    if (isDSL)
    {
      bitRates->wanAccessType = DSL;

      if (NeMoIntfGetFirstParameter(PCB_getClient(), PCB_getPeer(), "data", "UpstreamMaxRate", (uint32_t*) &(bitRates->upstreamMaxBitRate)) &&
          NeMoIntfGetFirstParameter(PCB_getClient(), PCB_getPeer(), "data", "DownstreamMaxRate", (uint32_t*) &(bitRates->downStreamMaxBitRate)) &&
          NeMoIntfGetFirstParameter(PCB_getClient(), PCB_getPeer(), "data", "UpstreamCurrRate", (uint32_t*) &(bitRates->UpstreamCurrRate)) &&
          NeMoIntfGetFirstParameter(PCB_getClient(), PCB_getPeer(), "data", "DownstreamCurrRate", (uint32_t*) &(bitRates->DownstreamCurrRate))) {
          // values are given in kbps, so convert to bps
          bitRates->upstreamMaxBitRate *= 1000;
          bitRates->downStreamMaxBitRate *=1000;
          bitRates->UpstreamCurrRate *= 1000;
          bitRates->DownstreamCurrRate *=1000;
          rc = true;
      }
    }
    else
    {
      bitRates->wanAccessType = ETHERNET;
      uint32_t maxRate = 0;

      if (NeMoIntfGetFirstParameter(PCB_getClient(), PCB_getPeer(), "data", "MaxBitRateSupported", &maxRate)) {
          // recalculate: for ethernet the result is in Mbps and we need bps
          maxRate *= 1048576;
          bitRates->upstreamMaxBitRate = bitRates->downStreamMaxBitRate = bitRates->UpstreamCurrRate = bitRates->DownstreamCurrRate = maxRate;
          rc = true;
      }
    }
  }

  if (!rc)
  {
    trace(1, "%s: Unable to retrieve UpstreamMaxRate, DownstreamMaxRate values", __FUNCTION__);
  }

leave:
  return rc;
}


