/**
 * This file is part of Nokia InternetGatewayDevice v2 reference implementation
 * Copyright © 2009 Nokia Corporation and/or its subsidiary(-ies).
 * Contact: mika.saaranen@nokia.com
 * Developer(s): jaakko.pasanen@tieto.com, opensource@tieto.com
 *
 * This file is part of igd2-for-linux project
 * Copyright © 2011 France Telecom.
 * Contact: fabrice.fontaine@orange-ftgroup.com
 * Developer(s): fabrice.fontaine@orange-ftgroup.com, rmenard.ext@orange-ftgroup.com
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program, see the /doc directory of this program. If
 * not, see http://www.gnu.org/licenses/.
 *
 */

#include <stddef.h>
#include <unistd.h>
#include <regex.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <syslog.h>
#include <arpa/inet.h>
#include <linux/sockios.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <wchar.h>
#include <wctype.h>
#include <ctype.h>
#include <upnp/upnp.h>
#include <upnp/ixml.h>
#include "globals.h"
#include "util.h"

#ifdef LOG_TO_FILE
FILE *logfileP = NULL;
#endif


/**
 * Open new socket.
 *
 * @return created socket if success, -1 if failure.
 */
static int get_sockfd(void)
{
    static int sockfd = -1;

    if (sockfd == -1)
    {
        if ((sockfd = socket(PF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
        {
            perror("user: socket creating failed");
            return (-1);
        }
    }
    return sockfd;
}

/**
 * Create union from items separated with space in strings given.
 * Use free() for created string.
 *
 * @param str1 First string to unionize
 * @param str2 Second string to unionize
 * @return New allocated union string or NULL if fails.
 */
char* createUnion(const char *str1, const char *str2)
{
    if (!str1 || !str2)
        return NULL;

    char *unionStr = NULL;
    size_t size = strlen(str1) + strlen(str2) + 2;
    unionStr = (char *)malloc(size);
    if (!unionStr)
        return NULL;

    // base for the union is created by copying whole str1 to it
    strcpy(unionStr,str1);

    // make copy of str2 so we can tokenize it
    char copystr[strlen(str1)];
    // go through all roles in list
    strcpy(copystr,str2);
    char *item = strtok(copystr, " ");
    if (item)
    {
        do
        {
            // do "raw" check that this item isn't already in current items
            if ( strstr(unionStr,item) == NULL )
            {
                // add new item at the end of rolelist
                if (strlen(unionStr) > 0)
                {
                    strcat(unionStr, " ");
                }
                strcat(unionStr, item);
            }
        } while ((item = strtok(NULL, " ")));

    }

    return unionStr;
}

/**
 * Get values for send bytes and packets and received bytes and packets for
 * external interface from /proc/net/dev
 *
 * @param stats Unsigned long array with size of STATS_LIMIT
 * @return 0 if fails to open file, 1 if succeed to get values.
 */
int readStats(unsigned long stats[STATS_LIMIT])
{
    FILE *proc = NULL;
    int read = 0;
    int rc = 0;
    size_t lineLength = 128;
    char *lineBuffer = malloc(lineLength);

    if (NULL == lineBuffer)
    {
         fprintf(stderr, "memory allocation failed\n");
         return 0;
    }

    if (NULL == ( proc = fopen("/proc/net/dev", "r")))
    {
        fprintf(stderr, "failed to open\n");
        goto out;
    }

    // skip first two lines
    if ( -1 == getline(&lineBuffer, &lineLength, proc) ||
			   -1 == getline(&lineBuffer, &lineLength, proc))
    {
        fprintf(stderr, "failed to read file\n");
        goto out;
    }

    // parse stats
    while(1)
    {
      char *stringStart = NULL;
      char *intf = NULL;
      // getline will re-allocate extra memory (and adjust lineLength) in case the lineBuffer is too small
      if (-1 == getline(&lineBuffer, &lineLength, proc))
      {
        break;  // EOF
      }

      stringStart = lineBuffer;
      // remove heading spaces
      while (*stringStart==' ') {
        stringStart++;
      }
      if (NULL == ( intf = (char*) strtok(stringStart, ":")) ||
          NULL == ( stringStart = (char*) strtok(NULL, ":")))
      {
        continue;
      }

      read = sscanf(stringStart, " %lu %lu %*u %*u %*u %*u %*u %*u %lu %lu %*u %*u %*u %*u %*u %*u\n",
                    &stats[STATS_RX_BYTES], &stats[STATS_RX_PACKETS], &stats[STATS_TX_BYTES], &stats[STATS_TX_PACKETS]);

      if (read == 4 && strncmp(intf, g_vars.extInterfaceName, IFNAMSIZ) == 0) {
        rc = 1;
        break;
      }
    }

    fclose(proc);

    if (!rc)
    {
       stats[STATS_RX_BYTES] =0;
       stats[STATS_RX_PACKETS] =0;
       stats[STATS_TX_BYTES] = 0;
       stats[STATS_TX_PACKETS] = 0;
    }

out:
    free(lineBuffer);

    return rc;
}

/**
 * Trims leading and trailing white spaces from a string.
 *
 * @param str String to trim.
 * @return Trimmed string or NULL
 */
char *trimString(char *str)
{
    char *ptr;
    if (!str)
        return NULL;
    if (!*str)
        return str;

    // trim leading white spaces
    while(isspace(*str)) str++;

    // trim trailing white spaces
    ptr = str + strlen(str);
    while (isspace(*--ptr))
        *ptr = '\0';

    return str;
}

/**
 * THIS FUNCTION IS NOT ACTUALLY NEEDED, if you use UpnpMakeActionResponse and such
 * functions for creating responses. libupnp then takes care of escaping xmls.
 * Unescaping on the other hand must be done by ourself here.
 *
 * Change given XML string in escaped form.
 * Following characters are converted:
 *  '<'  -->  "&lt;"
 *  '>'  -->  "&gt;"
 *  '"'  -->  "&quot;"
 *  '''  -->  "&apos;"
 *  '&'  -->  "&amp;"
 *
 * User should free returned pointer.
 *
 * @param xml String to turn escaped xml.
 * @return Escaped xml string or NULL if failure.
 */
char* escapeXMLString(char *xml)
{
    xml = trimString(xml);
    if (xml == NULL)
        return NULL;

    char *escXML = NULL;
    size_t size = strlen(xml);
    size_t alloc = size +1;

    escXML = realloc(NULL, alloc);
    if (!escXML)
        return NULL;

    int i,j; // i goes through original xml and j through escaped escXML
    for (i=0,j=0; i < size; i++)
    {
        switch (xml[i])
        {
            case '<' :
            {
                char *new_buf;
                alloc += strlen("&lt;");
                new_buf = realloc (escXML, alloc);
                if (!new_buf) {
                    return NULL;
                }
                escXML = new_buf;

                strcpy(escXML+j, "&lt;");
                j += strlen("&lt;");
                break;
            }
            case '>' :
            {
                char *new_buf;
                alloc += strlen("&gt;");
                new_buf = realloc (escXML, alloc);
                if (!new_buf) {
                    return NULL;
                }
                escXML = new_buf;

                strcpy(escXML+j, "&gt;");
                j += strlen("&gt;");
                break;
            }
            case '"' :
            {
                char *new_buf;
                alloc += strlen("&quot;");
                new_buf = realloc (escXML, alloc);
                if (!new_buf) {
                    return NULL;
                }
                escXML = new_buf;

                strcpy(escXML+j, "&quot;");
                j += strlen("&quot;");
                break;
            }
            case '\'' :
            {
                char *new_buf;
                alloc += strlen("&apos;");
                new_buf = realloc (escXML, alloc);
                if (!new_buf) {
                    return NULL;
                }
                escXML = new_buf;

                strcpy(escXML+j, "&apos;");
                j += strlen("&apos;");
                break;
            }
            case '&' :
            {
                char *new_buf;
                alloc += strlen("&amp;");
                new_buf = realloc (escXML, alloc);
                if (!new_buf) {
                    return NULL;
                }
                escXML = new_buf;

                strcpy(escXML+j, "&amp;");
                j += strlen("&amp;");
                break;
            }
            default :
            {
                escXML[j++] = xml[i];
                break;
            }
        }
    }

    if (j > 0)
        escXML[j] = '\0';

    return escXML;
}

/**
 * Change given XML string in unescaped form.
 * Following characters are converted:
 *  "&lt;"    -->  '<'
 *  "&gt;"    -->  '>'
 *  "&quot;"  -->  '"'
 *  "&apos;"  -->  '''
 *  "&amp;"   -->  '&'
 *
 * User should free returned pointer.
 *
 * @param escXML String to turn unescaped xml.
 * @return Unescaped xml string or NULL if failure.
 */
char* unescapeXMLString(char *escXML)
{
    escXML = trimString(escXML);
    if (escXML == NULL)
        return NULL;

    char *xml = NULL;
    size_t size = strlen(escXML);

    xml = (char *)malloc(size);
    if (!xml)
        return NULL;

    memset(xml, '\0', size);

    int i,j; // i goes through unescaped xml and j through escaped escXML
    for (i=0,j=0; i < size && j < size; i++)
    {
        if (strncmp(escXML+j, "&lt;", strlen("&lt;")) == 0)
        {
            xml[i] = '<';
            j += strlen("&lt;");
        }
        else if (strncmp(escXML+j, "&gt;", strlen("&gt;")) == 0)
        {
            xml[i] = '>';
            j += strlen("&gt;");
        }
        else if (strncmp(escXML+j, "&quot;", strlen("&quot;")) == 0)
        {
            xml[i] = '"';
            j += strlen("&quot;");
        }
        else if (strncmp(escXML+j, "&apos;", strlen("&apos;")) == 0)
        {
            xml[i] = '\'';
            j += strlen("&apos;");
        }
        else if (strncmp(escXML+j, "&amp;", strlen("&amp;")) == 0)
        {
            xml[i] = '&';
            j += strlen("&amp;");
        }
        else
        {
            xml[i] = escXML[j];
            j++;
        }
    }

    xml[i] = '\0';

    return xml;
}


/**
 * Change given string in uppercase. Converts given string first as wide-character string
 * and then transliterate that to upper case. Finally convert upper case wide-character string
 * back to character string and return it. Such a complex procedure guarantees that umlaut chars
 * are uppercased correctly, not sure if even all utf-8 chars.
 *
 * User should free returned pointer.
 *
 * @param str String to turn uppercase.
 * @return Upper cased string or NULL if failure.
 */
char *toUpperCase(const char * str)
{
    if (str == NULL)
        return NULL;

    int slen = strlen(str);
    int wcslen;
    wchar_t wc[2*slen];  // doubling original string length should guarantee that there is enough space for wchar_t
    char *UPPER = (char *)malloc(slen+1);

    wcslen = mbsrtowcs(wc, &str, slen, NULL); // to wide-character string

    int i;
    for (i=0; i<wcslen; i++)
    {
        wc[i] = towupper(wc[i]);   // to upper-case
    }

    const wchar_t *ptr = wc;
    wcslen = wcsrtombs(UPPER, &ptr, slen, NULL);  // to character string, requires that wide-char string is constant

    if (wcslen != slen) {
        free (UPPER);
        return NULL;
    }

    UPPER[slen] = '\0';
    return UPPER;
}

/**
 * Do case insensitive string comparison for strings.
 * Works like normal strcmp(), but turns both string to uppercase before comparing
 *
 * @param str1 First string to compare.
 * @param str2 Second string to compare.
 * @return An integer greater than, equal to or less than 0, if the string pointed to by s1 is greater than, equal to or less than the string pointed to by s2 respectively.
 *         ~0 if fails to turn strings to uppercase
 */
int caseInsesitive_strcmp(const char *str1, const char *str2)
{
    int ret = ~0;
    char *STR1 = toUpperCase(str1);
    char *STR2 = toUpperCase(str2);

    if (STR1 && STR2)
    {
        ret = strcmp(STR1, STR2);
    }
    free(STR1);
    free(STR2);

    return ret;
}

/**
 * Get MAC address of given network interface.
 *
 * @param address MAC address is wrote into this.
 * @param ifname Interface name.
 * @return 1 if success, 0 if failure. MAC address is returned in address parameter.
 */
int GetMACAddressStr(unsigned char *address, int addressSize, char *ifname)
{
    struct ifreq ifr;
    int fd;
    int succeeded = 0;

    fd = get_sockfd();
    if (fd >= 0 )
    {
        strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
        if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0)
        {
            memcpy(address, ifr.ifr_hwaddr.sa_data, addressSize);
            succeeded = 1;
        }
        else
        {
            syslog(LOG_ERR, "Failure obtaining MAC address of interface %s", ifname);
            succeeded = 0;
        }
    }
    return succeeded;
}

/**
 * Get IP address assigned for given network interface.
 * If fails to get IP, sets value of address to empty string.
 *
 * @param address IP address is wrote into this.
 * @param ifname Interface name.
 * @return 1 if success, 0 if failure. IP address is returned in ipAddress parameter.
 */
int GetIPAddressFromIntfName(const char *intfName, char *ipAddress) {
  int fd;
  int rc = 0;

  if (!intfName || !ipAddress) {
    trace(1, "GetIPAddressFromIntfName: Invalid parameters intfName(%s), ipAddress(%s)",
          intfName?intfName: "NULL", ipAddress?ipAddress: "NULL");
    return 0;
  }

  if( -1 != (fd  = socket(AF_INET, SOCK_DGRAM, 0))) {
    struct ifreq ifr;

    strncpy(ifr.ifr_name, intfName, IFNAMSIZ-1);

    if(-1 != ioctl(fd,SIOCGIFADDR, &ifr)) {
			struct sockaddr_in * sockaddr = (struct sockaddr_in *) &(ifr.ifr_addr);
			struct in_addr* sin_addr = (struct in_addr*) &(sockaddr->sin_addr);
			strcpy(ipAddress, inet_ntoa(*sin_addr));
      rc = 1;
    }
    else {
      trace(1, "GetIPAddressFromIntfName: Unable to retrieve ip address for interface \"%s\"", intfName);
        }

    close(fd);
        }
  else {
    trace(1, "GetIPAddressFromIntfName: Unable to open socket");
    }

  return rc;
}

/**
 * Get connection status string used as ConnectionStatus state variable.
 * If interface has IP, status id connected. Else disconnected
 * There are also manually adjusted states Unconfigured, Connecting and Disconnecting!
 *
 * @param conStatus Connection status string is written in this.
 * @param ifname Interface name.
 * @return 1 if success, 0 if failure. Connection status is returned in conStatus parameter.
 */
int GetConnectionStatus(char *conStatus, char *ifname)
{
    char tmp[INET_ADDRSTRLEN];

    int status = GetIPAddressFromIntfName(ifname, tmp);

    if (status == 1)
        strcpy(conStatus,"Connected");
    else
        strcpy(conStatus,"Disconnected");

    return status;
}

/**
 * Check if address is either valid IP address or network host address.
 *
 * @param address String to check
 * @return 1 if it is IP or host address, 0 else.
 */
int IsIpOrDomain(char *address)
{
    int result;

    // is it IP
    regex_t re_IP;
    regcomp(&re_IP, REGEX_IP_LASTBYTE, REG_EXTENDED|REG_NOSUB);
    result = regexec(&re_IP, address, (size_t) 0, NULL, 0);
    regfree(&re_IP);
    if (result == 0) {
        return 1;
    }

    // is it domain name
    regex_t re_host;
    regcomp(&re_host, REGEX_DOMAIN_NAME, REG_EXTENDED|REG_NOSUB|REG_ICASE);
    result = regexec(&re_host, address, (size_t) 0, NULL, 0);
    regfree(&re_host);
    if (result == 0) {
        return 1;
    }

    return 0;
}

/**
 * Check if IP of control point is same as internal client address in portmapping.
 *
 * @param ICAddresscon IP of Internalclient in portmapping.
 * @param in_ad IP of control point.
 * @return 1 if match, 0 else.
 */
int ControlPointIP_equals_InternalClientIP(char *ICAddress, struct sockaddr_storage *ss)
{
    char cpAddress[INET_ADDRSTRLEN];
    int result;
    int succeeded = 0;

    if(ss->ss_family == AF_INET6) {
        //TODO
        return succeeded;
    }

    struct in_addr *in_ad=&(((struct sockaddr_in *)ss)->sin_addr);

    inet_ntop(AF_INET, in_ad, cpAddress, INET_ADDRSTRLEN);

    result = strcmp(ICAddress, cpAddress);

    // Check the compare result InternalClient IP address is same than Control Point
    if (result == 0)
    {
        succeeded = 1;
    }
    else
    {
        syslog(LOG_ERR, "CP and InternalClient IP addresees won't match:  %s %s", ICAddress, cpAddress);
        succeeded = 0;
    }

    return succeeded;
}

void upnpd_openlog(int foreground)
{
#ifdef LOG_TO_FILE
    logfileP = fopen("/tmp/upnpd.log", "w");
#else
    openlog("upnpd", LOG_CONS | LOG_NDELAY | LOG_PID | (foreground ? LOG_PERROR : 0), LOG_LOCAL6);
#endif
}

void trace(int debuglevel, const char *format, ...)
{
    va_list ap;
    va_start(ap,format);
    if (g_vars.debug>=debuglevel)
    {
#ifdef LOG_TO_FILE
        if (logfileP)
        {
            struct timeval current_time;
            gettimeofday(&current_time, NULL);

            fprintf(logfileP, "(%d.%3.3d)", (int)current_time.tv_sec, (int)current_time.tv_usec/1000);
            vfprintf(logfileP, format, ap);
            fprintf(logfileP, "\n");
            fflush(logfileP);
        }
#else
        vsyslog(LOG_DEBUG,format,ap);
#endif
    }
    va_end(ap);
}

/**
 * Check if parameter string has a wildcard character '*', or if string is '0' which might be used as wildcard
 * for port number, or if string is empty string which wildcard form of ip addresses.
 *
 * @param str String to check.
 * @return 1 if found, 0 else.
 */
int checkForWildCard(const char *str)
{
    int retVal = 0;

    if ((strchr(str, '*') != NULL) || (strcmp(str,"0") == 0) || (strcmp(str,"") == 0))
        retVal = 1;

    return retVal;
}

/**
 * Add error data to event structure used by libupnp for creating response for action request message.
 *
 * @param ca_event Response structure used for response.
 * @param errorCode Error code number.
 * @param message Error message string.
 */
void addErrorData(struct Upnp_Action_Request *ca_event, int errorCode, char* message)
{
    ca_event->ErrCode = errorCode;
    strcpy(ca_event->ErrStr, message);
    ca_event->ActionResult = NULL;
}

/**
 * Resolve if given string is acceptable as boolean value used in upnp action request messages.
 * 'yes', 'true' and '1' currently acceptable values.
 *
 * @param value String to check.
 * @return 1 if true, 0 else.
 */
int resolveBoolean(char *value)
{
    if ( strcasecmp(value, "yes") == 0 ||
         strcasecmp(value, "true") == 0 ||
         strcasecmp(value, "1") == 0 )
    {
        return 1;
    }

    return 0;
}

void ParseXMLResponse(struct Upnp_Action_Request *ca_event, const char *result_str)
{
    IXML_Document *result = NULL;

    if ((result = ixmlParseBuffer(result_str)) != NULL)
    {
        ca_event->ActionResult = result;
        ca_event->ErrCode = UPNP_E_SUCCESS;
    }
    else
    {
        trace(1, "Error parsing response to %s: %s", ca_event->ActionName, result_str);
        ca_event->ActionResult = NULL;
        ca_event->ErrCode = UPNP_SOAP_E_INVALID_ARGS;
    }
}

/**
 * Form and parse upnp action result xml.
 *
 * @param ca_event Upnp action struct.
 * @param str Format string for extra parameters.
 * @param ... Extra parameters
 */
void ParseResult( struct Upnp_Action_Request *ca_event, const char *str, ... )
{
    char *result=NULL;
    char *parameters= NULL;
    va_list arg;

    // write all parameters into one string
    va_start( arg, str );
    if (vasprintf( &parameters, str, arg ) < 0)
        trace(3, "Error: %m");
    va_end( arg );

    if (!parameters)
    {
      ca_event->ActionResult = NULL;
      ca_event->ErrCode = UPNP_E_OUTOF_MEMORY;
      return;
    }

    // and form final xml
    if (asprintf( &result, "<%sResponse xmlns:%s=\"%s\">\n%s\n</%sResponse>",
                  ixmlNode_getNodeName( ca_event->ActionRequest->n.firstChild ),
                  ixmlNode_getPrefix( ca_event->ActionRequest->n.firstChild ),
                  ixmlNode_getNamespaceURI( ca_event->ActionRequest->n.firstChild ),
                  parameters,
                  ixmlNode_getNodeName( ca_event->ActionRequest->n.firstChild ) ) < 0)
        trace(3, "error: %m");

    if (!result)
    {
      free(parameters);
      ca_event->ActionResult = NULL;
      ca_event->ErrCode = UPNP_E_OUTOF_MEMORY;
      return;
    }

    ParseXMLResponse( ca_event, result );

    free(parameters);
    free(result);
}

/**
 * Get the number of parameters included in the SOAP action
 * given in parameter
 *
 * @param doc XML document where the parameters are defined
 * @return the number of parameters
 */
int GetNbSoapParameters( IN IXML_Document * doc)
{
    IXML_NodeList *nodeList = NULL;
    IXML_Node* tmpNode = NULL;
    IXML_Node* childNode = NULL;
    int nbchild = 0;

    if(doc == NULL) return 0;

    nodeList = ixmlDocument_getElementsByTagName( doc,
        ixmlNode_getNodeName( doc->n.firstChild ) );

    if ( nodeList )
    {
        if ( ( tmpNode = ixmlNodeList_item( nodeList, 0 ) ) )
        {
            ixmlNodeList_free( nodeList );
            nodeList = ixmlNode_getChildNodes(tmpNode);
            nbchild = ixmlNodeList_length(nodeList);
            if(nbchild > 0){
                childNode = ixmlNode_getFirstChild(tmpNode);
                if(childNode == NULL)
                  nbchild = 0;
            }
        }
    }

    if ( nodeList )
        ixmlNodeList_free( nodeList );

    return nbchild;
}

/**
 * test if the string is an Integer
 *
 * @param string the string to test
 * @return 1 if the string is only composed of digit,
 * 0 otherwise
 */
int isStringInteger(char * string)
{
    int i = 0;
    for(; i < strlen(string) ; i++) {
        if(!isdigit(string[i])) return 0;
    }
    return 1;
}

/**
 * Resolve up/down status of given network interface and insert it into given string.
 * Status is up if interface is listed in /proc/net/dev_mcast -file, else down.
 *
 * @param ethLinkStatus Pointer to string where status is wrote.
 * @param iface Network interface name.
 * @return 0 if status is up, 1 if down or failed to open dev_mcast file.
 */
int setEthernetLinkStatus(char *ethLinkStatus, char *iface)
{
    FILE *fp;
    char str[60];

    // check from dev_mcast if interface is up (up if listed in file)
    // This could be done "finer" with reading registers from socket. Check from ifconfig.c or mii-tool.c. Do if nothing better to do.
    if((fp = fopen("/proc/net/dev_mcast", "r"))==NULL) {
        syslog(LOG_ERR, "Cannot open /proc/net/dev_mcast");
        return 1;
    }

    while(!feof(fp)) {
        if(fgets(str,60,fp) && strstr(str,iface))
        {
            strcpy(ethLinkStatus,"Up");
            fclose(fp);
            return 0;
        }
    }

    fclose(fp);
    strcpy(ethLinkStatus,"Down");
    return 1;
}

/**
 * Read integer value from given file. File should only contain this one numerical value.
 *
 * @param file Name of file to read.
 * @param iface Network interface name.
 * @return Value read from file. -1 if fails to open file, -2 if no value found from file.
 */
int readIntFromFile(char *file)
{
    FILE *fp;
    int value = -1;

    trace(3,"Read integer value from %s", file);

    if((fp = fopen(file, "r"))==NULL) {
        return -1;
    }

    while(!feof(fp)) {
        if (fscanf(fp,"%d", &value) < 0)
            trace(3,"Error: %m");

        if (value > -1)
        {
            fclose(fp);
            return value;
        }
    }
    fclose(fp);
    return -2;
}

/**
 * Kill DHCP client. After killing check that IP of iface has been released.
 *
 * @param iface Network interface name.
 * @return 1 if DHCP client is killed and IP released, 0 else.
 */
int killDHCPClient(char *iface)
{
    char tmp[30];
    int pid;

    trace(2,"Killing DHCP client...");
    snprintf(tmp, 30, "/var/run/%s.pid", iface);
    pid = readIntFromFile(tmp);
    if (pid > -1)
    {
        snprintf(tmp, 30, "kill %d", pid);
        trace(3,"system(%s)",tmp);
        if (system(tmp) < 0)
            trace(3, "error executing %s: %m", tmp);
    }
    else
    {
        // brute force
        trace(3,"No PID file available for %s of %s",g_vars.dhcpc,iface);
        snprintf(tmp, 30, "killall -9 %s", g_vars.dhcpc);
        trace(3,"system(%s)",tmp);
        if (system(tmp) < 0)
            trace(3, "error executing %s: %m", tmp);
    }

    sleep(2); // wait that IP is released

    if (!GetIPAddressFromIntfName(iface, tmp))
    {
        trace(3,"Success IP of %s is released",iface);
        return 1;
    }
    else
    {
        trace(3,"Failure IP of %s: %s",iface,tmp);
        return 0;
    }
}

/**
 * Start DHCP client. After starting check that iface has IP.
 *
 * @param iface Network interface name.
 * @return 1 if DHCP client is started and iface has IP, 0 else.
 */
int startDHCPClient(char *iface)
{
    char tmp[100];

    trace(2,"Starting DHCP client...");
    if (strcmp(g_vars.dhcpc,"dhclient") == 0)
    {
        if (snprintf(tmp, 100, "%s -pf /var/run/%s.pid %s", g_vars.dhcpc, iface, iface))
            trace(3, "error: %m");
    }
    else
    {
        if (snprintf(tmp, 100, "%s -i %s -R -p /var/run/%s.pid", g_vars.dhcpc, iface, iface) < 0)
            trace(3, "error: %m");
    }
    trace(3,"system(%s)",tmp);
    if (system(tmp) < 0)
        trace(3, "error executing %s: %m", tmp);

    sleep(2); // wait that IP is acquired

    if (GetIPAddressFromIntfName(iface, tmp))
    {
        trace(3,"Success IP of %s: %s",iface,tmp);
        return 1;
    }
    else
    {
        trace(3,"Failure %s doens't have IP",iface);
        return 0;
    }
}

/**
 * Release IP address of given interface.
 *
 * @param iface Network interface name.
 * @return 1 if iface doesn't have IP, 0 else.
 */
int releaseIP(char *iface)
{
    char tmp[INET6_ADDRSTRLEN];
    int success = 0;

    // check does IP exist
    if (!GetIPAddressFromIntfName(iface, tmp))
        return 1;

    // if used dhcp client is dhclient, use -r parameter to release IP
    if (strcmp(g_vars.dhcpc,"dhclient") == 0)
    {
        char tmp[50];
        trace(2,"Releasing IP...");
        snprintf(tmp, 50, "%s -r %s", g_vars.dhcpc, iface);
        trace(3,"system(%s)",tmp);
        if (system(tmp) < 0)
            trace(3, "%m");

        sleep(2); // wait that IP is released

        if (!GetIPAddressFromIntfName(iface, tmp))
        {
            trace(3,"Success IP of %s is released",iface);
            return 1;
        }
        else
        {
            trace(3,"Failure IP of %s: %s",iface,tmp);
            return 0;
        }
    }
    // if used dhcp client is udhcpc, IP is released when udhcpc is killed if udhcpc was started with -R parameter
    else
    {
        // kill already running udhcpc-client for given iface and check if IP was released
        if (killDHCPClient(iface))
            success = 1; //OK
        else
        {
            // start udhcpc-clientdaemon with parameter -R which will release IP after quiting daemon
            startDHCPClient(iface);

            // then kill udhcpc-client running. Now there shouldn't be IP anymore.
            if(killDHCPClient(iface))
                success = 1;
        }
    }
    return success;
}



//-----------------------------------------------------------------------------
//
//                      Common extensions for ixml
//
//-----------------------------------------------------------------------------
/**
 * Get document item which is at position index in nodelist (all nodes with same name item).
 * Index 0 means first, 1 second, etc.
 *
 * @param doc XML document where item is fetched.
 * @param item Name of xml-node to fetch.
 * @param index Which one of nodes with same name is selected.
 * @return Value of desired node.
 */
char* GetDocumentItem(IXML_Document * doc, const char *item, int index)
{
    IXML_NodeList *nodeList = NULL;
    IXML_Node *textNode = NULL;
    IXML_Node *tmpNode = NULL;

    char *ret = NULL;

    nodeList = ixmlDocument_getElementsByTagName( doc, ( char * )item );

    if ( nodeList )
    {
        if ( ( tmpNode = ixmlNodeList_item( nodeList, index ) ) )
        {
            textNode = ixmlNode_getFirstChild( tmpNode );
            if (textNode != NULL)
            {
                ret = strdup( GET_STRING_VALUE(ixmlNode_getNodeValue( textNode )) );
            }
            // if desired node exist, but textNode is NULL, then value of node propably is ""
            else
                ret = strdup("");
        }
    }

    ixmlNodeList_free( nodeList );
    return ret;
}

/**
 * Get first document item in nodelist with name given in item parameter.
 *
 * @param doc XML document where item is fetched.
 * @param item Name of xml-node to fetch.
 * @return Value of desired node.
 */
char* GetFirstDocumentItem( IN IXML_Document * doc,
                            IN const char *item )
{
    return GetDocumentItem(doc,item,0);
}

#ifdef CONFIG_SAH_SERVICES_LINUX_IGD_MAC_ADDRESS_REMEMBERING_ENABLED

/** Retrieve the MAC address matching an IPv4 address by reading file /proc/net/arp.
 *
 * @param ipaddress, the ip address from which we should retrieve the corresponding MAC
 * @param mac, buffer of size MAC_ADDRESS_STRING_LENGTH to store the data in
 * @return true if success, false if error
 */
int GetMACAddressFromProc(const char *ipaddress, char *mac)
{
    int rc = 0;
    FILE *proc = NULL;
    char ipaddr[128];
    char hwaddr[128];
    char mask[128];
    char device[128];
    char line[128];
    int hwtype = 0;
    int flags = 0;
    int count = 0;

    if (!ipaddress || !mac) {
        trace(1, "Invalid input");
        goto out;
    }

    memset( mac, '\0', MAC_ADDRESS_STRING_LENGTH);

    if (NULL == ( proc = fopen(PROC_ARP_FILE, "r")))
    {
        trace(1, "Failed to open %s", PROC_ARP_FILE);
        goto out;
    }

    if (fgets(line, sizeof(line), proc) != NULL) {  // skip the header
        trace(3, "Skipping header");
    }

    // IP address       HW type     Flags       HW address            Mask     Device
    // 192.168.18.1     0x1         0x2         00:21:1b:95:95:c8     *        eth0
    while (fgets(line, sizeof(line), proc) != NULL) {
        if (sscanf(line, "%s 0x%x 0x%x %s %s %s\n", ipaddr, &hwtype, &flags, hwaddr, mask, device) == 6 ) {
            if (strcmp(ipaddress, ipaddr) == 0) {
                strncpy(mac, hwaddr, 17);
                count++;
            }
        }
    }

    if (count == 1) {
        rc = 1;        
    }
    else {
        trace(1, "Error: found %d MAC addresses for %s", count, ipaddress);
    }

    fclose(proc);
out:
    return rc;
}

int ControlPointMAC_equals_portmappingMAC(struct sockaddr_storage *ss, const char *portmappingMAC)
{
    int rc = 0;    
    char cpAddress[INET_ADDRSTRLEN];
    char controlPointMAC[MAC_ADDRESS_STRING_LENGTH];

    if(ss->ss_family == AF_INET6) {
        goto out;
    }

    struct in_addr *in_ad=&(((struct sockaddr_in *)ss)->sin_addr);
    inet_ntop(AF_INET, in_ad, cpAddress, INET_ADDRSTRLEN);

    if (!portmappingMAC) {
        trace(1, "Invalid input");
        goto out;
    }

    if (!GetMACAddressFromProc(cpAddress, controlPointMAC)) {
        goto out;
    }

    trace(4, "IPv4=%s has MAC=%s", cpAddress, controlPointMAC);

    if (strcmp(controlPointMAC, portmappingMAC) == 0) {
        rc = 1;
    }
    else {
        trace(1, "MAC address mismatch, portmapping MAC(%s) - requester MAC(%s)", controlPointMAC, portmappingMAC);
    }

out:
    return rc;
}

#endif // CONFIG_SAH_SERVICES_LINUX_IGD_MAC_ADDRESS_REMEMBERING_ENABLED
